/* Definitions for PDP-10 target machine using MACRO assembler.
   Copyright (C) 2001 Free Software Foundation, Inc.
   Contributed by Lars Brinkhoff (lars@nocrew.org), funded by XKL, LLC.
   
This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#ifndef __PDP10_MACRO_H__
#define __PDP10_MACRO_H__

#define ASM_FILE_START(STREAM) macro_file_start (STREAM)

#define ASM_FILE_END(STREAM) macro_file_end (STREAM)

#define ASM_COMMENT_START ";"

#define ASM_APP_ON ""

#define ASM_APP_OFF ""

#define ASM_OUTPUT_SECTION_NAME(STREAM, DECL, NAME, RELOC)	\
do {								\
  fputs ("\t.PSECT\t", STREAM);					\
  fputs (NAME, STREAM);						\
  fputc ('\n', STREAM);						\
} while (0)

#define IS_ASM_LOGICAL_LINE_SEPARATOR(C) 0

#define TEXT_SECTION_ASM_OP \
  (TARGET_EXTENDED ? "\t.PSECT .text/ronly" : "")

#define DATA_SECTION_ASM_OP \
  (TARGET_EXTENDED ? "\t.PSECT .data/rwrite" : "")

#define RODATA_SECTION_ASM_OP \
  (TARGET_EXTENDED ? "\t.PSECT .rodata/ronly" : "")

#define ASM_SECTION_END(STREAM)			\
do {						\
  if (TARGET_EXTENDED)				\
    {						\
      pdp10_output_byte (asm_out_file, 0, 0);	\
      fputs ("\t.ENDPS\n", STREAM);		\
    }						\
} while (0)

#define EXTRA_SECTIONS in_rodata

#define EXTRA_SECTION_FUNCTIONS \
  RODATA_SECTION_FUNCTION

#define RODATA_SECTION_FUNCTION				\
void							\
pdp10_rodata_section ()					\
{							\
  if (in_section != in_rodata)				\
    {							\
      ASM_SECTION_END (asm_out_file);		\
      fputs (RODATA_SECTION_ASM_OP, asm_out_file);	\
      fputc ('\n', asm_out_file);			\
      in_section = in_rodata;				\
    }							\
}

#define READONLY_DATA_SECTION pdp10_rodata_section

#define ASM_OUTPUT_ALIGN(STREAM, POWER) \
  do { } while (0)

#define ASM_OUTPUT_INT(STREAM, EXP)		\
  pdp10_output_byte (STREAM, 0, 0);		\
  fputc ('\t', STREAM);				\
  pdp10_print_operand (STREAM, EXP, 0);		\
  fputc ('\n', STREAM);

#define ASM_OUTPUT_CHAR(STREAM, EXP) \
  pdp10_output_byte (STREAM, EXP, 9)

#define ASM_OUTPUT_SHORT(STREAM, EXP) \
  pdp10_output_byte (STREAM, EXP, 18)

#define ASM_OUTPUT_SKIP(STREAM, NBYTES)		\
do {						\
  int i = (NBYTES);				\
  while (i--)					\
  pdp10_output_byte (STREAM, const0_rtx, 9);	\
} while (0)

#define ASM_OUTPUT_ASCII(STREAM, PTR, LEN) \
  asm_output_asciz (STREAM, PTR, LEN)

#define ASM_GLOBALIZE_LABEL(STREAM, NAME) \
  macro_globalize_label (STREAM, NAME)

#define ASM_OUTPUT_EXTERNAL(STREAM, DECL, NAME)		\
do {							\
  if (strcmp (NAME, "__builtin_constant_p") == 0)	\
    break;						\
  pdp10_output_byte (STREAM, 0, 0);			\
  fputs ("\tEXTERN\t", STREAM),				\
  assemble_name (STREAM, NAME);				\
  fputc ('\n', STREAM);					\
} while (0)

#define ASM_OUTPUT_EXTERNAL_LIBCALL(STREAM, SYMREF)	\
  fputs ("\tEXTERN\t", STREAM),				\
  assemble_name (STREAM, XSTR (SYMREF, 0));		\
  fputc ('\n', STREAM)

#define ASM_OUTPUT_COMMON(STREAM, NAME, SIZE, ROUNDED)	\
do {							\
  data_section ();					\
  fputs ("\tINTERN\t", STREAM);				\
  assemble_name (STREAM, NAME);				\
  fputc ('\n', STREAM);					\
  assemble_name (STREAM, NAME);				\
  fputs (":\n\tBLOCK\t", STREAM);			\
  fprintf (STREAM, "%o\n", (ROUNDED + 3) / 4);		\
} while (0)

#define ASM_OUTPUT_ALIGNED_LOCAL(STREAM, NAME, SIZE, ALIGN)	\
do {								\
  ASM_OUTPUT_LABEL (STREAM, NAME);				\
  fprintf (STREAM, "\tBLOCK\t%o\n", (((SIZE) + 3) / 4));	\
} while (0)

#define ASM_IDENTIFY_GCC(FILE)				\
do {							\
  const char *gnu_comp = "gnu.";			\
  const char *lang = lang_identify ();			\
  int len = strlen (lang) + sizeof (gnu_comp) + 1;	\
  char *s = (char *) alloca (len);			\
  fputs ("gcc2.compiled:\n", FILE);			\
  sprintf (s, "%s%s", gnu_comp, lang);			\
  ASM_OUTPUT_LABEL (asm_out_file, s);			\
} while (0)

#define ASM_OUTPUT_DEF(FILE, LABEL, VALUE)	\
do {						\
  assemble_name (FILE, LABEL);			\
  fputs ("==", FILE);				\
  assemble_name (FILE, VALUE);			\
  fprintf (FILE, "\n");				\
} while (0)

#include "pdp10.h"

#endif /* __PDP10_MACRO_H__ */
