/* Prototypes for exported functions defined in pdp10.c.
   Copyright (C) 2001 Free Software Foundation, Inc.
   Contributed by Lars Brinkhoff (lars@nocrew.org), funded by XKL, LLC.
   
This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#ifndef GCC_PDP10_PROTOS_H
#define GCC_PDP10_PROTOS_H

#ifndef HOST_WIDE_INT
#include "hwint.h"
#endif

#ifndef HAVE_MACHINE_MODES
#include "machmode.h"
#endif

#ifdef RTX_CODE
#define Rtx rtx
#define Rcode enum rtx_code
#else
#define Rtx void *
#define Rcode int
#endif

#ifndef TREE_CODE
union tree_node;
typedef union tree_node *tree;
#endif

#define R_PARAMS PARAMS
#define T_PARAMS PARAMS
#define RT_PARAMS PARAMS

/* Run-time Target Specification.  */
extern void	pdp10_override_options PARAMS ((void));

/* Storage Layout.  */
extern Rtx	pdp10_align_stack_pointer R_PARAMS ((Rtx x, unsigned align));

/* Register Classes.  */
extern int	pdp10_const_ok_for_letter_p PARAMS ((HOST_WIDE_INT, int));
extern int	pdp10_const_double_ok_for_letter_p R_PARAMS ((Rtx x, int c));
extern int	pdp10_extra_constraint PARAMS ((Rtx x, int c));

/* Stack Layout and Calling Conventions.  */
extern int	pdp10_starting_frame_offset PARAMS ((void));
extern int	pdp10_stack_pointer_offset PARAMS ((void));
extern int	pdp10_first_parm_offset PARAMS ((void));
extern int	pdp10_stack_dynamic_offset PARAMS ((void));
extern int	pdp10_initial_frame_pointer_offset PARAMS ((void));
extern void	pdp10_output_func_prologue PARAMS ((FILE *, int));
extern void	pdp10_output_func_epilogue PARAMS ((FILE *, int));
extern void	pdp10_function_profiler PARAMS ((FILE *, int));

/* Addressing Modes.  */
extern int	pdp10_legitimate_address_p R_PARAMS ((enum machine_mode mode,
						      Rtx x,
						      int strict));
extern Rtx	pdp10_legitimize_address R_PARAMS ((Rtx x,
						    Rtx oldx,
						    enum machine_mode mode));

/* Describing Relative Costs of Operations.  */
extern int	pdp10_const_int_costs R_PARAMS ((HOST_WIDE_INT i,
						 Rcode outer_code));
extern int	pdp10_const_double_costs R_PARAMS ((Rtx x,
						    Rcode code,
						    Rcode outer_code));
extern int	pdp10_operand_cost R_PARAMS ((Rtx x,
					      Rcode outer_code));
extern int	pdp10_address_cost R_PARAMS ((Rtx address));

/* Dividing the Output into Sections.  */
extern void	pdp10_encode_section_info T_PARAMS ((tree decl));

/* Defining the Output Assembler Language.  */
extern void	pdp10_asm_declare_object_name T_PARAMS ((FILE *stream,
						   const char *name,
						   tree decl));
extern void	pdp10_asm_output_labelref PARAMS ((FILE *file,
						   const char *name));
extern void	pdp10_asm_output_symbol_ref R_PARAMS ((FILE *stream, Rtx sym));
extern const char *
		pdp10_asm_output_opcode PARAMS ((FILE *stream,
						 const char *ptr));
extern void	pdp10_print_operand R_PARAMS ((FILE *stream, Rtx x, int code));
extern void	pdp10_print_operand_address R_PARAMS ((FILE *stream, Rtx x,
						       enum machine_mode));

/* Built-in Functions.  */
extern void	pdp10_init_builtins PARAMS ((void));
extern Rtx	pdp10_expand_builtin RT_PARAMS ((tree exp,
						 Rtx target,
						 Rtx subtarget,
						 enum machine_mode mode,
						 int ignore));

/* Machine-specific Attributes.  */
extern int	pdp10_valid_machine_attribute T_PARAMS ((tree type,
							 tree attributes,
							 tree identifier,
							 tree args));

/* Data Movement.  */
extern const char *
		pdp10_output_movhi R_PARAMS ((Rtx *operands, int sign_extend));
extern void	pdp10_expand_extv R_PARAMS ((Rtx *operands));
extern const char *
		pdp10_output_extv R_PARAMS ((Rtx *operands));
extern const char *
		pdp10_output_extzv R_PARAMS ((Rtx *operands));
extern const char *
		pdp10_output_insv R_PARAMS ((Rtx *operands));
extern int	pdp10_expand_movstrsi R_PARAMS ((Rtx *operands));
extern int	pdp10_expand_clrstrsi R_PARAMS ((Rtx *operands));
extern int	pdp10_expand_cmpstrsi R_PARAMS ((Rtx *operands));

/* Integer Arithmetic.  */
extern int	pdp10_expand_subsi3 R_PARAMS ((Rtx *operands));

/* Pointer Arithmetic.  */
extern void	pdp10_expand_addpsi3 R_PARAMS ((Rtx *operands));
extern void	pdp10_expand_subpsi3 R_PARAMS ((Rtx *operands));

/* Function Prologue and Epilogue.  */
extern void	pdp10_expand_prologue PARAMS ((void));
extern void	pdp10_expand_epilogue PARAMS ((void));
extern void	pdp10_expand_sibcall_epilogue PARAMS ((void));




#ifdef RTX_CODE
extern void	pdp10_output_byte PARAMS ((FILE *stream, Rtx exp,
					   int byte_size));
extern Rtx	pdp10_compare_op0, pdp10_compare_op1;
#endif

extern void	macro_file_start PARAMS ((FILE *));
extern void	macro_file_end PARAMS ((FILE *));

extern void	output_ascii_pseudo_op PARAMS ((FILE *stream,
						unsigned char *p,
						int len));
extern void	asm_output_ascii PARAMS ((FILE *stream, const char *ptr,
					  int len));
extern void	asm_output_asciz PARAMS ((FILE *stream, const char *ptr,
					  int len));
extern void	macro_globalize_label PARAMS ((FILE *stream,
					       const char *name));
extern void	pdp10_rodata_section PARAMS ((void));

#endif /* GCC_PDP10_PROTOS_H */
