/* Output routines for the PDP-10.
   Copyright (C) 2001 Free Software Foundation, Inc.
   Contributed by Lars Brinkhoff (lars@nocrew.org), funded by XKL, LLC.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


/**********************************************************************

	Index

**********************************************************************/

/*	  Front Page
	  Index
	  Includes
	  To-do List
	  Miscellaneous
	pdp10.h support:
	  Run-time Target Specification
	  Storage Layout
	  Layout of Source Language Data Types
	  Register Classes
	  Stack Layout and Calling Conventions
	  Addressing Modes
	  Dividing the Output into Sections
	  Describing Relative Costs of Operations
	  Defining the Output Assembler Language
	  Special Predicates
	  Built-in Functions
	  Machine-specific Attributes
	  MACRO, MIDAS, and FAIL
	pdp10.md support:
	  Data Movement
	  Pointer Arithmetic
	  Function Prologue and Epilogue */
    

/**********************************************************************

	Includes

**********************************************************************/

#include "config.h"
#include "system.h"
#include "rtl.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "real.h"
#include "insn-config.h"
#include "conditions.h"
#include "insn-flags.h"
#include "output.h"
#include "insn-attr.h"
#include "flags.h"
#include "reload.h"
#include "tree.h"
#include "expr.h"
#include "toplev.h"
#include "recog.h"
#include "function.h"
#include "ggc.h"
#include "pdp10-protos.h"
#include "c-tree.h"
#include "hash.h"
#include <ctype.h>


/**********************************************************************

	To-do List

**********************************************************************/

/* Candidates for inlining once the code settles:
   pdp10_align_stack_pointer, const_ok_for_letter_p,
   const_double_ok_for_letter_p, extra_constraint,
   starting_frame_offset, stack_pointer_offset, first_parm_offset,
   stack_dynamic_offset, initial_frame_pointer_offset,
   output_func_prologue, output_func_epilogue, function_profiler,
   legitimate_address_p, legitimize_address, pdp10_const_int_costs,
   pdp10_const_double_costs, pdp10_operand_cost, encode_section_info,
   asm_declare_object_name, asm_output_labelref,
   asm_output_symbol_ref, print_operand_address.  */

/* Investigate whether it's safe for expand_blt to use BLT when
   operands[2] is not constant.  */


/**********************************************************************

	Miscellaneous

**********************************************************************/

#define SAVE		1
#define RESTORE		0

#define WORD_MODE_P(MODE) \
  (GET_MODE_SIZE (MODE) >= UNITS_PER_WORD || (MODE) == BLKmode)
#define BYTE_MODE_P(MODE) (!WORD_MODE_P (MODE))

//#define WORD_TYPE_P(TYPE) ((int_size_in_bytes (TYPE) % UNITS_PER_WORD) == 0)
//#define WORD_TYPE_P(TYPE) ((TYPE_ALIGN (TYPE) % UNITS_PER_WORD) == 0)
#define WORD_TYPE_P(TYPE) (pointer_bytesize (TYPE) == 36)
#define BYTE_TYPE_P(TYPE) (!WORD_TYPE_P (TYPE))

#define VALID_BYTE_SIZE_P(SIZE)						    \
 ((SIZE) == 6 || (SIZE) == 7 || (SIZE) == 8 || (SIZE) == 9 || (SIZE) == 18)

#define GEN_MEM_SP_PLUS(mode, offset)			\
  gen_rtx_MEM (mode,					\
	       gen_rtx_PLUS (Pmode,			\
			     stack_pointer_rtx,		\
			     GEN_INT (offset)))

#define lengthof(A) (sizeof (A) / sizeof (A)[0])

static void	pdp10_print_number PARAMS ((FILE *stream,
					    rtx x,
					    int div,
					    int bits,
					    int zero));

#define pdp10_print_address(stream, x) \
  pdp10_print_number (stream, x, UNITS_PER_WORD, TARGET_EXTENDED ? 30 : 18, 0)
#define pdp10_print_offset30(stream, x) \
  pdp10_print_number (stream, x, UNITS_PER_WORD, TARGET_EXTENDED ? -30 : -18,1)
#define pdp10_print_offset(stream, x) \
  pdp10_print_number (stream, x, UNITS_PER_WORD, -18, 1)
#define pdp10_print_integer(stream, x) \
  pdp10_print_number (stream, x, 1, -36, 0)
#define pdp10_print_unsigned(stream, x) \
  pdp10_print_number (stream, x, 1, 36, 0)


/**********************************************************************

	Run-time Target Specifiation

**********************************************************************/

static void	normalize_name PARAMS ((char *name));

/* Receives a pointer to the value of the -mcpu option.  */
char *pdp10_cpu;

static struct { const char *name; int cpu; int enable; int disable; }
cpu_table[] =
{
  /* name,	cpu,		enable,		disable */

  { "166",	CPU_166,	0,		CPU_EXTENDED	},
  { "ka10",	CPU_KA10,	0,		CPU_EXTENDED	},
  { "ki10",	CPU_KI10,	0,		CPU_EXTENDED	},
  { "kl10",	CPU_KL10,	CPU_EXTENDED,	0		},
  { "ks10",	CPU_KS10,	0,		CPU_EXTENDED	},
#if 0
  { "kc10",	CPU_KC10,	CPU_EXTENDED,	0		},
  { "maxc",	CPU_MAXC,	0,		CPU_EXTENDED	},
  { "f1",	CPU_F1,		0,		CPU_EXTENDED	},
  { "f2",	CPU_F2,		0,		CPU_EXTENDED	},
  { "f3",	CPU_F3,		0,		CPU_EXTENDED	},
  { "f4",	CPU_F4,		0,		CPU_EXTENDED	},
  { "f5",	CPU_F5,		0,		0		},
  { "26",	CPU_26,		0,		0		},
  { "26kl",	CPU_26KL,	0,		0		},
  { "sc20",	CPU_SC20,	CPU_EXTENDED,	0		},
  { "sc25",	CPU_SC25,	CPU_EXTENDED,	0		},
  { "sc30m",	CPU_SC30M,	CPU_EXTENDED,	0		},
  { "sc40",	CPU_SC40,	CPU_EXTENDED,	0		},
#endif
  { "xkl1",	CPU_XKL1,	CPU_EXTENDED,	0		},
  { "xkl2",	CPU_XKL2,	CPU_EXTENDED,	0		},
};

static struct { const char *alias; const char *name; }
cpu_alias[] =
{
  { "pdp6",	"166"	},
  { "ka",	"ka10"	},
  { "1010",	"ka10"	},
  { "1020",	"ka10"	},
  { "1030",	"ka10"	},
  { "1040",	"ka10"	},
  { "1050",	"ka10"	},
  { "1055",	"ka10"	},
  { "ki",	"ki10"	},
  { "1070",	"ki10"	},
  { "1077",	"ki10"	},
  { "kl",	"kl10"	},
  { "1080",	"kl10"	},
  { "1088",	"kl10"	},
  { "1090",	"kl10"	},
  { "1090t",	"kl10"	},
  { "1091",	"kl10"	},
  { "1095",	"kl10"	},
  { "1099",	"kl10"	},
  { "2040",	"kl10"	},
  { "2050",	"kl10"	},
  { "2060",	"kl10"	},
  { "2065",	"kl10"	},
  { "ks",	"ks10"	},
  { "2020",	"ks10"	},
/*{ "minnow",	"?"	},*/
/*{ "dolphin",	"?"	},*/
  { "kc",	"kc10"	},
  { "2080",	"kc10"	},
  { "4050",	"kc10"	},
  { "jupiter",	"kc10"	},
  { "superfoonly","f1"	},
  { "sc30",	"sc30m"	},
  { "xkl",	"xkl1"	},
  { "td1",	"xkl1"	},
  { "toad",	"xkl1"	},
  { "toad1",	"xkl1"	},
};

static void
normalize_name (name)
     char *name;
{
  char *p;

  for (p = name; *p != 0; p++)
    {
      if (*p != '-')
	*name++ = tolower (*p);
    }
  *name = 0;
}

/* Called by the OVERRIDE_OPTIONS macro.  */
void
pdp10_override_options ()
{
  unsigned int i;

  if (pdp10_cpu != NULL)
    {
      normalize_name (pdp10_cpu);

      for (i = 0; i < lengthof (cpu_alias); i++)
	{
	  if (strcmp (pdp10_cpu, cpu_alias[i].alias) == 0)
	    {
	      pdp10_cpu = (char *)cpu_alias[i].name;
	      break;
	    }
	}

      for (i = 0; i < lengthof (cpu_table); i++)
	{
	  if (strcmp (pdp10_cpu, cpu_table[i].name) == 0)
	    {
	      target_flags = (target_flags & ~CPU_MASK) | cpu_table[i].cpu;
	      target_flags |= cpu_table[i].enable;
	      target_flags &= ~cpu_table[i].disable;
	      break;
	    }
	}

      if (i == lengthof (cpu_table))
	fprintf (stderr, "Unknown CPU: %s\n", pdp10_cpu);
    }
}


/**********************************************************************

	Storage Layout

**********************************************************************/

/* Called by the ALIGN_STACK_POINTER macro.  */
rtx
pdp10_align_stack_pointer (x, align)
     rtx x;
     unsigned align;
{
  if (align % UNITS_PER_WORD != 0)
    abort ();

  align /= UNITS_PER_WORD;
  if (align > 1)
    {
      x = expand_binop (Pmode, add_optab, x, GEN_INT (align - 1), NULL_RTX,
			1, OPTAB_LIB_WIDEN);
      x = expand_divmod (0, TRUNC_DIV_EXPR, Pmode, x, GEN_INT (align),
			 NULL_RTX, 1);
      x = expand_mult (Pmode, x, GEN_INT (2), NULL_RTX, 1);
    }

  return x;
}


/**********************************************************************

	Register Classes

**********************************************************************/

/* Called by the CONST_OK_FOR_LETTER_P macro.  */
int
pdp10_const_ok_for_letter_p (value, c)
     HOST_WIDE_INT value;
     int c;
{
  switch (c)
    {
    case 'I': return value >= 0 && value <= 0777777;
    case 'J': return value <= 0 && value >= -0777777;
    case 'K': return value >= -0400000 && value <= 0377777;
    case 'L': return ((value & 0777777) == 0
		      && pdp10_const_ok_for_letter_p (value >> 18, 'I'));
    case 'N': return pdp10_const_ok_for_letter_p (~value, 'L');
    case 'O': return value == 0;
    case 'P': return pdp10_const_ok_for_letter_p (~value, 'I');
    }
  return 0;
}

/* Called by the CONST_DOUBLE_OK_FOR_LETTER_P macro.  */
int
pdp10_const_double_ok_for_letter_p (x, c)
     rtx x;
     int c;
{
  REAL_VALUE_TYPE r;
  long y[3];

  REAL_VALUE_FROM_CONST_DOUBLE (r, x);
  REAL_VALUE_TO_TARGET_DOUBLE (r, y);

  switch (GET_MODE (x))
    {
    case SFmode:
      return
	c == 'G'
	&& (y[0] & 0x00003fffUL) == 0
	&& (y[1] & 0xf0000000UL) == 0;
    case DFmode:
      return
	c == 'G'
	&& (y[0] & 0x00003fffUL) == 0
	&& (y[1] & 0xffffffffUL) == 0
	&& (y[2] & 0xff000000UL) == 0;
    default:
      abort ();
    }
}

/* Called by the EXTRA_CONSTRAINT macro.  */
int
pdp10_extra_constraint (x, c)
     rtx x;
     int c;
{
  switch (c)
    {
    case 'S':
      return
	!TARGET_EXTENDED && GET_CODE (x) == SYMBOL_REF;

    default:
      abort ();
    }
}


/**********************************************************************

	Stack Layout and Calling Conventions

**********************************************************************/

static int	clobbered_regs PARAMS ((int *clobbered));

/* Total size of all callee-saved registers clobbered by the current
   function.  */
static int current_function_clobbered_size;

/* Total size of local stack variables needed by the current function.  */
static int current_function_frame_size;

/* Called by the STARTING_FRAME_OFFSET macro.  */
int
pdp10_starting_frame_offset ()
{
  return current_function_clobbered_size;
}

/* Called by the STACK_POINTER_OFFSET macro.  */
int
pdp10_stack_pointer_offset ()
{
  return UNITS_PER_WORD;
}

/* Called by the FIRST_PARM_OFFSET macro.  */
int
pdp10_first_parm_offset ()
{
  return UNITS_PER_WORD;
}

/* Called by the STACK_DYNAMIC_OFFSET macro.  */
int
pdp10_stack_dynamic_offset ()
{
  return 0;
}

/* Called by the INITIAL_FRAME_POINTER_OFFSET macro.  */
int
pdp10_initial_frame_pointer_offset ()
{
  return -(current_function_clobbered_size +
	   current_function_frame_size +
	   current_function_outgoing_args_size + UNITS_PER_WORD);
}

/* Stack frame layout:

                ...
         [     arg 6    ]
         [     arg 5    ]
         [    (arg 4)   ]
         [    (arg 3)   ]   The  register arguments are pushed to stack
         [    (arg 2)   ]   only if necessary, in a vararg function.
   fp -> [    (arg 1)   ]   Stack space is always reserved, though.
         [ saved reg  5 ]
         [ saved reg  6 ]
         [ saved reg  7 ]
         [ saved reg 10 ]
         [ saved reg 11 ]   Registers 5 - 16 are saved if clobbered by
         [ saved reg 12 ]   the function.
         [ saved reg 13 ]
         [ saved reg 14 ]
         [ saved reg 15 ]
         [ saved reg 16 ]
         [ frame slot 1 ]
         [ frame slot 2 ]   Local variables.
         [ frame slot 3 ]
               ...
         [outgoing arg 3]
         [outgoing arg 2]   Reserved space for outgoing arguments.
   sp -> [outgoing arg 1]
*/

/* Stack layout, before function prologue:

                ...
         [     arg 6    ]
         [     arg 5    ]
   sp -> [return address]

   Stack layout, after fixed-arg function prologue:

                ...
         [     arg 6    ]
  fp? -> [     arg 5    ]   frame pointer where?
	 [return address]
         [ saved reg  5 ]
         [ saved reg  6 ]
         [ saved reg  7 ]
         [ saved reg 10 ]
         [ saved reg 11 ]   Registers 5 - 16 are saved if clobbered by
         [ saved reg 12 ]   the function.
         [ saved reg 13 ]
         [ saved reg 14 ]
         [ saved reg 15 ]
         [ saved reg 16 ]
         [ frame slot 1 ]
         [ frame slot 2 ]   Local variables.
         [ frame slot 3 ]
               ...
         [outgoing arg 3]
         [outgoing arg 2]   Reserved space for outgoing arguments.
   sp -> [outgoing arg 1]

   Stack layout, after vararg function prologue:

                ...
         [     arg 6    ]
         [     arg 5    ]
         [     arg 4    ]   Args 1-4 has been pushed to stack.
         [     arg 3    ]
         [     arg 2    ]
   fp -> [     arg 1    ]
	 [return address]   Return address has been moved.
         [ saved reg  5 ]
         [ saved reg  6 ]
         [ saved reg  7 ]
         [ saved reg 10 ]
         [ saved reg 11 ]   Registers 5 - 16 are saved if clobbered by
         [ saved reg 12 ]   the function.
         [ saved reg 13 ]
         [ saved reg 14 ]
         [ saved reg 15 ]
         [ saved reg 16 ]
         [ frame slot 1 ]
         [ frame slot 2 ]   Local variables.
         [ frame slot 3 ]
               ...
         [outgoing arg 3]
         [outgoing arg 2]   Reserved space for outgoing arguments.
   sp -> [outgoing arg 1]

   Stack layout, after vararg function epilogue, but before return:

                ...
         [     arg 6    ]
         [     arg 5    ]
   sp -> [return address]   Return address has been moved back again.
*/

/* Called by the FUNCTION_PROLOGUE macro.  */
void
pdp10_output_func_prologue (stream, frame_size)
     FILE *stream ATTRIBUTE_UNUSED;
     int frame_size ATTRIBUTE_UNUSED;
{
#if 0
  int clobbered[FIRST_PSEUDO_REGISTER];
  int i;

  fprintf (stream, "\t%s current_function_args_size = %d\n",
	   ASM_COMMENT_START, current_function_args_size);
  fprintf (stream, "\t%s current_function_pretend_args_size = %d\n",
	   ASM_COMMENT_START, current_function_pretend_args_size);
  fprintf (stream, "\t%s frame_size = %d\n",
	   ASM_COMMENT_START, frame_size);
  fprintf (stream, "\t%s current_function_outgoing_args_size = %d\n",
	   ASM_COMMENT_START, current_function_outgoing_args_size);
  fprintf (stream, "\t%s frame_pointer_needed = %d\n",
	   ASM_COMMENT_START, frame_pointer_needed);

  clobbered_regs (clobbered);
  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
    {
      if (clobbered[i])
	{
	  fputc ('\t', stream);
	  fputs (ASM_COMMENT_START, stream);
	  fputs (" callee-saved accumulator ");
	  print_reg (stream, i);
	  fputs (" clobbered\n", stream);
	}
    }
#endif
}

/* Called by the FUNCTION_EPILOGUE macro.  */
void
pdp10_output_func_epilogue (f, frame_size)
     FILE *f;
     int frame_size ATTRIBUTE_UNUSED;
{
  fputc ('\n', f);
}

/* Called by the FUNCTION_PROFILER macro.  */
void
pdp10_function_profiler (file, labelno)
     FILE *file;
     int labelno;
{
}


/**********************************************************************

	Addressing Modes

**********************************************************************/

static int	index_register PARAMS ((rtx x, int strict));

static int
index_register (x, strict)
     rtx x;
     int strict;
{
    return
      GET_CODE (x) == REG
      && (REGNO_REG_CLASS (REGNO (x)) == INDEX_REGS
	  || (!strict && REGNO (x) >= FIRST_PSEUDO_REGISTER));
}

/* Called by the GO_IF_LEGITIMATE_ADDRESS macro.  */
int
pdp10_legitimate_address_p (mode, x, strict)
     enum machine_mode mode;
     rtx x;
     int strict;
{
  if (CONSTANT_ADDRESS_P (x))
    return 1;

  switch (GET_CODE (x))
    {
    case REG:
      return 1;

    case PLUS:
      if (!index_register (XEXP (x, 0), strict))
	return 0;
      x = XEXP (x, 1);
      switch (GET_CODE (x))
	{
	case CONST_INT:
	  if (TARGET_EXTENDED)
	    return INTVAL (x) >= -0x80000 && INTVAL (x) < 0x80000;
	  else
	    return INTVAL (x) > -0x100000 && INTVAL (x) < 0x100000;
	case LABEL_REF:
	case SYMBOL_REF:
	  return !TARGET_EXTENDED;
	default:
	  return 0;
	}

    case PRE_INC:
      x = XEXP (x, 0);
      return
	BYTE_MODE_P (mode)
	&& (GET_CODE (x) == REG
	    || (GET_CODE (x) == MEM
		&& pdp10_legitimate_address_p (Pmode, XEXP (x, 0), strict)));

    case MEM:
      x = XEXP (x, 0);
      if (GET_CODE (x) == REG)
	return index_register (x, strict);
      else if (GET_CODE (x) == PLUS)
	return
	  index_register (XEXP (x, 0), strict)
	  && ((!TARGET_EXTENDED && CONSTANT_ADDRESS_P (XEXP (x, 1)))
	      || (GET_CODE (XEXP (x, 1)) == CONST_INT
		  && pdp10_const_ok_for_letter_p (INTVAL (XEXP (x, 1)), 'K')));

    default:
      return 0;
    }
}

#define gen_ADDRESS(X) \
  gen_rtx_UNSPEC (Pmode, gen_rtvec (1, (X)), UNSPEC_ADDRESS)

/* Called by the LEGITIMIZE_ADDRESS macro.  */
rtx
pdp10_legitimize_address (x, oldx, mode)
     rtx x;
     rtx oldx ATTRIBUTE_UNUSED;
     enum machine_mode mode;
{
  HOST_WIDE_INT max_offset = TARGET_EXTENDED ? 0x7ffff : 0xfffff;
  HOST_WIDE_INT min_offset = TARGET_EXTENDED ? -0x80000 : -0xfffff;
  rtx x0, x1;

  if (mode == BLKmode)
    abort ();

  if (GET_CODE (x) != PLUS)
    return NULL_RTX;

  x0 = XEXP (x, 0);
  x1 = XEXP (x, 1);

  /* Let's see if GCC ever passes anything but a REG in the first
     operand of the PLUS.  If so, write code to swap x0 and x1.  */
  if (GET_CODE (x0) != REG)
    abort ();

  if (GET_CODE (x0) == REG
      && GET_MODE_SIZE (mode) >= UNITS_PER_WORD
      && GET_CODE (x1) == CONST_INT
      && (INTVAL (x1) < min_offset || INTVAL (x1) > max_offset))
    {
      HOST_WIDE_INT offset = INTVAL (x1);
      HOST_WIDE_INT mask = 0xfffff;
      HOST_WIDE_INT sign = 0x80000;
      rtx high_reg = gen_reg_rtx (Pmode);
      rtx base_reg = gen_reg_rtx (Pmode);
      rtx high_part;

      if (!TARGET_EXTENDED)
	error ("address too large for unextended mode");
      else if (offset <= -(HOST_WIDE_INT)0x100000000LL
	       || offset >= (HOST_WIDE_INT)0x100000000LL)
	error ("address too large for extended mode");

      high_part = GEN_INT (((offset + sign) & ~mask) / UNITS_PER_WORD);
#if 1
      emit_insn	(gen_rtx_SET (VOIDmode, high_reg, gen_ADDRESS (high_part)));
#else
      emit_move_insn (high_reg, high_part);
#endif
      emit_insn (gen_addwp3 (base_reg, x0, high_reg));
      return plus_constant (base_reg, ((offset & mask) ^ sign) - sign);
    }

  return NULL_RTX;
}


/**********************************************************************

	Dividing the Output into Sections

**********************************************************************/

static struct hash_entry *
		symbol_type_newfunc PARAMS ((struct hash_entry *he,
					     struct hash_table *table,
					     hash_table_key string));
static tree	lookup_symbol_type PARAMS ((const char *name));
static void	record_symbol_type PARAMS ((const char *name, tree type));

struct symbol_type_entry {
  /* The key will be the symbol name.  */
  struct hash_entry he;
  /* The symbol type.  */
  tree type;
};

static struct hash_table symbol_type_table;

static struct hash_entry *
symbol_type_newfunc (he, table, string)
     struct hash_entry *he;
     struct hash_table *table;
     hash_table_key string ATTRIBUTE_UNUSED;
{
  struct symbol_type_entry *entry = (struct symbol_type_entry *) he;

  if (entry == NULL)
    {
      entry = ((struct symbol_type_entry *)
	       hash_allocate (table, sizeof (struct symbol_type_entry)));
      if (entry == NULL)
	return NULL;
    }

  entry->type = NULL_TREE;
  return (struct hash_entry *) entry;
}

static tree
lookup_symbol_type (name)
     const char *name;
{
  struct symbol_type_entry *entry;

  entry = ((struct symbol_type_entry *)
	   hash_lookup (&symbol_type_table, (const hash_table_key) name, 
			false, string_copy));
  return entry ? entry->type : NULL_TREE;
}

static int
type_bytesize (tree type)
{
  if (type == NULL_TREE)
    return -1;

  switch (TREE_CODE (type))
    {
    case ARRAY_TYPE:
      type = TREE_TYPE (type);
      /* Fall through.  */
    case INTEGER_TYPE:
      return TREE_INT_CST_LOW (TYPE_SIZE (type));
    default:
      return -1;
    }
}

static void
record_symbol_type (name, type)
     const char *name;
     tree type;
{
  struct symbol_type_entry *entry;

  entry = ((struct symbol_type_entry *)
	   hash_lookup (&symbol_type_table, (const hash_table_key) name, 
			true, string_copy));
  if (!entry)
    abort ();

  entry->type = type;
}

/* Called by the ENCODE_SECTION_INFO macro.  */
void
pdp10_encode_section_info (decl)
     tree decl;
{
  static int symbol_type_table_inited = 0;
  const char *name;

  if (!symbol_type_table_inited)
    {
      hash_table_init (&symbol_type_table,
		       symbol_type_newfunc,
		       string_hash,
		       string_compare);
      symbol_type_table_inited = 1;
    }

  switch (TREE_CODE (decl))
    {
      /* Record whether the function is known to be defined in this
         unit or not.  */
    case FUNCTION_DECL:
      if (!TREE_PUBLIC (decl) && !DECL_WEAK (decl))
	SYMBOL_REF_INTERNAL (XEXP (DECL_RTL (decl), 0)) = 1;
      /* Fall through.  */

    case VAR_DECL:
      name = XSTR (XEXP (DECL_RTL (decl), 0), 0);
      record_symbol_type (name, TREE_TYPE (decl));
      break;

    case STRING_CST:
      name = XSTR (XEXP (decl->string.rtl, 0), 0);
      record_symbol_type (name, TREE_TYPE (decl));
      break;

    case ARRAY_TYPE:
    case RECORD_TYPE:
    case UNION_TYPE:
      abort ();

    default:
      break;
    }
}


/**********************************************************************

	Describing Relative Costs of Operations

**********************************************************************/

/* Called by the CONST_COSTS macro.  */
int
pdp10_const_int_costs (value, outer_code)
     HOST_WIDE_INT value;
     enum rtx_code outer_code;
{
  char c = 'I';

  switch (outer_code)
    {
    case ASHIFT:
    case ASHIFTRT:
    case LSHIFTRT:
    case ROTATE:
    case ROTATERT:
      /* A signed 18-bit value.  */
      c = 'K';
      break;

      /* TLZ and TRZ take 18-bit one's complemented values.  AND takes
	 an 18-bit value.  */
    case AND:
      if (pdp10_const_ok_for_letter_p (value, 'N')	/* TLZ */
	  || pdp10_const_ok_for_letter_p (value, 'P'))	/* TRZ */
	return 0;
	/* AND takes an unsigned 18-bit value.  */
      break;

    case IOR:
    case XOR:
      if (pdp10_const_ok_for_letter_p (value, 'L'))	/* TLO and TLC */
	return 0;
      /* IOR and XOR take an unsigned 18-bit value.  */
      break;

    case SET:
      if (pdp10_const_ok_for_letter_p (value, 'L'))	/* MOVS */
	return 0;
      /* Fall through.  */

    case PLUS:
      if (pdp10_const_ok_for_letter_p (value, 'J'))	/* SUBI and MOVN */
	return 0;
      /* ADDI and MOVEI take an unsigned 18-bit value.  */
      break;

    case NEG:
      /* Changing this may affect right-shifts.  */
      return 0;

    default:
      break;
    }

  return pdp10_const_ok_for_letter_p (value, c) ? 0 : COSTS_N_INSNS (1) / 4;
}

int
pdp10_const_double_costs (x, code, outer_code)
     rtx x;
     enum rtx_code code;
     enum rtx_code outer_code;
{
  return COSTS_N_INSNS (1);
}

int
pdp10_operand_cost (x, outer_code)
     rtx x;
     enum rtx_code outer_code;
{
  switch (GET_CODE (x))
    {
    CONST_COSTS (x, GET_CODE (x), outer_code)

    case REG:
    case PLUS:
      return COSTS_N_INSNS (1);

    case MEM:
      return
      (GET_MODE (x) == SImode ?
       COSTS_N_INSNS (1) / 4 : COSTS_N_INSNS (1) / 2)
      + ADDRESS_COST (XEXP (x, 0));

    case NOT:
    case NEG:
    case SUBREG:
      return pdp10_operand_cost (XEXP (x, 0), outer_code);

      /* Not a valid operand.  */
    case MINUS:
      return COSTS_N_INSNS (1000);

    default:
      abort ();
    }
}

int
pdp10_address_cost (x)
     rtx x;
{
  /* Assumptions:
     Using a bare index register or address has no cost.
     Adding an index register to an address costs 1 unit.
     Incrementing a byte pointer costs 2 units.
     Fetching a word from memory costs 10 units.  */

  if (CONSTANT_ADDRESS_P (x))
    return
      TARGET_EXTENDED
      && GET_CODE (x) != LABEL_REF
      && !(GET_CODE (x) == SYMBOL_REF && SYMBOL_REF_INTERNAL (x))
      ? COSTS_N_INSNS (1) / 2 : COSTS_N_INSNS (1) / 4;
  else if (GET_CODE (x) == REG)
    return COSTS_N_INSNS (1) / 2;
  else if (GET_CODE (x) == PLUS)
    return COSTS_N_INSNS (1) / 2;
  else if (GET_CODE (x) == PRE_INC)
    return GET_CODE (XEXP (x, 0)) == REG ?
      COSTS_N_INSNS (1) : COSTS_N_INSNS (2);
  else if (GET_CODE (x) == MEM)
    {
      x = XEXP (x, 0);
      if (GET_CODE (x) == REG)
	return COSTS_N_INSNS (2);
      else if (GET_CODE (x) == PLUS)
	return COSTS_N_INSNS (2);
    }

  /* Not a valid address.  */
  return -1;
}


/**********************************************************************

	Defining the Output Assembler Language

**********************************************************************/

static const char *
		condition_string PARAMS ((enum rtx_code condition));
static void	print_reg PARAMS ((FILE *stream, int regno));
static void	print_byte_symbol PARAMS ((FILE *stream, const char *name,
					   int bytesize, int aggregate));
static void	print_byte_pointer PARAMS ((FILE *stream, rtx x, int));
static void	print_word_pointer PARAMS ((FILE *stream, rtx x, int));

/* In case of a PRE_INC, POST_INC, PRE_DEC, POST_DEC memory reference, we
   must report the mode of the memory reference from PRINT_OPERAND to
   PRINT_OPERAND_ADDRESS.  */
static enum machine_mode output_memory_reference_mode;

/* Called by the ASM_DECLARE_OBJECT_NAME macro.  */
void
pdp10_asm_declare_object_name (stream, name, decl)
     FILE *stream;
     const char *name;
     tree decl;
{
  int byte_size;

  ASM_OUTPUT_LABEL (stream, name);

  /* If this is a scalar integer variable smaller than a word, justify
     it to the right inside a word.  */
  if (TREE_CODE (decl) == VAR_DECL
      && TREE_CODE (TREE_TYPE (decl)) == INTEGER_TYPE
      && (byte_size = TREE_INT_CST_LOW (TYPE_SIZE (TREE_TYPE (decl)))) < 36)
    {
      pdp10_output_byte (stream, const0_rtx, 36 - byte_size);
    }
}

/* Called by the ASM_OUTPUT_LABELREF macro.  */
void
pdp10_asm_output_labelref (file, name)
     FILE *file;
     const char *name;
{
  int len;
  char *label;
  int i;

#ifndef __PDP10_GAS_H__
  pdp10_output_byte (file, 0, 0);
#endif

  if (*name == '*')
    {
      fputs (name + 1, file);
      return;
    }

#ifdef __PDP10_GAS_H__
  label = name;
#else
  len = strlen (name);
  label = alloca (len + 1);
  strcpy (label, name);
  for (i = 0; i < len; i++)
    {
      if (label[i] == '_')
	label[i] = '.';
    }
#endif

  asm_fprintf (file, "%U%s", label);
}

/* Called by the ASM_OUTPUT_SYMBOL_REF macro.  */
void
pdp10_asm_output_symbol_ref (stream, sym)
     FILE *stream;
     rtx sym;
{
  const char *name = XSTR (sym, 0);
  int len;
  char *label;
  int i;
  tree type;
  int bytesize;

  type = lookup_symbol_type (name);
  if (type && TREE_CODE (type) == RECORD_TYPE)
    bytesize
      = output_memory_reference_mode == BLKmode
      ? BITS_PER_WORD
      : GET_MODE_SIZE (output_memory_reference_mode) * BITS_PER_UNIT;
  else
    bytesize = type ? type_bytesize (type) : -1;
  if (bytesize != -1 && bytesize < 36
      && output_memory_reference_mode != word_mode)
    {
      print_byte_symbol (stream, name, bytesize, AGGREGATE_TYPE_P (type));
      return;
    }

  if (*name == '*')
    {
      fputs (name + 1, stream);
      return;
    }

#ifdef __PDP10_GAS_H__
  label = name;
#else
  len = strlen (name);
  label = alloca (len + 1);
  strcpy (label, name);
  for (i = 0; i < len; i++)
    {
      if (label[i] == '_')
	label[i] = '.';
    }
#endif

  asm_fprintf (stream, "%U%s", label);
}

/* Called by the ASM_OUTPUT_OPCODE macro.  */
const char *
pdp10_asm_output_opcode (FILE *stream, const char *ptr)
{
  while (*ptr != ' ' && *ptr != 0 && *ptr != '%')
    {
      int ch = *ptr++;

      if (target_flags & ASM_UPPER)
	ch = TOUPPER (ch);

      fputc (ch, stream);
    }

  if (*ptr == ' ')
    {
      if (target_flags & ASM_TAB)
	fputc ('\t', stream);
      else
	fputc (' ', stream);
      ptr++;
    }

  return ptr;
}

static const char *
condition_string (condition)
     enum rtx_code condition;
{
  static const char *upper[] = { "", "L", "E", "LE", "A", "GE", "N", "G" };
  static const char *lower[] = { "", "l", "e", "le", "a", "ge", "n", "g" };
  const char **string = target_flags & ASM_UPPER ? upper : lower;

  switch (condition)
    {
   /* never: return string[0]; */
    case LT: return string[1];
    case EQ: return string[2];
    case LE: return string[3];
  /* always: return string[4]; */
    case GE: return string[5];
    case NE: return string[6];
    case GT: return string[7];
    default: abort ();
    }
}

void
pdp10_print_operand (stream, x, code)
     FILE *stream;
     rtx x;
     int code;
{
  switch (code)
    {
      /* Start a comment.  */
    case ';':
      fputs (ASM_COMMENT_START, stream);
      return;

      /* Compute one's complement of an integer constant.  */
    case 'C':
      if (GET_CODE (x) != CONST_INT)
	abort ();
      pdp10_print_operand (stream, GEN_INT (~INTVAL (x)), 0);
      return;

      /* Print double.  */
    case 'D':
      if (GET_CODE (x) != CONST_INT)
	abort ();

      fputs ("0\n\t\t", stream);
      pdp10_print_integer (stream, x);
      return;

      /* Print floating-point immediate (constraint 'G').  */
    case 'G':
      if (GET_MODE (x) != SFmode && GET_MODE (x) != DFmode)
	abort ();

      if (1)
	{
	  REAL_VALUE_TYPE r;
	  char str[30];

	  REAL_VALUE_FROM_CONST_DOUBLE (r, x);
	  REAL_VALUE_TO_DECIMAL (r, "%.3g", str);
	  fputc ('(', stream);
	  fputs (str, stream);
	  fputc (')', stream);
	}
      else
	{
	  REAL_VALUE_TYPE r;
	  HOST_WIDE_INT z;
	  long y[3];

	  REAL_VALUE_FROM_CONST_DOUBLE (r, x);
	  REAL_VALUE_TO_TARGET_DOUBLE (r, y);
	  z = (((HOST_WIDE_INT) y[0]) >> 14) & 0777777;
	  pdp10_print_unsigned (stream, GEN_INT (z));
	}

      return;

      /* Print local symbolic address constant, if possible and appropriate. */
    case 'L':
      if (GET_CODE (x) != SYMBOL_REF
	  && GET_CODE (x) != LABEL_REF)
	break;

      if (TARGET_EXTENDED
	  && GET_CODE (x) == SYMBOL_REF
	  && !SYMBOL_REF_INTERNAL (x))
	{
	  fputs ("@[GIW ", stream);
	  output_addr_const (stream, x);
	  fputc (']', stream);
	}
      else
	output_addr_const (stream, x);

      return;

      /* Print negated integer constant.  */
    case 'N':
      pdp10_print_integer (stream, GEN_INT (-INTVAL (x)));
      return;

      /* Print stack register in a pdp10_push_operand or
         pdp10_pop_operand.  */
    case 'P':
      if (GET_CODE (x) != MEM
	  || (GET_MODE (x) != SImode && GET_MODE (x) != Pmode))
	abort ();
      x = XEXP (x, 0);
      if (GET_CODE (x) != PRE_INC && GET_CODE (x) != POST_DEC)
	abort ();
      x = XEXP (x, 0);
      if (GET_CODE (x) != REG)
	abort ();
      print_reg (stream, REGNO (x));
      return;

      /* Print reversed condition.  */
    case 'R':
      fputs (condition_string (reverse_condition (GET_CODE (x))), stream);
      return;

      /* Swap half-words of an integer constant.  */
    case 'S':
      if (GET_CODE (x) != CONST_INT)
	abort ();
      {
	HOST_WIDE_INT i =  INTVAL (x)        & 0777777;
	HOST_WIDE_INT j = (INTVAL (x) >> 18) & 0777777;
	pdp10_print_operand (stream, GEN_INT ((i << 18) | j), 0);
      }
      return;

      /* Compute one's complement and swap half-words of an integer
         constant.  */
    case 'T':
      if (GET_CODE (x) != CONST_INT)
	abort ();
      {
	HOST_WIDE_INT i =  (~INTVAL (x))        & 0777777;
	HOST_WIDE_INT j = ((~INTVAL (x)) >> 18) & 0777777;
	pdp10_print_operand (stream, GEN_INT ((i << 18) | j), 0);
      }
      return;

      /* Force word address.  */
    case 'W':
      if (GET_CODE (x) != MEM)
	break;
      pdp10_print_operand_address (stream, XEXP (x, 0), word_mode);
      return;

      /* Print next register.  */
    case 'Z':
      if (GET_CODE (x) != REG)
	abort ();
      print_reg (stream, (REGNO (x) + 1) % 020);
      return;
    }

  switch (GET_CODE (x))
    {
    case REG:
      print_reg (stream, REGNO (x));
      break;

    case MEM:
      output_memory_reference_mode = GET_MODE (x);
      pdp10_print_operand_address (stream, XEXP (x, 0), GET_MODE (x));
      break;

    case CONST_DOUBLE:
      if (GET_MODE (x) != SFmode && GET_MODE (x) != DFmode)
	abort ();

      if (1)
	{
	  REAL_VALUE_TYPE r;
	  char str[30];

	  REAL_VALUE_FROM_CONST_DOUBLE (r, x);
	  REAL_VALUE_TO_DECIMAL (r,
				 GET_MODE (x) == SFmode ? "%.9g" : "%.19g",
				 str);
	  fputs (str, stream);
	}
      else
	{
	  REAL_VALUE_TYPE r;
	  HOST_WIDE_INT z;
	  long y[3];

	  REAL_VALUE_FROM_CONST_DOUBLE (r, x);
	  REAL_VALUE_TO_TARGET_DOUBLE (r, y);

	  z = ((HOST_WIDE_INT)y[0] << 4)
	    + (HOST_WIDE_INT)((y[1] >> 28) & 0xful);
	  pdp10_print_unsigned (stream, GEN_INT (z));

	  if (GET_MODE (x) == DFmode)
	    {
	      z = (((HOST_WIDE_INT)y[1] << 8) & 0xffffffful)
		+ (HOST_WIDE_INT)((y[2] >> 24) & 0xfful);
	      fputc (',', stream);
	      pdp10_print_unsigned (stream, GEN_INT (z));
	    }
	}

      break;

    case CONST:
      x = XEXP (x, 0);
      if (GET_CODE (x) == PLUS
	  && GET_CODE (XEXP (x, 0)) == SYMBOL_REF
	  && GET_CODE (XEXP (x, 1)) == CONST_INT)
	{
	  pdp10_print_operand (stream, XEXP (x, 0), 0);
	  x = XEXP (x, 1);
	  if (INTVAL (x) >= UNITS_PER_WORD)
	    fputc ('+', stream);
	  pdp10_print_offset30 (stream, x);
	}
      else
	abort ();
      break;

    case CONST_INT:
      pdp10_print_integer (stream, x);
      break;

    case SYMBOL_REF:
    case LABEL_REF:
      if (type_bytesize (lookup_symbol_type (XSTR (x, 0))) != -1
	  && type_bytesize (lookup_symbol_type (XSTR (x, 0))) < 36)
	print_byte_pointer (stream, x, 0);
      else
	print_word_pointer (stream, x, 0);
      break;

    case LT: case EQ: case LE: case GE: case NE: case GT:
      fputs (condition_string (GET_CODE (x)), stream);
      return;

    default:
      abort ();
    }
}

static void
print_reg (stream, regno)
     FILE *stream;
     int regno;
{
  fputs (REGISTER_PREFIX, stream);
  fputs (reg_names[regno], stream);
}

static void
print_byte_pointer (stream, x, indirect)
     FILE *stream;
     rtx x;
     int indirect;
{
  if (indirect)
    fputc ('[', stream);

  if (GET_CODE (x) == CONST_INT)
    abort (); /*pdp10_print_address (stream, x);*/
  else if (GET_CODE (x) == CONST)
    {
      x = XEXP (x, 0);
      output_addr_const (stream, XEXP (x, 0));
      x = XEXP (x, 1);
      if (INTVAL (x) >= UNITS_PER_WORD)
	fputc ('+', stream);
      pdp10_print_offset30 (stream, x);
    }
  else
    output_addr_const (stream, x);

  if (indirect)
    fputc (']', stream);
}

static void
print_byte_symbol (stream, name, byte_size, aggregate)
     FILE *stream;
     const char *name;
     int byte_size;
     int aggregate;
{
  static const char *byte6_selector[] = { "46", "47", "50", "51", "52", "53" };
  static const char *byte7_selector[] = { "62", "63", "64", "65", "66" };
  static const char *byte8_selector[] = { "55", "56", "57", "60" };
  static const char * char_selector[] = { "70", "71", "72", "73" };
  static const char *short_selector[] = { "75", "76" };
  static const char *int32_selector[] = { "??" };
  const char **byte_selector;
  int bytes_per_word;
  int which_byte;

  switch (byte_size)
    {
    case  6: byte_selector = byte6_selector; bytes_per_word = 6; break;
    case  7: byte_selector = byte7_selector; bytes_per_word = 5; break;
    case  8: byte_selector = byte8_selector; bytes_per_word = 4; break;
    case  9: byte_selector =  char_selector; bytes_per_word = 4; break;
    case 16: byte_selector = short_selector; bytes_per_word = 2; break;
    case 18: byte_selector = short_selector; bytes_per_word = 2; break;
    case 32: byte_selector = int32_selector; bytes_per_word = 1; break;
    default: abort ();
    }

  if (aggregate)
    which_byte = 0;
  else
    which_byte = bytes_per_word - 1;

  if (TARGET_EXTENDED)
    fprintf (stream, "OWGBP %s,", byte_selector[which_byte]);
  else
    fprintf (stream, "POINT %d,", byte_size);

  assemble_name (stream, name);

  if (!TARGET_EXTENDED)
    fprintf (stream, ",%d", byte_size * which_byte + byte_size - 1);
}

static void
print_word_pointer (stream, x, indirect)
     FILE *stream;
     rtx x;
     int indirect;
{
  if (TARGET_EXTENDED)
    {
      if (indirect)
	fputs ("@[", stream);
      fputs ("GIW ", stream);
    }

  switch (GET_CODE (x))
    {
    case CONST:
      x = XEXP (x, 0);
      if (GET_CODE (x) == PLUS
	  && GET_CODE (XEXP (x, 0)) == SYMBOL_REF
	  && GET_CODE (XEXP (x, 1)) == CONST_INT)
	{
	  output_addr_const (stream, XEXP (x, 0));
	  x = XEXP (x, 1);
	  if (INTVAL (x) >= UNITS_PER_WORD)
	    fputc ('+', stream);
	  pdp10_print_offset30 (stream, x);
	}
      else
	abort ();
      break;

    case CONST_INT:
      pdp10_print_address (stream, x);
      break;

    case SYMBOL_REF:
    case LABEL_REF:
      output_addr_const (stream, x);
      break;

    default:
      abort ();
    }

  if (TARGET_EXTENDED && indirect)
    fputc (']', stream);
}

void
pdp10_print_operand_address (stream, x, mode)
     FILE *stream;
     rtx x;
     enum machine_mode mode;
{
  int byte_pointer = BYTE_MODE_P (mode);

  switch (GET_CODE (x))
    {
    case CONST_INT:
#if 0
      if (byte_pointer)
	print_byte_pointer (stream, x, 1);
      else
#endif
	print_word_pointer (stream, x, 1);
      break;

    case CONST:
    case SYMBOL_REF:
      {
	rtx sym = x;
	tree type;
	if (GET_CODE (x) == CONST)
	  sym = XEXP (XEXP (x, 0), 0);
	type = lookup_symbol_type (XSTR (sym, 0));
	if ((TREE_CODE (type) == RECORD_TYPE && byte_pointer)
	    || (type_bytesize (type) != -1 && type_bytesize (type) < 36))
	  print_byte_pointer (stream, x, 1);
	else
	  print_word_pointer (stream, x, 1);
	break;
      }

    case REG:
      if (!byte_pointer)
	{
	  if (REGNO (x) == 0)
	    fputc ('@', stream);
	  else
	    fputc ('(', stream);
	}
      print_reg (stream, REGNO (x));
      if (!byte_pointer && REGNO (x) != 0)
	fputc (')', stream);
      break;

    case PLUS:
      if (GET_CODE (XEXP (x, 0)) != REG
	  || (GET_CODE (XEXP (x, 1)) != CONST_INT
	      && GET_CODE (XEXP (x, 1)) != LABEL_REF
	      && GET_CODE (XEXP (x, 1)) != SYMBOL_REF)
	  || REGNO (XEXP (x, 0)) == 0)
	abort ();

      if (GET_CODE (XEXP (x, 1)) == CONST_INT)
	pdp10_print_offset (stream, XEXP (x, 1));
      else
	output_addr_const (stream, XEXP (x, 1));

      fputc ('(', stream);
      print_reg (stream, REGNO (XEXP (x, 0)));
      fputc (')', stream);
      break;

    case MEM:
      if (!byte_pointer)
	fputc ('@', stream);
      pdp10_print_operand_address (stream, XEXP (x, 0), mode);
      break;

    case PRE_INC:
      /* PRE_INC is used with the ILDB and IDPB instructions.  */
      if (!byte_pointer)
	abort ();
      pdp10_print_operand_address (stream, XEXP (x, 0), mode);
      break;

    case UNSPEC:
      if (XINT (x, 1) != UNSPEC_ADDRESS)
	abort ();
      pdp10_print_operand_address (stream, XVECEXP (x, 0, 0), mode);
      break;

    default:
      abort ();
    }
}


/**********************************************************************

	Special Predicates

**********************************************************************/

static int	pointer_alignment PARAMS ((rtx x));

static int
pointer_alignment (x)
     rtx x;
{
  /* This function must be prepared to handle anything
     GO_IF_LEGITIMATE_ADDRESS says is a valid address.  */

  switch (GET_CODE (x))
    {
    case CONST:
    case SYMBOL_REF:
    case LABEL_REF:
      return UNITS_PER_WORD;
    case CONST_INT:
      return INTVAL (x) & (~INTVAL (x) + 1);
    case REG:
      return REGNO_POINTER_ALIGN (REGNO (x)) / BITS_PER_UNIT;
    case PLUS:
      x = XEXP (x, 0);
      if (GET_CODE (XEXP (x, 1)) == CONST_INT)
	return REGNO_POINTER_ALIGN (REGNO (XEXP (x, 0))) / BITS_PER_UNIT;
      else
	return UNITS_PER_WORD;
    case PRE_INC:
      x = XEXP (x, 0);
      if (GET_CODE (x) == REG)
	return REGNO_POINTER_ALIGN (REGNO (x)) / BITS_PER_UNIT;
      return 1;
    case MEM:
      return 1;
    default:
      abort ();
    }
}

int
pdp10_comparison_operator (x, mode)
     rtx x;
     enum machine_mode mode ATTRIBUTE_UNUSED;
{
  switch (GET_CODE (x))
    {
    case EQ:
    case NE:
    case LT:
    case LE:
    case GT:
    case GE:
      return 1;
    default:
      return 0;
    }
}

int
pdp10_equality_operator (x, mode)
     rtx x;
     enum machine_mode mode ATTRIBUTE_UNUSED;
{
  enum rtx_code code = GET_CODE (x);
  return code == EQ || code == NE;
}

int
reg_or_mem_operand (x, mode)
     rtx x;
     enum machine_mode mode ATTRIBUTE_UNUSED;
{
  return register_operand (x, mode) || memory_operand (x, mode);
}

int
pdp10_aligned_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) != MEM)
    return 0;

  if (mode != VOIDmode && GET_MODE (op) != mode)
    return 0;

  return pointer_alignment (XEXP (op, 0)) % UNITS_PER_WORD == 0;
}

int
pdp10_push_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) != MEM)
    return 0;

  if (mode != VOIDmode && GET_MODE (op) != mode)
    return 0;

  op = XEXP (op, 0);

  if (GET_CODE (op) != PRE_INC)
    return 0;

  return GET_CODE (XEXP (op, 0)) == REG;
}

int
pdp10_pop_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) != MEM)
    return 0;

  if (mode != VOIDmode && GET_MODE (op) != mode)
    return 0;

  op = XEXP (op, 0);

  if (GET_CODE (op) != POST_DEC)
    return 0;

  return GET_CODE (XEXP (op, 0)) == REG;
}

int
pdp10_inline_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  HOST_WIDE_INT x;

  if (mode != VOIDmode && GET_MODE (op) != mode && GET_CODE (op) != CONST_INT)
    return 0;

  x  = INTVAL (op);
  return
    !pdp10_const_ok_for_letter_p (x, 'I')
    && !pdp10_const_ok_for_letter_p (x, 'S')
    && !pdp10_const_ok_for_letter_p (x, 'L')
    && !pdp10_const_ok_for_letter_p (x, 'N')
    && !pdp10_const_ok_for_letter_p (x, 'P');
}

int
pdp10_const_double_0_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  REAL_VALUE_TYPE r;
  long y[3];

  if (mode != VOIDmode && GET_MODE (op) != mode)
    return 0;

  REAL_VALUE_FROM_CONST_DOUBLE (r, op);
  REAL_VALUE_TO_TARGET_DOUBLE (r, y);

  switch (mode)
    {
    case SFmode:
      return
	(y[0] & 0xffffffffUL) == 0
	&& (y[1] & 0xf0000000UL) == 0;
    case DFmode:
      return
	(y[0] & 0xffffffffUL) == 0
	&& (y[1] & 0xffffffffUL) == 0
	&& (y[2] & 0xff000000UL) == 0;
    default:
      abort ();
    }
}


/**********************************************************************

	Built-in Functions

**********************************************************************/

static rtx	expand_builtin_jffo PARAMS ((tree arglist, rtx target));
static rtx	expand_builtin_fsc PARAMS ((tree arglist, rtx target,
					    enum rtx_code code));

void
pdp10_init_builtins ()
{
#define PTRTYPE(a)	built_pointer_type (a)
#define NIL		void_list_node
#define CONS(a,b)	tree_cons (NULL_TREE, a, b)
#define FUNTYPE(a,b)	build_function_type (a, b)
#define FUNTYPE0(a)	FUNTYPE (a, NIL)
#define FUNTYPE1(a,b)	FUNTYPE (a, CONS (b, NIL))
#define FUNTYPE2(a,b,c)	FUNTYPE (a, CONS (b, CONS (c, NIL)))
#define VOID		void_type_node
#define VOIDP		ptr_type_node
#define INT		integer_type_node
#define CHAR		char_type_node
#define LONGLONG	long_long_integer_type_node
#define FLOAT		float_type_node
#define DOUBLE		double_type_node

  builtin_function ("__builtin_fsc",
		    FUNTYPE2 (FLOAT, FLOAT, INT),
		    PDP10_BUILTIN_FSC,
		    BUILT_IN_MD,
		    NULL_PTR);

  if (TARGET_KA10up)
    {
      builtin_function ("__builtin_jffo",
			FUNTYPE2 (VOID, LONGLONG, VOIDP),
			PDP10_BUILTIN_JFFO,
			BUILT_IN_MD,
			NULL_PTR);
    }

  if (TARGET_KL10up && TARGET_GFLOAT)
    {
      builtin_function ("__builtin_gfsc",
			FUNTYPE2 (DOUBLE, DOUBLE, INT),
			PDP10_BUILTIN_GFSC,
			BUILT_IN_MD,
			NULL_PTR);

    }

#undef PTRTYPE
#undef NIL
#undef CONS
#undef FUNTYPE0
#undef FUNTYPE1
#undef FUNTYPE2
#undef VOID
#undef VOIDP
#undef INT
#undef CHAR
#undef LONGLONG
#undef FLOAT
#undef DOUBLE
}

static rtx
expand_builtin_jffo (arglist, target)
     tree arglist;
     rtx target;
{
  tree arg0, arg1;
  rtx op0, op1, pat;

  arg0 = TREE_VALUE (arglist);
  arg1 = TREE_VALUE (TREE_CHAIN (arglist));
  op0 = expand_expr (arg0, NULL_RTX, VOIDmode, 0);
  op1 = expand_expr (arg1, NULL_RTX, VOIDmode, 0);

  //reg = gen_reg_rtx (DImode);
  //emit_move_insn (gen_rtx_SUBREG (SImode, reg, 0), op0);
  //emit_move_insn (gen_rtx_SUBREG (SImode, reg, 1), constm1_rtx);

  op0 = force_reg (DImode, op0);
  pat = gen_JFFO (op0, op1);
  if (!pat)
    return 0;
  emit_jump_insn (pat);

  return target; //gen_rtx_SUBREG (SImode, reg, 1);
}

static rtx
expand_builtin_fsc (arglist, target, code)
     tree arglist;
     rtx target ATTRIBUTE_UNUSED;
     enum pdp10_builtins code;
{
  tree arg0, arg1;
  rtx op0, op1, pat;

  arg0 = TREE_VALUE (arglist);
  arg1 = TREE_VALUE (TREE_CHAIN (arglist));
  op0 = expand_expr (arg0, NULL_RTX, VOIDmode, 0);
  op1 = expand_expr (arg1, NULL_RTX, VOIDmode, 0);
  op0 = force_reg (DFmode, op0);
  op1 = force_not_mem (op1);

  if (!target)
    target = gen_reg_rtx (code == PDP10_BUILTIN_FSC ? SFmode : DFmode);

  if (code == PDP10_BUILTIN_FSC)
    pat = gen_FSC (target, op0, op1);
  else
    pat = gen_GFSC (target, op0, op1);
  if (!pat)
    return 0;
  emit_insn (pat);

  return target;
}

rtx
pdp10_expand_builtin (exp, target, subtarget, mode, ignore)
     tree exp;
     rtx target;
     rtx subtarget ATTRIBUTE_UNUSED;
     enum machine_mode mode ATTRIBUTE_UNUSED;
     int ignore ATTRIBUTE_UNUSED;
{
  tree arglist = TREE_OPERAND (exp, 1);
  enum pdp10_builtins code =
    DECL_FUNCTION_CODE (TREE_OPERAND (TREE_OPERAND (exp, 0), 0));

  switch (code)
    {
    case PDP10_BUILTIN_JFFO:
      return expand_builtin_jffo (arglist, target);
    case PDP10_BUILTIN_FSC:
    case PDP10_BUILTIN_GFSC:
      return expand_builtin_fsc (arglist, target, code);
    default:
      abort ();
    }
}


/**********************************************************************

	Machine-specific Attributes

**********************************************************************/

static void	make_byte_type PARAMS ((tree type, int precision,
					int unsignedp));
static void	build_machine_types PARAMS ((void));
static void	layout_byte_decl PARAMS ((tree decl));

static tree char6_signed_type_node;
static tree char6_unsigned_type_node;
static tree char7_signed_type_node;
static tree char7_unsigned_type_node;
static tree char8_signed_type_node;
static tree char8_unsigned_type_node;
static tree short16_integer_type_node;
static tree short16_unsigned_type_node;
static tree int32_integer_type_node;
static tree int32_unsigned_type_node;

static void
make_byte_type (type, precision, unsignedp)
     tree type;
     int precision;
     int unsignedp;
{
  TYPE_PRECISION (type) = precision;

  if (unsignedp)
    {
      TYPE_MIN_VALUE (type) = build_int_2 (0, 0);
      TYPE_MAX_VALUE (type)
	= build_int_2 (precision - HOST_BITS_PER_WIDE_INT >= 0
		       ? -1 : ((HOST_WIDE_INT) 1 << precision) - 1,
		       precision - HOST_BITS_PER_WIDE_INT > 0
		       ? ((unsigned HOST_WIDE_INT) ~0
			  >> (HOST_BITS_PER_WIDE_INT
			      - (precision - HOST_BITS_PER_WIDE_INT)))
		       : 0);
    }
  else
    {
      TYPE_MIN_VALUE (type)
	= build_int_2 ((precision - HOST_BITS_PER_WIDE_INT > 0
			? 0 : (HOST_WIDE_INT) (-1) << (precision - 1)),
		       (((HOST_WIDE_INT) (-1)
			 << (precision - HOST_BITS_PER_WIDE_INT - 1 > 0
			     ? precision - HOST_BITS_PER_WIDE_INT - 1
			     : 0))));
      TYPE_MAX_VALUE (type)
	= build_int_2 ((precision - HOST_BITS_PER_WIDE_INT > 0
			? -1 : ((HOST_WIDE_INT) 1 << (precision - 1)) - 1),
		       (precision - HOST_BITS_PER_WIDE_INT - 1 > 0
			? (((HOST_WIDE_INT) 1
			    << (precision - HOST_BITS_PER_WIDE_INT - 1))) - 1
			: 0));
    }

  TREE_TYPE (TYPE_MIN_VALUE (type)) = type;
  TREE_TYPE (TYPE_MAX_VALUE (type)) = type;

  TREE_UNSIGNED (type) = unsignedp;

  TYPE_MODE (type) = smallest_mode_for_size (TYPE_PRECISION (type),
					     MODE_INT);
  TYPE_SIZE (type) = bitsize_int (TYPE_PRECISION (type));
  TYPE_SIZE_UNIT (type)
    = size_int ((TYPE_PRECISION (type) + BITS_PER_UNIT - 1) / BITS_PER_UNIT);

  TYPE_ALIGN (type) = precision;
}

static void
build_machine_types ()
{
  char6_signed_type_node = make_node (INTEGER_TYPE);
  make_byte_type (char6_signed_type_node, 6, 0);
  char6_unsigned_type_node = make_node (INTEGER_TYPE);
  make_byte_type (char6_unsigned_type_node, 6, 1);

  char7_signed_type_node = make_node (INTEGER_TYPE);
  make_byte_type (char7_signed_type_node, 7, 0);
  char7_unsigned_type_node = make_node (INTEGER_TYPE);
  make_byte_type (char7_unsigned_type_node, 7, 1);

  char8_signed_type_node = make_node (INTEGER_TYPE);
  make_byte_type (char8_signed_type_node, 8, 0);
  char8_unsigned_type_node = make_node (INTEGER_TYPE);
  make_byte_type (char8_unsigned_type_node, 8, 1);

  short16_integer_type_node = make_node (INTEGER_TYPE);
  make_byte_type (short16_integer_type_node, 16, 0);
  short16_unsigned_type_node = make_node (INTEGER_TYPE);
  make_byte_type (short16_unsigned_type_node, 16, 1);

  int32_integer_type_node = make_node (INTEGER_TYPE);
  make_byte_type (int32_integer_type_node, 32, 0);
  int32_unsigned_type_node = make_node (INTEGER_TYPE);
  make_byte_type (int32_unsigned_type_node, 32, 1);
}

static void
layout_byte_decl (decl)
     tree decl;
{
  register tree type = TREE_TYPE (decl);
  register enum tree_code code = TREE_CODE (decl);

  TREE_UNSIGNED (decl) = TREE_UNSIGNED (type);
  DECL_MODE (decl) = TYPE_MODE (type);
  DECL_SIZE (decl) = TYPE_SIZE (type);
  DECL_SIZE_UNIT (decl) = TYPE_SIZE_UNIT (type);

  /* Force alignment required for the data type.
     But if the decl itself wants greater alignment, don't override that.
     Likewise, if the decl is packed, don't override it.  */
  if (! (code == FIELD_DECL && DECL_BIT_FIELD (decl))
      && (DECL_ALIGN (decl) == 0
	  || (! (code == FIELD_DECL && DECL_PACKED (decl))
	      && TYPE_ALIGN (type) > DECL_ALIGN (decl))))
    {	      
      DECL_ALIGN (decl) = TYPE_ALIGN (type);
      DECL_USER_ALIGN (decl) = TYPE_USER_ALIGN (type);
    }

  /* For fields, set the bit field type and update the alignment.  */
  if (code == FIELD_DECL)
    {
      DECL_BIT_FIELD_TYPE (decl) = DECL_BIT_FIELD (decl) ? type : 0;
      if (maximum_field_alignment != 0)
	DECL_ALIGN (decl) = MIN (DECL_ALIGN (decl), maximum_field_alignment);
      else if (DECL_PACKED (decl))
	DECL_USER_ALIGN (decl) = 0;
    }

  /* Evaluate nonconstant size only once, either now or as soon as safe.  */
  if (DECL_SIZE (decl) != 0 && TREE_CODE (DECL_SIZE (decl)) != INTEGER_CST)
    DECL_SIZE (decl) = variable_size (DECL_SIZE (decl));
  if (DECL_SIZE_UNIT (decl) != 0
      && TREE_CODE (DECL_SIZE_UNIT (decl)) != INTEGER_CST)
    DECL_SIZE_UNIT (decl) = variable_size (DECL_SIZE_UNIT (decl));

  /* If requested, warn about definitions of large data objects.  */
  if (warn_larger_than
      && (code == VAR_DECL || code == PARM_DECL)
      && ! DECL_EXTERNAL (decl))
    {
      tree size = DECL_SIZE_UNIT (decl);

      if (size != 0 && TREE_CODE (size) == INTEGER_CST
	  && compare_tree_int (size, larger_than_size) > 0)
	{
	  unsigned int size_as_int = TREE_INT_CST_LOW (size);

	  if (compare_tree_int (size, size_as_int) == 0)
	    warning_with_decl (decl, "size of `%s' is %d bytes", size_as_int);
	  else
	    warning_with_decl (decl, "size of `%s' is larger than %d bytes",
			       larger_than_size);
	}
    }
}

#define PDP10_TYPE_FROM_SIZE(SIZE, UNSIGNEDP)				    \
  ((SIZE) == 6 ?							    \
   ((UNSIGNEDP) ? char6_unsigned_type_node : char6_signed_type_node) :	    \
   (SIZE) == 7 ?							    \
   ((UNSIGNEDP) ? char7_unsigned_type_node : char7_signed_type_node) :	    \
   (SIZE) == 8 ?							    \
   ((UNSIGNEDP) ? char8_unsigned_type_node : char8_signed_type_node) :	    \
   (SIZE) == 9 ?							    \
   ((UNSIGNEDP) ? unsigned_char_type_node : signed_char_type_node) :	    \
   (SIZE) == 16 ?							    \
   ((UNSIGNEDP) ? short16_unsigned_type_node : short16_integer_type_node) : \
   (SIZE) == 18 ?							    \
   ((UNSIGNEDP) ? short_unsigned_type_node : short_integer_type_node) :	    \
   (SIZE) == 32 ?							    \
   ((UNSIGNEDP) ? int32_unsigned_type_node : int32_integer_type_node) :	    \
   NULL_TREE)

int
pdp10_valid_machine_attribute (node, attributes, identifier, args)
     tree node;
     tree attributes ATTRIBUTE_UNUSED;
     tree identifier;
     tree args;
{
  const char *name = IDENTIFIER_POINTER (identifier);
  enum tree_code code = TREE_CODE (node);
  static int init = 1;

  if (init)
    {
      build_machine_types ();
      init = 0;
    }

  if (strcmp (name, "size") == 0
      && (code == INTEGER_TYPE
	  || code == VAR_DECL
	  || code == PARM_DECL
	  || (code == FIELD_DECL && !DECL_BIT_FIELD (node))))
    {
      int size = 0;

      if (args && TREE_VALUE (args)
	  && TREE_CODE (TREE_VALUE (args)) == INTEGER_CST)
	size = TREE_INT_CST_LOW (TREE_VALUE (args));

      if (!VALID_BYTE_SIZE_P (size) && size != 16 && size != 32)
	{
	  error ("invalid value for `size' attribute");
	  return 0;
	}

      if ((code != INTEGER_TYPE
	   && TREE_CODE (TREE_TYPE (node)) != INTEGER_TYPE)
	  || TREE_CHAIN (args) != NULL_TREE
	  || size == 0)
	return 0;

      if (code == INTEGER_TYPE)
	{
#if 1
	  make_byte_type (node, size, TREE_UNSIGNED (node));
#else
	  TYPE_PRECISION (node) = size;
	  if (TREE_UNSIGNED (node))
	    fixup_unsigned_type (node);
	  else
	    fixup_signed_type (node);
#endif
	}
      else
	{
	  int unsignedp
	    = TREE_UNSIGNED (code == FIELD_DECL ? node : TREE_TYPE (node));
	  TREE_TYPE (node) = PDP10_TYPE_FROM_SIZE (size, unsignedp);
	  layout_byte_decl (node);
	}

      return 1;
    }

  return 0;
}


/**********************************************************************

	MACRO, MIDAS, and FAIL

**********************************************************************/

#ifndef __PDP10_GAS_H__

static int	asciz_ok PARAMS ((const char *ptr, int len));
static void	macro_output_ascii_bytes PARAMS ((FILE *stream,
						  const char *ptr,
						  int len,
						  int byte_size));

void
macro_file_start (file)
     FILE *file;
{
  const char *title;

  if (main_input_filename)
    {
      title = main_input_filename + strlen (main_input_filename);
      while (title != main_input_filename && !IS_DIR_SEPARATOR (title[-1]))
	title--;
      fprintf (file, "\tTITLE\t%s\n", title);
    }

  fprintf (file, "\tSEARCH\tMONSYM\n\n");

  if (TARGET_KA10up)
    {
      fputs ("\t.DIRECTIVE\t", file);
      fputs (TARGET_KL10up ? "KL10" :
	     TARGET_KI10up ? "KI10,KL10" :
			     "KA10,KI10,KL10",
	     file);
      fputs ("\n\n", file);
    }

  if (TARGET_EXTENDED)
    {
      fprintf (file,
	       "DEFINE OWGBP (PS,Y)<\t%s one-word global byte pointer\n"
	       "<PS>B5+Y*ONE>\n",
	       ASM_COMMENT_START);
      fprintf (file,
	       "DEFINE GIW (Y)<Y*ONE>\t%s global indirect word\n",
	       ASM_COMMENT_START);
      fputs (TEXT_SECTION_ASM_OP, file);
      fputc ('\n', file);
    }
}

void
macro_file_end (file)
     FILE *file;
{
  if (TARGET_EXTENDED)
    fputs ("\t.ENDPS\n", file);
  fputs ("\tEND\n", file);
}

static void
macro_output_ascii_bytes (stream, ptr, len, byte_size)
     FILE *stream;
     const char *ptr;
     int len;
     int byte_size;
{
  if (len == 0)
    return;

  while (len--)
    pdp10_output_byte (stream, GEN_INT (*ptr++), byte_size);
}

static int
asciz_ok (ptr, len)
     const char *ptr;
     int len;
{
  int c;

  return 0;

  while (len--)
    {
      c = *ptr++;
      if (c >= 32 && c <= 126)
	continue;

      if (c == '\r' && len > 0 && *ptr == '\n')
	{
	  ptr++;
	  len--;
	  continue;
	}

      return 0;
    }

  return 1;
}

void
asm_output_asciz (stream, ptr, len)
     FILE *stream;
     const char *ptr;
     int len;
{
  int c;

  if (!asciz_ok (ptr, len))
    {
      macro_output_ascii_bytes (stream, ptr, len, 9);
      return;
    }

  fputs ("\tASCIZ\t/", stream);

  while (len--)
    {
      c = *ptr++;
      if (c == '\r')
	{
	  len--;
	  ptr++;
	  fputc ('\n', stream);
	}
      else if (c == 0 && len == 0)
	;
      else
	fputc (c, stream);
    }

  fputs ("/\n", stream);
}

static int bits_in_word;
static rtx last_output_byte = 0;
static int last_byte_size;

void
pdp10_output_byte (stream, exp, byte_size)
     FILE *stream;
     rtx exp;
     int byte_size;
{
  if (last_output_byte == NULL)
    {
      if (exp == NULL)
	return;
      fprintf (stream, "\tBYTE (%d)", byte_size);
      pdp10_print_integer (stream, exp);
      bits_in_word = byte_size;
    }
  else if (exp)
    {
      if (bits_in_word + byte_size > 36)
	{
	  pdp10_output_byte (stream, 0, 0);
	  pdp10_output_byte (stream, exp, byte_size);
	  return;
	}
      if (byte_size == last_byte_size)
	fputc (',', stream);
      else
	fprintf (stream, "(%d)", byte_size);
      pdp10_print_integer (stream, exp);
      bits_in_word += byte_size;
    }
  else
    fputc ('\n', stream);

  last_byte_size = byte_size;
  last_output_byte = exp;
}

void
macro_globalize_label (stream, name)
     FILE *stream;
     const char *name;
{
  pdp10_output_byte (stream, 0, 0);

  fputs ("\tENTRY\t", stream);
  assemble_name (stream, name);
  fputc ('\n', stream);
}

#endif /* !__PDP10_GAS_H__ */


/**********************************************************************

	Data Movement

**********************************************************************/

static int	pointer_offset PARAMS ((rtx x));
static const char *
		output_halfword_move PARAMS ((rtx reg, rtx mem, int store,
					      int sign_extend));
static const char *
		output_bitfield_move PARAMS ((int store, int sign, rtx reg,
					      rtx x, int len, int bit));
static int	use_xblt PARAMS ((rtx destination, rtx source, rtx length));
static int	expand_blt PARAMS ((rtx destination, rtx source, rtx length));
static int	expand_xblt PARAMS ((rtx destination, rtx source, rtx length));
static int	expand_blt_clear PARAMS ((rtx destination, rtx length));
static int	expand_xblt_clear PARAMS ((rtx destination, rtx length));
#if TARGET_STRING
static int	expand_movslj PARAMS ((rtx dest, rtx source, rtx length));
static int	expand_movst_clear PARAMS ((rtx destination, rtx length));
#endif

static int
pointer_offset (x)
     rtx x;
{
  /* This function must be prepared to handle anything
     GO_IF_LEGITIMATE_ADDRESS says is a valid address.  */

  switch (GET_CODE (x))
    {
    case CONST_INT:
    case SYMBOL_REF:
    case LABEL_REF:
    case REG:
    case PRE_INC:
    case MEM:
      return 0;
    case PLUS:
      x = XEXP (x, 0);
      if (GET_CODE (XEXP (x, 1)) == CONST_INT)
	return INTVAL (XEXP (x, 1));
      else
	return REGNO_POINTER_ALIGN (REGNO (XEXP (x, 0))) / BITS_PER_UNIT;
    case CONST:
      return INTVAL (XEXP (XEXP (x, 0), 0));
    default:
      abort ();
    }
}

static const char *
output_halfword_move (reg, mem, store, sign_extend)
     rtx reg;
     rtx mem;
     int store;
     int sign_extend;
{
  int mem_right = (pointer_offset (mem) % UNITS_PER_WORD != 0);

  if (GET_CODE (reg) == REG)
    {
      if (store)
	{
	  if (mem_right)
	    return "hrrm %0,%W1";
	  else
	    return "hrlm %0,%W1";
	}
      else
	{
	  const char *hrrz_hrre[2] =
	  { "hrrz %0,%W1", "hrre %0,%W1" };
	  const char *hlrz_hlre[2] =
	  { "hlrz %0,%W1", "hlre %0,%W1" };

	  if (mem_right)
	    return hrrz_hrre[sign_extend];
	  else
	    return hlrz_hlre[sign_extend];
	}
    }
  else if (GET_CODE (reg) == SUBREG)
    {
      const char *hll[2] =
      { "hll %0,%W1", "hllm %0,%W1" };
      const char *hrl[2] =
      { "hrl %0,%W1", "hlrm %0,%W1" };
      const char *hrr[2] =
      { "hrr %0,%W1", "hrrm %0,%W1" };
      const char *hlr[2] =
      { "hlr %0,%W1", "hrlm %0,%W1" };
      int reg_right = (SUBREG_WORD (reg) % 2 == 0);

      switch (mem_right + 2 * reg_right)
	{
	case 0: return hrr[store];
	case 1: return hlr[store];
	case 2: return hrl[store];
	case 3: return hll[store];
	default:
	  abort ();
	}
    }
  else
    abort ();
}

const char *
pdp10_output_movhi (operands, sign_extend)
     rtx *operands;
     int sign_extend;
{
  int store = memory_operand (operands[0], HImode);
  rtx x = store ? operands[0] : operands[1];
  int align;

  if (GET_CODE (x) == PRE_INC)
    {
      if (store)
	return "idpb %1,%0\t%; movhi";
      else
	return sign_extend ? "ildbe %0,%1\t%; movhi" : "ildb %0,%1\t%; movhi";
    }
  else if (GET_CODE (x) == POST_INC)
    {
      if (store)
	return "dpbi %1,%0\t%; movhi";
      else
	return sign_extend ? "ldbei %0,%1\t%; movhi" : "ldbi %0,%1\t%; movhi";
    }
  else if ((align = pointer_alignment (x)) != 0
	   && align % UNITS_PER_WORD == 0)
    {
      rtx reg = store ? operands[1] : operands[0];
      return output_halfword_move (reg, x, store, sign_extend);
    }
  else
    {
      if (store)
	return "dpb %1,%0\t%; movhi";
      else
	return sign_extend ? "ldbe %0,%1\t%; movhi" : "ldb %0,%1\t%; movhi";
    }
}

void
pdp10_expand_extv (operands)
     rtx *operands;
{
  int len = INTVAL (operands[2]);
  unsigned HOST_WIDE_INT sign = 1LL << len;
  rtx label = gen_label_rtx ();

  emit_insn (gen_extzv (operands[0], operands[1],
			operands[2], operands[3]));
  emit_insn (gen_test_extract_skip (gen_rtx_EQ (VOIDmode,
						NULL_RTX,
						NULL_RTX),
				    operands[0],
				    operands[2],
				    GEN_INT (1),
				    label));
  emit_insn (gen_iorsi3 (operands[0], operands[0], GEN_INT (sign - 1)));
  emit_label (label);
}

static const char *
output_bitfield_move (store, sign, reg, x, len, bit)
     int store;
     int sign;
     rtx reg;
     rtx x;
     int len;
     int bit;
{
  if (!TARGET_EXTENDED)
    return
      store ? "dpb %3,[POINT %1,%0,%2]" :
      sign ? "ldbe %0,[POINT %2,%1,%3]" :
	     "ldb %0,[POINT %2,%1,%3]";

  if (target_flags & ASM_UPPER)
    fputs (store ? "\tDPB" : sign ? "\tLDBE" : "\tLDB", asm_out_file);
  else
    fputs (store ? "\tdpb" : sign ? "\tldbe" : "\tldb", asm_out_file);
  fputc (target_flags & ASM_TAB ? '\t' : ' ', asm_out_file);
  print_reg (asm_out_file, REGNO (reg));
  fputs (",[", asm_out_file);

  if (VALID_BYTE_SIZE_P (len)
      && (bit % len == 0)
      /* Bitfield must be in register or memory.  */
      && (GET_CODE (x) == REG
	  || (GET_CODE (x) == MEM
	      /* Memory address must be a constant.  */
	      && CONSTANT_ADDRESS_P (XEXP (x, 0)))))
    {
      /* Use a one-word global byte pointer.  */

      static const char *byte6_selector[] = { "46", "47", "50",
					      "51", "52", "53" };
      static const char *byte7_selector[] = { "62", "63", "64", "65", "66" };
      static const char *byte8_selector[] = { "55", "56", "57", "60" };
      static const char * char_selector[] = { "70", "71", "72", "73" };
      static const char *short_selector[] = { "75", "76" };
      const char **byte_selector;

      switch (len)
	{
	case  6: byte_selector = byte6_selector; break;
	case  7: byte_selector = byte7_selector; break;
	case  8: byte_selector = byte8_selector; break;
	case  9: byte_selector =  char_selector; break;
	case 18: byte_selector = short_selector; break;
	}

      fputs ("OWGBP ", asm_out_file);
      fputs (byte_selector[bit / len], asm_out_file);
      fputc (',', asm_out_file);

      if (GET_CODE (x) == REG)
	print_reg (asm_out_file, REGNO (x));
      else
	{
	  x = XEXP (x, 0);
	  switch (GET_CODE (x))
	    {
	    case CONST_INT:
	      pdp10_print_address (asm_out_file, x);
	      break;
	    case SYMBOL_REF:
	      output_addr_const (asm_out_file, x);
	      break;
	    case CONST:
	      x = XEXP (x, 0);
	      output_addr_const (asm_out_file, XEXP (x, 0));
	      x = XEXP (x, 1);
	      if (INTVAL (x) >= UNITS_PER_WORD)
		fputc ('+', asm_out_file);
	      pdp10_print_offset30 (asm_out_file, x);
	      break;
	    default:
	      abort ();
	    }
	}
    }
  else
    {
      /* Use a two-word global byte pointer.  */

      fprintf (asm_out_file, "%02o%02o40,,0\n\t\t", 36 - bit - len, len);

      if (GET_CODE (x) == REG
	  || (GET_CODE (x) == MEM && GET_CODE (XEXP (x, 0)) == PLUS))
	pdp10_print_operand (asm_out_file, x, 0);
      else
	print_word_pointer (asm_out_file, XEXP (x, 0), 0);
    }

  fputs ("]\n", asm_out_file);
  return "";
}

const char *
pdp10_output_extv (operands)
     rtx *operands;
{
  return output_bitfield_move (0, 1, operands[0], operands[1],
			       INTVAL (operands[2]), INTVAL (operands[3]));
}

const char *
pdp10_output_extzv (operands)
     rtx *operands;
{
  return output_bitfield_move (0, 0, operands[0], operands[1],
			       INTVAL (operands[2]), INTVAL (operands[3]));
}

const char *
pdp10_output_insv (operands)
     rtx *operands;
{
  return output_bitfield_move (1, 0, operands[3], operands[0],
			       INTVAL (operands[1]), INTVAL (operands[2]));
}

static int
use_xblt (destination, source, length)
     rtx destination;
     rtx source;
     rtx length;
{
  if (!HAVE_XBLT || !TARGET_EXTENDED)
    return 0;

  if (GET_CODE (destination) != CONST_INT
      || (source && GET_CODE (source) != CONST_INT)
      || GET_CODE (length) != CONST_INT)
    return 1;

  if (INTVAL (length) >> 32 > 0)
    return 1;

  if (source)
    {
      HOST_WIDE_INT destination_section = INTVAL (destination) >> 32;
      HOST_WIDE_INT source_section = INTVAL (source) >> 32;

      if (destination_section != source_section)
	return 1;
    }

  return 0;
}

static int
expand_blt (destination, source, length)
     rtx destination;
     rtx source;
     rtx length;
{
  rtx ac, e;

  if (GET_CODE (length) != CONST_INT
      || INTVAL (length) % UNITS_PER_WORD != 0)
    return 0;

  ac = gen_reg_rtx (PSImode);
  if (CONSTANT_ADDRESS_P (destination)
      && CONSTANT_ADDRESS_P (source))
    {
      emit_insn (gen_move_two_halves (ac, source, destination));
    }
  else
    {
      emit_insn (gen_MOVEI (ac, destination));
      emit_insn (gen_HRLI (ac, source));
    }

  {
    HOST_WIDE_INT len = INTVAL (length) - UNITS_PER_WORD;
    if (CONSTANT_ADDRESS_P (destination))
      e = plus_constant (destination, len);
    else
      {
	e = gen_reg_rtx (PSImode);
	emit_insn (gen_addwp3 (e, destination, GEN_INT (len)));
      }
  }

  /* Beginning at the location addressed by the left half-word of AC,
     move words to another area beginning at the location addressed by
     the right half-word of AC.  Continue until a word is moved to
     location E.  */
  emit_insn (gen_BLT (ac, gen_rtx_MEM (BLKmode, e)));

  return 1;
}

static int
expand_xblt (destination, source, length)
     rtx destination;
     rtx source;
     rtx length;
{
  rtx acs;

  if (GET_CODE (length) != CONST_INT
      || INTVAL (length) % UNITS_PER_WORD != 0)
    return 0;

  acs = gen_reg_rtx (TImode);
  emit_move_insn (gen_rtx_SUBREG (SImode, acs, 0),
		  GEN_INT (INTVAL (length) / UNITS_PER_WORD));
  emit_move_insn (gen_rtx_SUBREG (Pmode, acs, 1), source);
  emit_move_insn (gen_rtx_SUBREG (Pmode, acs, 2), destination);
  emit_insn (gen_XBLT (acs));

  return 1;
}

#if TARGET_STRING
static int
expand_movslj (destination, source, length)
     rtx destination;
     rtx source;
     rtx length;
{
  abort ();
}
#endif

int
pdp10_expand_movstrsi (operands)
     rtx *operands;
{
  /* operands[0] is the block destination address
     operands[1] is the block source address
     operands[2] is the block length
     operands[3] is the known alignment.  */

#if TARGET_STRING
  if (TARGET_KL10up && TARGET_STRING)
    return expand_movslj (operands[0], operands[1], operands[2]);
#endif

  operands[0] = XEXP (operands[0], 0);
  operands[1] = XEXP (operands[1], 0);

  if (GET_CODE (operands[3]) != CONST_INT
      || INTVAL (operands[3]) != UNITS_PER_WORD)
    return 0;

  if (use_xblt (operands[0], operands[1], operands[2]))
    return expand_xblt (operands[0], operands[1], operands[2]);
  else
    return expand_blt (operands[0], operands[1], operands[2]);
}

static int
expand_blt_clear (destination, length)
     rtx destination;
     rtx length;
{
  HOST_WIDE_INT len;
  rtx ac, e;

  if (GET_CODE (length) != CONST_INT
      || INTVAL (length) % UNITS_PER_WORD != 0)
    return 0;

  len = INTVAL (length);
  ac = gen_reg_rtx (PSImode);

  if (CONSTANT_ADDRESS_P (destination))
    {
      emit_insn (gen_move_two_halves (ac,
				      destination,
				      plus_constant (destination,
						     UNITS_PER_WORD)));
      e = plus_constant (destination, len - UNITS_PER_WORD);
    }
  else
    {
      emit_insn (gen_MOVSI (ac, destination));
      emit_insn (gen_HRRI (ac, plus_constant (destination, UNITS_PER_WORD)));
      e = gen_reg_rtx (PSImode);
      emit_insn (gen_addwp3 (e, destination,
			     GEN_INT (len - UNITS_PER_WORD)));
    }

  emit_move_insn (gen_rtx_MEM (SImode, destination), const0_rtx);

  /* Beginning at the location addressed by the left half-word of AC,
     move words to another area beginning at the location addressed by
     the right half-word of AC.  Continue until a word is moved to
     location E.  */
  emit_insn (gen_BLT (ac, gen_rtx_MEM (BLKmode, e)));

  return 1;
}

static int
expand_xblt_clear (destination, length)
     rtx destination;
     rtx length;
{
  rtx acs;

  if (GET_CODE (length) != CONST_INT
      || INTVAL (length) % UNITS_PER_WORD != 0)
    return 0;

  acs = gen_reg_rtx (TImode);
  emit_move_insn (gen_rtx_SUBREG (SImode, acs, 0),
		  GEN_INT (INTVAL (length) / UNITS_PER_WORD));
  emit_move_insn (gen_rtx_SUBREG (Pmode, acs, 1), destination);
  emit_insn (gen_MOVEI (gen_rtx_SUBREG (Pmode, acs, 2),
			plus_constant (destination, UNITS_PER_WORD)));
  emit_move_insn (gen_rtx_MEM (SImode, destination), const0_rtx);
  emit_insn (gen_XBLT (acs));

  return 1;
}

#if TARGET_STRING
static int
expand_movst_clear (destination, length)
     rtx destination;
     rtx length;
{
  abort ();
}
#endif

int
pdp10_expand_clrstrsi (operands)
     rtx *operands;
{
  /* operands[0] is the block destination address
     operands[1] is the block length
     operands[2] is the known alignment.  */

#if TARGET_STRING
  if (TARGET_KL10up && TARGET_STRING)
    return expand_movst_clear (operands[0], operands[1]);
#endif

  if (GET_CODE (operands[2]) != CONST_INT
      || INTVAL (operands[2]) != UNITS_PER_WORD)
    return 0;

  operands[0] = XEXP (operands[0], 0);

  if (use_xblt (operands[0], NULL_RTX, operands[1]))
    return expand_xblt_clear (operands[0], operands[1]);
  else
    return expand_blt_clear (operands[0], operands[1]);
}

#if TARGET_STRING
int
pdp10_expand_cmpstrsi (operands)
     rtx *operands ATTRIBUTE_UNUSED;
{
  /* operands[0] is the result
     operands[1] is the address of the first string
     operands[2] is the address of the second string
     operands[3] is the string length
     operands[4] is the known alignmen.  */

  if (!TARGET_KL10up || !TARGET_STRING)
    return 0;

  abort ();
}
#endif


/**********************************************************************

	Pointer Arithmetic

**********************************************************************/

static void	expand_adjust_byte_pointer PARAMS ((enum machine_mode mode,
						    rtx op0,
						    rtx op1,
						    rtx op2));
static int	maybe_expand_ibp PARAMS ((rtx op0, rtx op1, int increment));
static void	expand_addbp3 PARAMS ((enum machine_mode mode,
				       rtx op0, rtx op1, rtx op2));
static void	expand_add_or_sub_pointer PARAMS ((rtx *operands, int add));
static void	expand_subtract_pointers PARAMS ((rtx *operands));

static int
pointer_bytesize (type)
     tree type;
{
  int size, max_size, min_size;
  tree t, x;

  if (TREE_CODE (type) != RECORD_TYPE)
    return TYPE_ALIGN (type);

  min_size = 36;
  max_size = 1;
  for (x = TYPE_FIELDS (type); x; x = TREE_CHAIN (x))
    {
      if (TREE_CODE (x) != FIELD_DECL)
	continue;

      t = TREE_TYPE (x);
      if (TREE_CODE (t) == RECORD_TYPE)
	size = pointer_bytesize (t);
      else
	size = TREE_INT_CST_LOW (TYPE_SIZE (t));
      if (size < min_size)
	min_size = size;
      if (size > max_size)
	max_size = size;
    }

  if (min_size == max_size)
    return min_size;

  return TYPE_ALIGN (type);
}

static tree
get_record_pointer_type (record_type, offset)
     tree record_type;
     HOST_WIDE_INT offset;
{
  tree type;
  tree x;

  offset *= BITS_PER_UNIT;

  for (x = TYPE_FIELDS (record_type); x && offset >= 0; x = TREE_CHAIN (x))
    {
      if (TREE_CODE (x) != FIELD_DECL)
	continue;

      type = TREE_TYPE (x);
      if (TREE_CODE (type) == RECORD_TYPE)
	{
	  tree t = get_record_pointer_type (t, offset);
	  if (t)
	    return t;
	}
      if (TREE_CODE (type) == UNION_TYPE)
	abort ();

      if (offset == 0)
	return type;

      offset -= TREE_INT_CST_LOW (TYPE_SIZE (type));
    }

  /* Bitfield.  */
  return integer_type_node;
}

tree
get_pointer_type (x)
     rtx x;
{
  tree type;

  switch (GET_CODE (x))
    {
    case REG:
      if (REG_POINTER (x))
	return REGNO_POINTER_TYPE (REGNO (x));
      else
	return NULL_TREE;

    case SUBREG:
      return get_pointer_type (XEXP (x, 0));

    case MEM:
      type = get_pointer_type (XEXP (x, 0));
      if (type == NULL_TREE || !POINTER_TYPE_P (type))
	return NULL_TREE;
      return TREE_TYPE (type);

    case CONST:
      x = XEXP (x, 0);
      if (GET_CODE (x) != PLUS)
	abort ();
      /* Fall through.  */

    case PLUS:
    case MINUS:
      type = get_pointer_type (XEXP (x, 0));
      if (!type)
	return NULL_TREE;
      switch (TREE_CODE (type))
	{
	case ARRAY_TYPE:
	  abort ();
	case POINTER_TYPE:
	case REFERENCE_TYPE:
	  abort ();
	case RECORD_TYPE:
	  if (GET_CODE (XEXP (x, 0)) != REG
	      && GET_CODE (XEXP (x, 0)) != SYMBOL_REF)
	    abort ();
	  if (GET_CODE (XEXP (x, 1)) != CONST_INT)
	    abort ();
	  type = get_record_pointer_type (type, INTVAL (XEXP (x, 1)));
	  if (!type)
	    abort ();
	  return type;
	default:
	  abort ();
	}

    case SYMBOL_REF:
      type = lookup_symbol_type (XSTR (x, 0));
      if (type && TREE_CODE (type) == FUNCTION_TYPE)
	type = build_pointer_type (type);
      return type;

    case CONST_INT:
      return NULL_TREE;

    case CONST_DOUBLE:
      return NULL_TREE;

    default:
      abort ();
    }
}

static void
expand_adjust_byte_pointer (mode, op0, op1, op2)
     enum machine_mode mode;
     rtx op0;
     rtx op1;
     rtx op2;
{
  int bytes_per_word = UNITS_PER_WORD / GET_MODE_SIZE (mode);
  rtx reg;
  rtx mod;

  abort ();

  /* char9:
     MOVE op0,op1
     MOVE op3,op2
     ANDI op3,3
     ASH op2,-2
     ADD op0,op2
     TRZE op3,1		op3=0 => 2 insns
      IBP op0		op3=1 => 3 insns
     JUMPE op3,%Lnn	op3=2 => 4 insns
      IBP op0		op3=3 => 5 insns
      IBP op0
   %Lnn:

     char7:
     MOVE op0,op1
     IDIV op2,5		op3=0 => 2 insns
     ADD op0,op2	op3=1 => 4 insns
     JRST %Lmm		op3=2 => 6 insns
%Lnn:IBP op0
%Lmm:SOJGE op3,%Lnn */

  /* Move the source to the destination.  */
  if (!register_operand (op0, Pmode))
    reg = force_reg (Pmode, op0);
  emit_move_insn (reg, op1);

  /* Divide the increment by the number of bytes per word.  */
  mod = gen_reg_rtx (Pmode);
  op2 = force_reg (Pmode, op2);
  switch (bytes_per_word)
    {
    case 2:
      emit_insn (gen_ashrpsi3 (op2, op2, GEN_INT (1)));
      break;
    case 4:
      emit_insn (gen_ashrpsi3 (op2, op2, GEN_INT (2)));
      break;
    case 5:
      emit_insn (gen_IDIVM (op2, op2, GEN_INT (5)));
      break;
    default:
      abort ();
    }

  emit_insn (gen_addwp3 (op0, op0, op2));

  if (!register_operand (op0, Pmode))
    emit_move_insn (op0, reg);
}

static int
maybe_expand_ibp (op0, op1, increment)
     rtx op0;
     rtx op1;
     int increment;
{
  int count_moves = 0;
  int op0_eq_op1, use_temp, reg0, reg1;
  int ibp_cost, adjbp_cost;
  rtx inc;
  int i;

  if (increment < 0)
    return 0;

  if (increment == 0)
    return 1;

  op0_eq_op1 = rtx_equal_p (op0, op1);
  reg0 = register_operand (op0, Pmode);
  reg1 = register_operand (op1, Pmode);
  use_temp = 0;

  /* The time costs are calculated like this:
      * an instruction costs 1 unit
      * a memory read or write costs 1 unit
      * an adjust/increment byte pointer operation costs 1 uni.  */

  if (reg0)
    {
      /* (MOVE op0,op1)
	 IBP op0
	 IBP op0
	 ... */
      if (optimize_size)
	{
	  ibp_cost = increment;
	  if (count_moves && !op0_eq_op1)
	    ibp_cost += 1;
	}
      else
	{
	  ibp_cost = 2 * increment;
	  if (count_moves && !op0_eq_op1)
	    ibp_cost += reg1 ? 1 : 2;
	}
    }
  else
    {
      /* MOVE tmp,op1
	 IBP tmp
	 IBP tmp
	 ...
	 MOVEM tmp,op0 */
      use_temp = 1;
      if (optimize_size)
	{
	  ibp_cost = increment;
	  if (count_moves)
	    ibp_cost += 2;
	}
      else
	{
	  ibp_cost = 2 * increment;
	  if (count_moves)
	    ibp_cost += (reg1 ? 1 : 2) + 2;
	}

      if (op0_eq_op1 || reg1)
	{
	  int ibp_cost_without_temp;

	  /* (MOVEM op1,op0)
	     IBP op0
	     ... */
	  if (optimize_size)
	    {
	      ibp_cost_without_temp = increment;
	      if (count_moves && !op0_eq_op1)
		ibp_cost_without_temp += 1;
	    }
	  else
	    {
	      ibp_cost_without_temp = 4 * increment;
	      if (count_moves && !op0_eq_op1)
		ibp_cost_without_temp += 1;
	    }

	  if (ibp_cost_without_temp < ibp_cost)
	    {
	      ibp_cost = ibp_cost_without_temp;
	      use_temp = 0;
	    }
	}
    }

  /* MOVE(I) op0/tmp,op2
     ADJBP op0/tmp,op1
     (MOVEM tmp,op0) */
  if (optimize_size)
    {
      adjbp_cost = 2;
      if (count_moves && !reg0)
	adjbp_cost += 1;
    }
  else
    {
      adjbp_cost = 1 + (reg1 ? 2 : 3);
      if (count_moves && !reg0)
	adjbp_cost += 2;
    }

  /* Don't use a sequence of IBPs if the cost is higher than using an
     ADJBP.

     On an older target, don't use a sequence of IBPs if the cost is
     higher than using another sequence to adjust a byte pointer.  */
  if ((HAVE_ADJBP && ibp_cost > adjbp_cost)
      || ibp_cost > 10)
    return 0;

  if (use_temp)
    {
      inc = gen_reg_rtx (Pmode);
      emit_move_insn (inc, op1);
    }
  else
    {
      inc = op0;
      if (!op0_eq_op1)
	emit_move_insn (op0, op1);
    }

  for (i = 0; i < increment; i++)
    emit_insn (gen_IBP (inc, inc));

  if (use_temp)
    emit_move_insn (op0, inc);

  return 1;
}

static void
expand_addbp3 (mode, op0, op1, op2)
     enum machine_mode mode;
     rtx op0;
     rtx op1;
     rtx op2;
{
  if (GET_CODE (op2) == CONST_INT)
    {
      if (maybe_expand_ibp (op0, op1, INTVAL (op2)))
	return;
    }

  /* If we get here, either the increment is not a constant, or a
     simple sequence of IBP instructions is too costly.  */

  if (HAVE_ADJBP)
    {
      rtx x;

      if (!register_operand (op0, Pmode))
	x = force_reg (Pmode, op0);
      else
	x = op0;
      emit_move_insn (x, op2);
      emit_insn (gen_ADJBP (x, op1, x));
      if (!register_operand (op0, Pmode))
	emit_move_insn (op0, x);
    }
  else
    {
      /* Emit sequence to adjust a byte pointer without ADJBP.  */
      expand_adjust_byte_pointer (mode, op0, op1, op2);
    }
}

static tree
check_that_pointers_match (rtx op0, rtx op1)
{
  tree type0, type1;

  type0 = get_pointer_type (op0);
  type1 = get_pointer_type (op1);

  if (!type0)
    abort ();
  if (!type1)
    abort ();
  if (type0 != type1
      && TREE_CODE (type0) != VOID_TYPE
      && TREE_CODE (type1) != VOID_TYPE)
    abort ();

  return type0;
}

static void
expand_add_or_sub_pointer (operands, add)
     rtx *operands;
     int add;
{
  tree type;

  type = check_that_pointers_match (operands[0], operands[1]);

  if (WORD_TYPE_P (type))
    {
      /* Add to a word pointer by first dividing the increment by the
         number of storage units per word, which is always four.
         Then, add the increment to the pointer.  */

      rtx increment;

      if (UNITS_PER_WORD != 4)
	abort ();

      if (GET_CODE (operands[2]) == CONST_INT)
	increment = GEN_INT (INTVAL (operands[2]) / UNITS_PER_WORD);
      else
	{
	  increment = gen_reg_rtx (Pmode);
	  emit_insn (gen_ashrpsi3 (increment, operands[2], GEN_INT (2)));
	}

      if (add)
	emit_insn (gen_addwp3 (operands[0], operands[1], increment));
      else
	emit_insn (gen_subwp3 (operands[0], operands[1], increment));
    }
  else
    {
      /* Add to a byte pointer by first dividing the increment by the
         number of bytes per word, which depends on the byte size.
         Then, use an IBP or ADJBP instruction.  */

      rtx increment;
      int size;

      size = pointer_bytesize (type);
      if (size == -1 && TREE_CODE (type) == VOID_TYPE)
	size = 9;

      switch (size)
	{
	case 7:
	case 8:
	case 9:
	  increment = operands[2];
	  break;

	case 16:
	case 18:
	  if (GET_CODE (operands[2]) == CONST_INT)
	    increment = GEN_INT (INTVAL (operands[2]) / 2);
	  else
	    {
	      increment = gen_reg_rtx (Pmode);
	      emit_insn (gen_ashrpsi3 (increment, operands[2], GEN_INT (1)));
	    }
	  break;

	default:
	  abort ();
	}

      if (!add)
	{
	  if (GET_CODE (increment) == CONST_INT)
	    increment = GEN_INT (-INTVAL (increment));
	  else
	    emit_insn (gen_negpsi2 (increment, increment));
	}

      expand_addbp3 (TYPE_MODE (type), operands[0], operands[1], increment);
    }

  /* This is to fool GCC into not generating a REG_EQUAL note for this
     sequence.  */
  emit_move_insn (operands[1], operands[1]);
}

static void
expand_subtract_pointers (operands)
     rtx *operands;
{
  tree type;

  type = check_that_pointers_match (operands[1], operands[2]);

  if (WORD_TYPE_P (type))
    {
      /* Calculate the difference between two word pointers by
	 subtracting and multiplying by the number of storage units
	 per word.  */

      rtx x = gen_reg_rtx (SImode);

      /*emit_insn (gen_subwp3 (x, operands[1], operands[2]));*/
      /*emit_insn (gen_ashlsi3 (operands[0], x, GEN_INT (2)));*/
      emit_insn (gen_subsi3_internal (x, operands[1], operands[2]));
      emit_insn (gen_ashlsi3_pointer (operands[0], x, GEN_INT (2)));
    }
  else
    {
      /* Calculate the difference between two byte pointers ...  */
      abort ();
    }

  /* This is to fool GCC into not generating a REG_EQUAL note for this
     sequence.  */
  emit_move_insn (operands[1], operands[1]);
}

int
pdp10_expand_subsi3 (operands)
     rtx *operands;
{
  if (!get_pointer_type (operands[1]) || !get_pointer_type (operands[2]))
    return 0;

  expand_subtract_pointers (operands);
  return 1;
}

void
pdp10_expand_addpsi3 (operands)
     rtx *operands;
{
  expand_add_or_sub_pointer (operands, 1);
}

void
pdp10_expand_subpsi3 (operands)
     rtx *operands;
{
  if (get_pointer_type (operands[1]) && get_pointer_type (operands[2]))
    expand_subtract_pointers (operands);
  else
    expand_add_or_sub_pointer (operands, 0);
}


/**********************************************************************

	Function Prologue and Epilogue

**********************************************************************/

static rtx	emit_frame_insn PARAMS ((rtx pattern));
static void	expand_save_or_restore_return_address PARAMS ((int save));
static void	expand_save_pretend_args PARAMS ((void));
static void	expand_save_or_restore_regs PARAMS ((int save,
						     int *clobbered,
						     int single));
static void	expand_set_up_frame_pointer PARAMS ((void));

static int
clobbered_regs (clobbered)
     int *clobbered;
{
  int i, n;

  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
    clobbered[i] = regs_ever_live[i] && !call_used_regs[i];

  clobbered[FRAME_POINTER_REGNUM] = frame_pointer_needed;
  clobbered[STACK_POINTER_REGNUM] = 0;

  n = 0;
  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
    {
      if (clobbered[i])
	n += UNITS_PER_WORD;
    }

  return n;
}

static rtx
emit_frame_insn (pattern)
     rtx pattern;
{
  rtx insn = emit_insn (pattern);
  RTX_FRAME_RELATED_P (insn) = 1;
  return insn;
}

static void
expand_save_or_restore_return_address (save)
     int save;
{
  int offset1, offset2;

  if (!current_function_pretend_args_size)
    return;

  /* If we need to store pretend args, move the return address out of
     their way.  */

  offset1 = - (current_function_outgoing_args_size +
	       current_function_frame_size +
	       current_function_clobbered_size +
	       current_function_pretend_args_size);
  offset2 = - (current_function_outgoing_args_size +
	       current_function_frame_size +
	       current_function_clobbered_size);

  if (!save)
    {
      int tmp = offset1;
      offset1 = offset2;
      offset2 = tmp;
    }

  emit_move_insn (gen_rtx_REG (SImode, 0),
		  GEN_MEM_SP_PLUS (SImode, offset1));
  emit_move_insn (GEN_MEM_SP_PLUS (SImode, offset2),
		  gen_rtx_REG (SImode, 0));
}

static void
expand_save_pretend_args ()
{
  int offset = -(current_function_outgoing_args_size +
		 current_function_frame_size +
		 current_function_clobbered_size +
		 current_function_pretend_args_size);
  int i;

  for (i = 4;
       i > 4 - current_function_pretend_args_size / UNITS_PER_WORD;
       i--)
    {
      emit_move_insn (GEN_MEM_SP_PLUS (SImode, offset),
		      gen_rtx_REG (SImode, i));
      offset += UNITS_PER_WORD;
    }
}

/* Save or restore clobbered callee-saved registers.  SAVE is nonzero
   when saving, and zero when restoring.  CLOBBERED indicates
   clobbered registers.  SINGLE is nonzero when a single PUSH or POP
   instruction can be used.  */
static void
expand_save_or_restore_regs (save, clobbered, single)
     int save;
     int *clobbered;
     int single;
{
  int max_move;
  int min_blt;
  int offset;
  int regs;
  int i, j, k;

  if (current_function_clobbered_size == 0)
    return;

  offset = -(current_function_outgoing_args_size +
	     current_function_frame_size +
	     current_function_clobbered_size) + UNITS_PER_WORD;

  if (HAVE_movdi)
    max_move = 2;
  else
    max_move = 1;
  min_blt = 3 * max_move + 1;

  for (i = 0; i < FIRST_PSEUDO_REGISTER; i += regs)
    {
      regs = 1;
      if (!clobbered[i])
	continue;

      /* If a single PUSH or POP instruction can be used, do so and
	 return.  */

      if (single)
	{
	  rtx stack;

	  if (save)
	    {
	      stack = gen_rtx_PRE_INC (Pmode, stack_pointer_rtx);
	      emit_frame_insn (gen_PUSH (gen_rtx_MEM (SImode, stack),
					 gen_rtx_REG (SImode, i)));
	    }
	  else
	    {
	      stack = gen_rtx_POST_DEC (Pmode, stack_pointer_rtx);
	      emit_frame_insn (gen_POP (gen_rtx_REG (SImode, i),
					gen_rtx_MEM (SImode, stack)));
	    }

	  return;
	}

      /* Find a consecutive block of clobbered registers.

	 When extended addressing is in use, the stack can span
	 multiple sections, so BLT can't be used.  In this case, the
	 search is limited to just examine the next register.  */

      k =
	TARGET_EXTENDED && i <= FIRST_PSEUDO_REGISTER - 2 ?
	i + 2 : FIRST_PSEUDO_REGISTER;
      for (j = i + 1; j < k; j++)
	{
	  if (clobbered[j])
	    regs++;
	  else
	    break;
	}
      
      if (regs >= min_blt)
	{
	  /* Use a BLT instruction.  */

	  rtx from_start, to_start, to_end;

	  if (save)
	    {
	      from_start = GEN_INT (4 * i);
	      to_start = gen_rtx_PLUS (Pmode,
				       stack_pointer_rtx,
				       GEN_INT (offset));
	      to_end = GEN_MEM_SP_PLUS (SImode,
					offset + (regs - 1) * UNITS_PER_WORD);
	    }
	  else
	    {
	      from_start = gen_rtx_PLUS (Pmode,
					 stack_pointer_rtx,
					 GEN_INT (offset));
	      to_start = GEN_INT (4 * i);
	      to_end = gen_rtx_MEM (SImode, GEN_INT (4 * (i + regs - 1)));
	    }

#define REG0 gen_rtx_REG (Pmode, 0)
	  emit_frame_insn (gen_MOVEI (REG0, to_start));
	  emit_frame_insn (gen_HRLI (REG0, from_start));
	  emit_frame_insn (gen_BLT (REG0, to_end));
#undef REG0
	}
      else
	{
	  /* Use a single MOVE or DMOVE instruction.  */

	  enum machine_mode mode;
	  rtx insn;

	  if (regs > max_move)
	    regs = max_move;
	  mode = (regs == 1 ? SImode : DImode);

	  if (save)
	    insn = emit_move_insn (GEN_MEM_SP_PLUS (mode, offset),
				   gen_rtx_REG (mode, i));
	  else
	    insn = emit_move_insn (gen_rtx_REG (mode, i),
				   GEN_MEM_SP_PLUS (mode, offset));
	  RTX_FRAME_RELATED_P (insn) = 1;
	}

      offset += regs * UNITS_PER_WORD;
    }
}

static void
expand_set_up_frame_pointer ()
{
  int offset;

  if (!frame_pointer_needed)
    return;

  offset = -(current_function_outgoing_args_size +
	     current_function_frame_size +
	     current_function_clobbered_size);

  emit_frame_insn (gen_MOVEI (frame_pointer_rtx,
			      gen_rtx_PLUS (Pmode,
					    stack_pointer_rtx,
					    GEN_INT (offset))));
}

void
pdp10_expand_prologue ()
{
  int clobbered[FIRST_PSEUDO_REGISTER];
  int stack_adjust;
  int single_push;

  current_function_frame_size = get_frame_size ();
  current_function_clobbered_size = clobbered_regs (clobbered);

  /* If only one word is to be saved on stack, it can be stored with a
     single PUSH instruction.  */
  single_push =
    current_function_clobbered_size == UNITS_PER_WORD
    && current_function_outgoing_args_size == 0
    && current_function_frame_size == 0;

  stack_adjust =
    current_function_outgoing_args_size
    + current_function_frame_size
    + current_function_clobbered_size
    + current_function_pretend_args_size;
  if (!single_push && stack_adjust > 0)
    emit_frame_insn (gen_ADJSP (stack_pointer_rtx,
				GEN_INT (stack_adjust / UNITS_PER_WORD)));

  expand_save_or_restore_return_address (SAVE);
  expand_save_pretend_args ();
  expand_save_or_restore_regs (SAVE, clobbered, single_push);
  expand_set_up_frame_pointer ();
}

void
pdp10_expand_epilogue ()
{
  int clobbered[FIRST_PSEUDO_REGISTER];
  int stack_adjust;
  int single_pop;

  current_function_clobbered_size = clobbered_regs (clobbered);

  /* If only one word was saved on stack, it can be restored with a
     single POP instruction.  */
  single_pop =
    current_function_clobbered_size == UNITS_PER_WORD
    && current_function_outgoing_args_size == 0
    && current_function_frame_size == 0;

  expand_save_or_restore_regs (RESTORE, clobbered, single_pop);
  expand_save_or_restore_return_address (RESTORE);

  stack_adjust =
    current_function_outgoing_args_size
    + current_function_frame_size
    + current_function_clobbered_size
    + current_function_pretend_args_size;
  if (!single_pop && stack_adjust > 0)
    emit_frame_insn (gen_ADJSP (stack_pointer_rtx,
				GEN_INT (-stack_adjust / UNITS_PER_WORD)));

  emit_jump_insn (gen_POPJ_17 ());
}



rtx pdp10_compare_op0, pdp10_compare_op1;

static void
pdp10_print_number (stream, x, div, bits, zero)
     FILE *stream;
     rtx x;
     int div;
     int bits;
     int zero;
{
  HOST_WIDE_INT i = INTVAL (x);
  HOST_WIDE_INT mask;

  if (GET_CODE (x) != CONST_INT)
    abort ();

#if 0
  if (i % div)
    fprintf (stream, "pdp10.c:%d:UNALIGNED:", __LINE__);
#endif
  if (i < 0)
    i -= div - 1;
  i /= div;

  if (bits < 0)
    {
      bits = -bits;
      if (i < 0)
	{
	  fputc ('-', stream);
	  i = -i;
	}
    }

  mask = ~0LL << bits;
  if ((i & mask) != 0 && (i & mask) != mask)
    fprintf (stream, "pdp10.c:%d:TOOBIG:", __LINE__);

  i &= ~mask;

  if (zero == 0 || i > 0)
    fprintf (stream, HOST_WIDE_INT_PRINT_OCT, i);
}

void
set_pointer_attributes (target, code, icode, op0, op1)
     rtx target;
     enum rtx_code code;
     int icode;
     rtx op0;
     rtx op1;
{
  tree type0;
  tree type1;
  int align;
  int indirect;

  indirect = 0;
  while (GET_CODE (target) != REG)
    switch (GET_CODE (target))
      {
      case MEM:
	target = XEXP (target, 0);
	indirect++;
	break;
      case PLUS:
      case SUBREG:
	target = XEXP (target, 0);
	break;
      case SYMBOL_REF:
	return;
      default:
	abort ();
      }

  if (REG_POINTER (target) && REGNO_POINTER_TYPE (REGNO (target)))
    return;

  type0 = get_pointer_type (op0);
  type1 = op1 ? get_pointer_type (op1) : NULL_TREE;
  if (type0 == NULL_TREE && type1 == NULL_TREE)
    return;

  switch (code)
    {
    case MINUS:
      if ((type0 != NULL_TREE && type1 == NULL_TREE)
	  || (type0 == NULL_TREE && type1 != NULL_TREE))
	{
	  if (type0)
	    align = REGNO_POINTER_ALIGN (REGNO (op0));
	  else if (type1)
	    {
	      type0 = type1;
	      align = REGNO_POINTER_ALIGN (REGNO (op1));
	    }

	  if (align == 0)
	    align = TYPE_ALIGN (type0);
	}
      else
	type0 = NULL_TREE;
      break;

    case PLUS:
    case MULT:
      align = GET_CODE (op0) == REG && REG_POINTER (op0) ?
	REGNO_POINTER_ALIGN (REGNO (op0)) : -1;
      if (!type0)
	{
	  type0 = type1;
	  align = GET_CODE (op1) == REG && REG_POINTER (op1) ?
	    REGNO_POINTER_ALIGN (REGNO (op1)) : -1;
	}
      if (type0 && align == -1)
	align = TYPE_ALIGN (type0);
      break;

    case SET:
    case ASHIFT:
    case ASHIFTRT:
    case ZERO_EXTEND:
      align = GET_CODE (op0) == REG && REG_POINTER (op0) ?
	REGNO_POINTER_ALIGN (REGNO (op0)) : -1;
      if (type0 && align == -1)
	align = TYPE_ALIGN (type0);
      break;

    case UNKNOWN:
      switch (icode)
	{
	case CODE_FOR_zero_extendpsisi2:
	  align = GET_CODE (op0) == REG && REG_POINTER (op0) ?
	    REGNO_POINTER_ALIGN (REGNO (op0)) : -1;
	  if (type0 && align == -1)
	    align = TYPE_ALIGN (type0);
	  break;
	case CODE_FOR_truncsipsi2:
	  return;
	default:
	  abort ();
	}
      break;

#if 0
    case UNOP:
      if (GET_CODE (op0) == REG && REG_POINTER (op0))
	mark_reg_pointer (temp, REGNO_POINTER_ALIGN (REGNO (op0)),
			  REGNO_POINTER_TYPE (REGNO (op0)));
      break;
#endif

    default:
      /* Not a valid pointer operation.  */
      abort ();
    }

  if (!type0)
    return;

  if (indirect)
    align = BITS_PER_WORD;

  while (indirect--)
    type0 = build_pointer_type (type0);

  mark_reg_pointer (target, align, type0);
}
