;;- Machine description for the PDP-10.
;;  Copyright (C) 2001 Free Software Foundation, Inc.
;;  Contributed by Lars Brinkhoff <lars@nocrew.org>, funded by XKL, LLC.

;; This file is part of GNU CC.

;; GNU CC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU CC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU CC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Index

;;   Front Page
;;   Index
;;   Constraints for Immediate Operands
;;   To-do List
;;   Instruction Wish-List
;;   Attributes
;;	length, skip
;;   Unspec Usage
;;	UNSPEC_ADJSP, UNSPEC_ADJBP, UNSPEC_ADDRESS, UNSPEC_FFO,
;;	VUNSPEC_BLT, VUNSPEC_FSC, VUNSPEC_XBLT, VUNSPEC_MOVSLJ, VUNSPEC_MOVST
;;   Constants
;;      SIGNBIT, SP_REGNUM
;;   Data Movement
;;	LDB, (ILDB), DPB, (IDPB),
;;	HRRM, HRLM, HRRZ, HRRE, HLRZ, HLRE, HRRI, HRR, HRLI, HRL, HLRM,
;;	SETZM, SETOM,
;;	MOVE, MOVEI, MOVNI, MOVSI, MOVEM,
;;	HRLOI, HRROI,
;;	EXCH,
;;	(SETZB),
;;	DMOVE, DMOVEM,
;;	(BLT), (XBLT), (MOVSLJ), (MOVST), (CMPS)
;;   Conditional Data Movement
;;	SKIPL, SKIPE, SKIPLE, SKIPGE, SKIPN, SKIPG
;;   Integer Arithmetic
;;	AOS, SOS,
;;	ADD, ADDI, ADDM, (ADDB), DADD,
;;	SUB, SUBI, SUBM, (SUBB), DSUB,
;;	IMUL, IMULI, IMULM, (IMULB), MUL, MULI, MULM, (MULB), DMUL,
;;	IDIV, IDIVI, IDIVM, DIV, DIVI, DIVM, DDIV,
;;	MOVN, MOVNM, MOVNS, MOVNI,
;;	DMOVN, DMOVNM,
;;	MOVM, MOVMM, MOVMS
;;   Integer Conversions
;;   Shifting and Rotating
;;	LSH, LSHC, ASH, ASHC, ROT, ROTC
;;   Logical Operations
;;	AND, ANDI, ANDM, (ANDB), TLZ, ANDCMI,
;;	ANDCA, ANDCAI, ANDCAM, (ANDCAB),
;;	ANDCM, ANDCMI, ANDCMM, (ANDCMB),
;;	XOR, XORI, XORM, (XORB), TLC,
;;	IOR, IORI, IORM, (IORB), TLO, ORCMI,
;;	ANDCB, ANDCBI, ANDCBM, (ANDCBB),
;;	EQV, EQVI, EQVM, (EQVB),
;;	SETCA, SETCAM, (SETCAB),
;;	SETCM, SETCMM, (SETCMB),
;;	ORCA, ORCAI, ORCAM, (ORCAB),
;;	ORCM, ORCMM, (ORCMB),
;;	ORCB, ORCBI, ORCBM, ORCBB
;;   Floating-point Arithmetic
;;	FADR, FADRI, FADRM, (FADRB), DFAD, GFAD,
;;	FSBR, FSBRI, FSBRM, (FSBRB), DFSB, GFSB,
;;	MOVM, MOVMM, MOVMS,
;;	MOVN, MOVNM, MOVNS, DMOVN, DMOVNM,
;;	FMPR, FMPRI, FMPRM, (FMPRB), DFMP, GFMP,
;;	FDVR, FDVRI, FDVRM, (FDVRB), DFDV, GFDV,
;;	FSC, GFSC
;;   Floating-point Conversions
;;	FIX, GFIX, GDFIX, FLTR, GFLTR, DGFLTR, GSNGL, GDBLE
;;   Pointer Arithmetic
;;	IBP, ADJBP
;;   Unconditional Jumps
;;	JFCL, JRST, PUSHJ
;;   Conditional Jumps
;;	TRN, TLN, TDN,
;;	TRO, TLO, TDO,
;;	TRZ, TLZ, TDZ,
;;	TRC, TLC, TDC,
;;	JUMP, SKIP, CAI, CAM,
;;	SOJ, SOS, AOJ, AOS,
;;	JFFO
;;   Function prologue and epilogue
;;	POPJ, ADJSP, PUSH, POP
;;
;; Instructions in parentheses are either commented out or not
;; generated by the compiler.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Constraints for Immediate Operands

;;   'G'	Any valid immediate floating-point constant
;;   'I'	0 .. 262143
;;   'J'	-262143 .. 0
;;   'K'	-131072 .. 131071
;;   'L'	xxxxxx,,000000 (MOVSI, TLN, TLO, TLC)
;;   'N'	xxxxxx,,777777 (HRLOI, TLZ)
;;   'O'	0
;;   'P'	777777,,xxxxxx (HRROI, TRZ, ORCMI, ANDCBI)
;;   'S'	Local symbol.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; To-do List

;; ILDB, IDPB.

;; udiv, umod

;; Add `casesi' or `tablejump' pattern.

;; Add `untyped_call', `untyped_return', patterns.

;; Add `cmpdi' pattern?

;; String instructions.  Avoid them for now.  "None of these is fast.
;; None of these is thoroughly tested.  They all require a lot of
;; registers for setup.  They are uncharacteristic of other PDP-10
;; instructions."

;; Test-modify-skip instructions?

;; Ok to negate double-word integer with DMOVN?

;; ANDCMI really faster than TRZ?

;; Add `lshldi3' pattern - involves adding LSHIFT to the RTL.
;; Implement ashrsi3 in terms of ASH, not LSH?  KCC doesn't.

;; AOBJP, AOBJN?

;; SUBREG of DImode causes bad register allocation:
;; divmodsi4, divmoddi4, expand_builtin_jffo, ffssi

;; Historical: KA10 software double precision floating-point.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Instruction Wish-List

;; ILDBE, LDBE - (increment and) load sign-extended byte

;; LDBI, LDBEI, DPBI - load/deposit byte and increment

;; SUBBP - subtract byte pointers

;; CMPBP - compare byte pointers

;; SEXT r,n - sign-extend n-bit byte in register r

;; clear/set byte

;; test byte >0 =0 <0

;; UDIVI, UDIV, UDDIV - unsigned division

;; UJUMP, USKIP, UCAI, UCAM - compare unsigned values.

;; MIN, MAX, UMIN, UMAX - min/max.

;; SQRT, DSQRT, GSQRT - square root (sqrtM2)

;; FFS - find first set, counting least significant bit first (ffsM2)

;; ?DFIX, DFIXR - Double Floating to Integer (fix_truncdfsi2)
;; ?DFIX, DFIXR - Single Floating to Double Precision Integer (fix_truncsfdi2)
;; DDFIX, DDFIXR - Double Floating to Double Precision Integer (fix_truncdfdi2)

;; ?DFLTR - Double Float and Round (floatsidf2)
;; ?DFLTR - Float Double Precision Integer and Round (floatdisf2)
;; DDFLTR - Double Float Double Precision Integer and Round (floatdidf2)

;; WAIT - wait for interrupt


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Attributes

;; Length of the instruction, in words.
(define_attr "length" "" (const_int 1))

;; The instruction is a skip instruction.
(define_attr "skip" "yes,no" (const_string "no"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Unspec Usage

(define_constants
  [(UNSPEC_ADJSP	0)	; ADJSP operation (PSImode):
				;   Operand 0 is the stack pointer.
				;   Operand 1 is the adjustment.
   (UNSPEC_ADJBP	1)	; ADJBP operation (PSImode):
				;   Operand 0 is the byte pointer.
				;   Operand 1 is the adjustment.
   (UNSPEC_ADDRESS	2)	; Effective-address calculation (PSImode):
				;   Operand 0 is the memory operand.
   (UNSPEC_FFO		3)	; Find-first-one operation in FFO (SImode):
				;   Operand 0 is the register.
   (UNSPEC_FSC		4)	; FSC operation (SFmode, DFmode):
				;   Operand 0 is the floating-point value.
				;   Operand 1 is the scale factor.
   (UNSPEC_SHIFT	5)	; Left-shift operation (SImode, PSImode):
				;   Used to compensate for unnecessary
				;   shifts generated by pointer arithmetic.
				;   Operand 0 is the value.
				;   Operand 1 is the shift count.
   (UNSPEC_SHIFTRT	6)])	; Right-shift operation (SImode, PSImode):
				;   Used to compensate for unnecessary
				;   shifts generated by pointer arithmetic.
				;   Operand 0 is the value.
				;   Operand 1 is the shift count.

(define_constants
  [(VUNSPEC_BLT		0)	; BLT operation (BLKmode):
				;   Operand 0 is the source address.
				;   Operand 1 is the destination address.
				;   Operand 2 is the block length.
   (VUNSPEC_XBLT	1)	; XBLT operation (BLKmode):
				;   operand 0 is the block length
				;   operand 1 is the source address
				;   operand 2 is the destination address
   (VUNSPEC_MOVSLJ	2)	; MOVSLJ operation (BLKmode):
				;   operand 0 is the register block
   (VUNSPEC_MOVST	3)	; MOVST operation (BLKmode):
				;   operand 0 is the register block
   (VUNSPEC_CMPS	4)])	; CMPS operation (BLKmode):
				;   operand 0 is the register block


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Constants

(define_constants
  [(SIGNBIT	34359738368)	; 2^35
   (SP_REGNUM	15)])		; Stack pointer register.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Data Movement

(define_expand "movqi"
  [(set (match_operand:QI 0 "reg_or_mem_operand" "")
        (match_operand:QI 1 "general_operand" ""))]
  ""
  "if (GET_CODE (operands[0]) == REG && GET_CODE (operands[1]) == REG)
     {
       rtx op0 = copy_rtx (operands[0]);
       rtx op1 = copy_rtx (operands[1]);
       PUT_MODE (op0, SImode);
       PUT_MODE (op1, SImode);
       emit_move_insn (op0, op1);
       DONE;
     }
   if (!no_new_pseudos)
     {
       if (GET_CODE (operands[1]) == CONST_INT)
   	 {
   	   rtx reg = gen_reg_rtx (SImode);
   
   	   emit_insn (gen_movsi (reg, operands[1]));
   	   operands[1] = gen_rtx_SUBREG (QImode, reg, 0);
   	 }
       if (GET_CODE (operands[0]) == MEM)
   	 operands[1] = force_reg (QImode, operands[1]);
     }")

(define_insn "*LDBzqi"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
        (zero_extend:SI (match_operand:QI 1 "memory_operand" "m,>")))]
  ""
  "@
   ldb %0,%1\\t%; movqi
   ildb %0,%1\\t%; movqi")

(define_insn "*LDBIzqi"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (zero_extend:SI (match_operand:QI 1 "memory_operand" "<")))]
  "TARGET_XKL2"
  "extend %0,[ldbi %1]\\t%; movqi")

(define_insn "*LDBEqi"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
        (sign_extend:SI (match_operand:QI 1 "memory_operand" "m,>,<")))]
  "TARGET_XKL2"
  "@
   extend %0,[ldbe %1]\\t%; movqi
   extend %0,[ildbe %1]\\t%; movqi
   extend %0,[ldbei %1]\\t%; movqi")

(define_insn "*LDBqi"
  [(set (match_operand:QI 0 "register_operand" "=r,r")
        (match_operand:QI 1 "memory_operand" "m,>"))]
  ""
  "@
   ldb %0,%1\\t%; movqi
   ildb %0,%1\\t%; movqi")

(define_insn "*LDBIqi"
  [(set (match_operand:QI 0 "register_operand" "=r")
        (match_operand:QI 1 "memory_operand" "<"))]
  "TARGET_XKL2"
  "extend %0,[ldbi %1]\\t%; movqi")

(define_insn "*DPBqi"
  [(set (match_operand:QI 0 "memory_operand" "=m,>")
	(match_operand:QI 1 "register_operand" "r,r"))]
  ""
  "@
   dpb %1,%0\\t%; movqi
   idpb %1,%0\\t%; movqi")

(define_insn "*DPBIqi"
  [(set (match_operand:QI 0 "memory_operand" "=<")
	(match_operand:QI 1 "register_operand" "r"))]
  "TARGET_XKL2"
  "dpbi %1,%0\\t%; movqi")

(define_insn "*ILDBqi"
  [(set (match_operand:PSI 1 "reg_or_mem_operand" "+rm")
	(unspec:PSI [(match_operand:PSI 2 "reg_or_mem_operand" "0")
		     (const_int 1)]
		    UNSPEC_ADJBP))
   (set (match_operand:QI 0 "register_operand" "=r")
	(mem:QI (unspec:PSI [(match_dup 1) (const_int 1)] UNSPEC_ADJBP)))]
  ""
  "ildb %0,%1\\t%; movqi")

(define_insn "*IDPBqi"
  [(set (match_operand:PSI 1 "reg_or_mem_operand" "+rm")
	(unspec:PSI [(match_operand:PSI 2 "reg_or_mem_operand" "0")
		     (const_int 1)]
		    UNSPEC_ADJBP))
   (set (mem:QI (unspec:PSI [(match_dup 1) (const_int 1)] UNSPEC_ADJBP))
	(match_operand:QI 0 "register_operand" "r"))]
  ""
  "idpb %0,%1\\t%; movqi")

(define_expand "movhi"
  [(set (match_operand:HI 0 "reg_or_mem_operand" "")
        (match_operand:HI 1 "general_operand" ""))]
  ""
  "if (!no_new_pseudos)
     {
       if (GET_CODE (operands[1]) == CONST_INT)
   	 {
   	   rtx reg = gen_reg_rtx (SImode);
   
   	   emit_insn (gen_movsi (reg, operands[1]));
   	   operands[1] = gen_rtx_SUBREG (HImode, reg, 0);
   	 }
       if (GET_CODE (operands[0]) == MEM)
   	 operands[1] = force_reg (HImode, operands[1]);
     }")

(define_insn "*loadhi"
  [(set (match_operand:HI 0 "register_operand" "=r")
        (match_operand:HI 1 "memory_operand" "m>"))]
  ""
  "*return pdp10_output_movhi (operands, 0);")

(define_insn "*zloadhi"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (zero_extend:SI (match_operand:HI 1 "memory_operand" "m>")))]
  ""
  "*return pdp10_output_movhi (operands, 0);")

(define_insn "*sloadhi"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (sign_extend:SI (match_operand:HI 1 "pdp10_aligned_operand" "m")))]
  ""
  "*return pdp10_output_movhi (operands, 1);")

(define_insn "storehi"
  [(set (match_operand:HI 0 "memory_operand" "=m>")
	(match_operand:HI 1 "register_operand" "r"))]
  ""
  "*return pdp10_output_movhi (operands, 0);")

(define_insn "*HLRE_shift"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extend:SI
	 (subreg:HI (ashiftrt:SI
		     (match_operand:SI 1 "register_operand" "r")
		     (neg:SI (const_int -18)))
		    0)))]
  ""
  "hlre %0,%1")

(define_insn "*HLRZ_shift"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extend:SI
	 (subreg:HI (ashiftrt:SI
		     (match_operand:SI 1 "register_operand" "r")
		     (neg:SI (const_int -18)))
		    0)))]
  ""
  "hlrz %0,%1")

(define_insn "*HRR"
  [(set (zero_extract:SI (match_operand:SI 0 "reg_or_mem_operand" "+r,r,m")
			 (const_int 18)
			 (const_int 18))
	(match_operand:SI 1 "general_operand" "I,r,r"))]
  ""
  "@
   hrri %0,%1
   hrr %0,%1
   hrrm %1,%W0")

(define_insn "*HRRZ"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extract:SI (match_operand:SI 1 "reg_or_mem_operand" "rm")
			 (const_int 18)
			 (const_int 18)))]
  ""
  "hrrz %0,%W1")

(define_insn "*HRRE"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extract:SI (match_operand:SI 1 "reg_or_mem_operand" "rm")
			 (const_int 18)
			 (const_int 18)))]
  ""
  "hrre %0,%W1")

(define_insn "HRL"
  [(set (zero_extract:SI (match_operand:SI 0 "reg_or_mem_operand" "+r,r,m")
			 (const_int 18)
			 (const_int 0))
	(match_operand:SI 1 "general_operand" "I,r,r"))]
  ""
  "@
   hrli %0,%1
   hrl %0,%1
   hlrm %1,%W0")

(define_insn "*HLRZ_extract"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extract:SI (match_operand:SI 1 "reg_or_mem_operand" "rm")
			 (const_int 18)
			 (const_int 0)))]
  ""
  "hlrz %0,%W1")

(define_insn "*HLRE_extract"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extract:SI (match_operand:SI 1 "reg_or_mem_operand" "rm")
			 (const_int 18)
			 (const_int 0)))]
  ""
  "hlre %0,%W1")

(define_expand "extv"
  [(set (match_operand:SI 0 "register_operand" "")
	(sign_extract:SI (match_operand:SI 1 "reg_or_mem_operand" "")
			 (match_operand:SI 2 "const_int_operand" "")
			 (match_operand:SI 3 "const_int_operand" "")))]
  ""
  "if (!TARGET_XKL2)
     {
       pdp10_expand_extv (operands);
       DONE;
     }")

(define_insn "*LDBE"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extract:SI (match_operand:SI 1 "reg_or_mem_operand" "rm")
			 (match_operand:SI 2 "const_int_operand" "")
			 (match_operand:SI 3 "const_int_operand" "")))]
  "TARGET_XKL2"
  "*return pdp10_output_extv (operands);")

(define_insn "extzv"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extract:SI (match_operand:SI 1 "reg_or_mem_operand" "rm")
			 (match_operand:SI 2 "const_int_operand" "")
			 (match_operand:SI 3 "const_int_operand" "")))]
  ""
  "*return pdp10_output_extzv (operands);")

(define_insn "insv"
  [(set (zero_extract:SI (match_operand:SI 0 "reg_or_mem_operand" "+rm")
			 (match_operand:SI 1 "const_int_operand" "")
			 (match_operand:SI 2 "const_int_operand" ""))
	(match_operand:SI 3 "register_operand" "r"))]
  ""
  "*return pdp10_output_insv (operands);")

(define_insn "*SETZM"
  [(set (match_operand:SI 0 "memory_operand" "=m")
	(const_int 0))]
  ""
  "setzm %0")

(define_insn "*SETOM"
  [(set (match_operand:SI 0 "memory_operand" "=m")
	(const_int -1))]
  ""
  "setom %0")

(define_insn "*movsi_inline"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (match_operand:SI 1 "pdp10_inline_operand" "i"))]
  "0 && TARGET_KL10up && !optimize_size"
  "skipa %0,.+1\\t%; movsi\; %1"
  [(set_attr "length" "2")])

(define_insn "movsi"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,r,r,r,r,m")
        (match_operand:SI 1 "general_operand" "rm,IS,L,N,P,i,r"))]
  ""
  "@
   move %0,%1
   movei %0,%1
   movsi %0,%S1\\t%; movsi
   hrloi %0,%T1\\t%; movsi
   hrroi %0,%C1\\t%; movsi
   move %0,[%1]
   movem %1,%0")

(define_insn "move_two_halves"
  [(set (zero_extract:PSI (match_operand:PSI 0 "register_operand" "+r")
			  (const_int 18)
			  (const_int 0))
	(match_operand:PSI 1 "immediate_operand" "p"))
   (set (zero_extract:PSI (match_dup 0) (const_int 18) (const_int 18))
	(match_operand:PSI 2 "immediate_operand" "p"))]
  ""
  "move %0,[%1,,%2]")

(define_insn "*EXCH"
  [(set (match_operand:SI 0 "register_operand" "+r")
        (match_operand:SI 1 "reg_or_mem_operand" "+rm"))
   (set (match_dup 1) (match_dup 0))]
  ""
  "exch %0,%1")

(define_insn "*SETZB"
  [(set (match_operand:SI 0 "register_operand" "=r") (const_int 0))
   (set (match_operand:SI 1 "reg_or_mem_operand" "=rm") (const_int 0))]
  ""
  "setzb %0,%1")

(define_insn "movdi"
  [(set (match_operand:DI 0 "reg_or_mem_operand" "=r,r,m")
        (match_operand:DI 1 "general_operand" "rm,i,r"))]
  "TARGET_KI10up"
  "@
   dmove %0,%1
   dmove %0,[%D1]
   dmovem %1,%0")

(define_insn "movsf"
  [(set (match_operand:SF 0 "reg_or_mem_operand" "=r,r,r,m")
        (match_operand:SF 1 "general_operand" "rm,G,F,r"))]
  ""
  "@
   move %0,%1\\t%; movsf
   movsi %0,%G1\\t%; movsf
   move %0,[%1]\\t%; movsf
   movem %1,%0\\t%; movsf")

(define_insn "*movdf_immediate"
  [(set (match_operand:DF 0 "register_operand" "=r")
        (match_operand:DF 1 "immediate_operand" "G"))]
  ""
  "movsi %0,%G1\\t%; movdf\;movei %Z0,0\\t%; movdf"
  [(set_attr "length" "2")])

(define_insn "movdf"
  [(set (match_operand:DF 0 "reg_or_mem_operand" "=r,r,m")
        (match_operand:DF 1 "general_operand" "rm,F,r"))]
  "TARGET_KI10up"
  "@
   dmove %0,%1\\t%; movdf
   dmove %0,[%1]\\t%; movdf
   dmovem %1,%0\\t%; movdf")

(define_expand "movstrsi"
  [(match_operand:BLK 0 "general_operand" "")
   (match_operand:BLK 1 "general_operand" "")
   (match_operand:SI 2 "general_operand" "")
   (match_operand:SI 3 "const_int_operand" "")]
  ""
  "if (pdp10_expand_movstrsi (operands))
     DONE;
   FAIL;")

(define_expand "clrstrsi"
  [(match_operand:BLK 0 "general_operand" "")
   (match_operand:SI 1 "general_operand" "")
   (match_operand:SI 2 "const_int_operand" "")]
  ""
  "if (pdp10_expand_clrstrsi (operands))
     DONE;
   FAIL;")

; (define_expand "cmpstrsi"
;   [(set (match_operand:SI 0 "register_operand" "")
; 	(compare:SI (match_operand:BLK 1 "general_operand" "")
; 		    (match_operand:BLK 2 "general_operand" "")))
;    (use (match_operand:SI 3 "general_operand" ""))
;    (use (match_operand:SI 4 "const_int_operand" ""))]
;   "TARGET_KL10up && TARGET_STRING"
;   "if (pdp10_expand_cmpstrsi (operands))
;      DONE;
;    FAIL;")

(define_insn "BLT"
  [(unspec_volatile:BLK [(match_operand:PSI 0 "register_operand" "+r")]
			VUNSPEC_BLT)
   (use (match_operand:BLK 1 "memory_operand" "m"))]
  ""
  "blt %0,%1")

(define_insn "XBLT"
  [(unspec_volatile:BLK [(match_operand:TI 0 "register_operand" "+r")]
			 VUNSPEC_XBLT)]
  "TARGET_KL10up"
  "extend %0,[xblt]")

(define_insn "MOVSLJ"
  [(unspec_volatile:BLK [(match_operand:SI 0 "register_operand" "+r")]
			 VUNSPEC_MOVSLJ)]
  "TARGET_KL10up && TARGET_STRING"
  "extend %0,[movslj]")

(define_insn "MOVST"
  [(unspec_volatile:BLK [(match_operand:SI 0 "register_operand" "+r")]
			 VUNSPEC_MOVST)]
  "TARGET_KL10up && TARGET_STRING"
  "extend %0,[movst]")

;(define_insn "CMPS"
;  [(unspec_volatile:BLK [(match_operand:SI 0 "register_operand" "+r")]
;			 VUNSPEC_MOVST)]
;  "TARGET_KL10up && TARGET_STRING"
;  "extend %0,[movst]")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Conditional Data Movement

;; (define_expand "cstoresi4" ...)
;; (define_expand "cstoresf4" ...)

(define_expand "sne"
  [(set (match_operand:SI 0 "register_operand" "")
	(if_then_else:SI (ne (match_dup 1) (const_int 0))
			 (const_int 1)
			 (const_int 0)))]
  ""
  "if (pdp10_compare_op1 != const0_rtx)
     FAIL;
   operands[1] = pdp10_compare_op0;")

(define_insn "*sne"
  [(set (match_operand:SI 0 "register_operand" "=x")
	(if_then_else:SI (ne (match_operand:SI 1 "reg_or_mem_operand" "rm")
			     (const_int 0))
			 (const_int 1)
			 (const_int 0)))]
  ""
  "*if (GET_CODE (operands[1]) == REG
        && REGNO (operands[1]) == REGNO (operands[0]))
      return \"skipe %1\; movei %0,1\\t%; sne\";
    else
      return \"skipe %0,%1\; movei %0,1\\t%; sne\";"
  [(set_attr "length" "2")])

;; (define_expand "cmovsi6" ...)
;; (define_expand "cmovsf6" ...)

(define_expand "movsicc"
  [(set (match_operand:SI 0 "register_operand" "")
	(if_then_else:SI (match_operator 1 "comparison_operator"
			  [(match_dup 4) (const_int 0)])
			 (match_operand:SI 2 "reg_or_mem_operand" "")
			 (match_operand:SI 3 "reg_or_mem_operand" "")))]
  ""
  "if (pdp10_compare_op1 != const0_rtx)
     FAIL;
   if (!rtx_equal_p (operands[2], pdp10_compare_op0))
     FAIL;
   if (!pdp10_comparison_operator (operands[1], VOIDmode))
     FAIL;
     /*{
       operands[4] = gen_reg_rtx (SImode);
       emit_move_insn (operands[4], pdp10_compare_op0);
       emit_insn (gen_xorsi3 (operands[4], operands[4], GEN_INT (SIGNBIT)));
     }*/
   else
     operands[4] = pdp10_compare_op0;")

(define_insn "*movsicc"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(if_then_else:SI (match_operator 1 "pdp10_comparison_operator"
			  [(match_operand:SI 2 "reg_or_mem_operand" "rm")
			   (const_int 0)])
			 (match_operand:SI 3 "reg_or_mem_operand" "2")
			 (match_operand:SI 4 "reg_or_mem_operand" "rm")))]
  "rtx_equal_p (operands[2], operands[3])"
  "*if (GET_CODE (operands[2]) == REG
	&& REGNO (operands[2]) == REGNO (operands[0]))
      return \"skip%1 %2\; move %0,%4\\t%; movsicc\";
    else
      return \"skip%1 %0,%2\; move %0,%4\\t%; movsicc\";"
  [(set_attr "length" "2")])

(define_expand "movsfcc"
  [(set (match_operand:SF 0 "register_operand" "")
	(if_then_else:SF (match_operator 1 "comparison_operator"
			  [(match_dup 4) (const_int 0)])
			 (match_operand:SF 2 "reg_or_mem_operand" "")
			 (match_operand:SF 3 "reg_or_mem_operand" "")))]
  ""
  "if (pdp10_compare_op1 != const0_rtx)
     FAIL;
   if (!rtx_equal_p (operands[2], pdp10_compare_op0))
     FAIL;
   if (!pdp10_comparison_operator (operands[1], VOIDmode))
     FAIL;
     /*{
       operands[4] = gen_reg_rtx (SFmode);
       emit_move_insn (operands[4], pdp10_compare_op0);
       emit_insn (gen_xorsi3 (operands[4], operands[4], GEN_INT (SIGNBIT)));
     }*/
   else
     operands[4] = pdp10_compare_op0;")

(define_insn "*movsfcc"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(if_then_else:SF (match_operator 1 "pdp10_comparison_operator"
			  [(match_operand:SF 2 "reg_or_mem_operand" "rm")
			   (const_int 0)])
			 (match_operand:SF 3 "reg_or_mem_operand" "2")
			 (match_operand:SF 4 "reg_or_mem_operand" "rm")))]
  "rtx_equal_p (operands[2], operands[3])"
  "*if (GET_CODE (operands[2]) == REG
	&& REGNO (operands[2]) == REGNO (operands[0]))
      return \"skip%1 %2\; move %0,%4\\t%; movsfcc\";
    else
      return \"skip%1 %0,%2\; move %0,%4\\t%; movsfcc\";"
  [(set_attr "length" "2")])


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Integer Arithmetic

(define_insn "*AOS"
  [(set (match_operand:SI 0 "memory_operand" "=m")
	(plus:SI (match_operand:SI 1 "memory_operand" "0")
		 (const_int 1)))]
  ""
  "aos %0")

(define_insn "*SOS"
  [(set (match_operand:SI 0 "memory_operand" "=m")
	(plus:SI (match_operand:SI 1 "memory_operand" "0")
		 (const_int -1)))]
  ""
  "sos %0")

(define_insn "addsi3"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,r,r,m")
	(plus:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0,0,0,0")
		 (match_operand:SI 2 "general_operand" "rm,I,J,i,r")))]
  ""
  "@
   add %0,%2
   addi %0,%2
   subi %0,%N2
   add %0,[%2]
   addm %2,%0")

(define_insn "*ADDB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI (match_operand:SI 1 "register_operand" "%0")
		 (match_operand:SI 2 "reg_or_mem_operand" "+rm")))
   (set (match_dup 2)
	(plus:SI (match_dup 1) (match_dup 2)))]
  ""
  "addb %0,%2")

(define_insn "adddi3"
  [(set (match_operand:DI 0 "register_operand" "=r,r")
	(plus:DI (match_operand:DI 1 "register_operand" "%0,0")
		 (match_operand:DI 2 "reg_or_mem_operand" "rm,i")))]
  "TARGET_KL10up"
  "@
   dadd %0,%2
   dadd %0,[%2]")

(define_expand "subsi3"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "")
	(minus:SI (match_operand:SI 1 "reg_or_mem_operand" "")
		  (match_operand:SI 2 "general_operand" "")))]
  ""
  "if (pdp10_expand_subsi3 (operands))
     DONE;")

(define_insn "subsi3_internal"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,m")
	(minus:SI (match_operand:SI 1 "reg_or_mem_operand" "0,r")
		  (match_operand:SI 2 "general_operand" "rm,0")))]
  ""
  "@
   sub %0,%2
   subm %1,%0")

(define_insn "*SUBB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(minus:SI (match_operand:SI 1 "register_operand" "0")
		  (match_operand:SI 2 "reg_or_mem_operand" "+rm")))
   (set (match_dup 2)
	(minus:SI (match_dup 1) (match_dup 2)))]
  ""
  "subb %0,%2")

(define_insn "subdi3"
  [(set (match_operand:DI 0 "register_operand" "=r,r")
	(minus:DI (match_operand:DI 1 "register_operand" "%0,0")
		  (match_operand:DI 2 "reg_or_mem_operand" "rm,i")))]
  "TARGET_KL10up"
  "@
   dsub %0,%2
   dsub %0,[%2]")

(define_insn "mulsi3"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,r,m")
	(mult:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0,0,0")
		 (match_operand:SI 2 "general_operand" "rm,I,i,r")))]
  ""
  "@
   imul %0,%2
   imuli %0,%2
   imul %0,[%2]
   imulm %2,%0")

(define_insn "*IMULB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(mult:SI (match_operand:SI 1 "register_operand" "%0")
		 (match_operand:SI 2 "reg_or_mem_operand" "+rm")))
   (set (match_dup 2)
	(mult:SI (match_dup 1) (match_dup 2)))]
  ""
  "imulb %0,%2")

(define_insn "mulsidi3"
  [(set (match_operand:DI 0 "reg_or_mem_operand" "=r,r,r,m")
	(mult:DI (sign_extend:DI
		  (match_operand:SI 1 "reg_or_mem_operand" "%0,0,0,0"))
		 (sign_extend:DI
		  (match_operand:SI 2 "general_operand" "rm,I,i,r"))))]
  ""
  "@
   mul %0,%2
   muli %0,%2
   mul %0,[%2]
   mulm %2,%0")

(define_insn "*MULB"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(mult:DI (sign_extend:DI
		  (match_operand:SI 1 "register_operand" "%0"))
		 (sign_extend:DI
		  (match_operand:SI 2 "reg_or_mem_operand" "+rm"))))
   (set (match_dup 2)
	(mult:SI (sign_extend:DI (match_dup 1))
		 (sign_extend:DI (match_dup 2))))]
  ""
  "mulb %0,%2")

(define_insn "mulditi3"
  [(set (match_operand:TI 0 "register_operand" "=r,r")
	(mult:TI
	 (sign_extend:TI
	  (subreg:DI (match_operand:TI 1 "register_operand" "%0,0") 0))
	 (sign_extend:TI
	  (match_operand:DI 2 "reg_or_mem_operand" "rm,i"))))]
  "TARGET_KL10up"
  "@
   dmul %0,%2
   dmul %0,[%2]")

(define_insn "IDIV"
  [(set (subreg:SI (match_operand:DI 0 "register_operand" "=r,r,r") 0)
        (div:SI (subreg:SI (match_operand:DI 1 "register_operand" "0,0,0") 0)
                (match_operand:SI 2 "general_operand" "rm,I,i")))
   (set (subreg:SI (match_dup 0) 1)
        (mod:SI (match_dup 1) (match_dup 2)))]
  ""
  "@
   idiv %0,%2
   idivi %0,%2
   idiv %0,[%2]")

(define_insn "IDIVM"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=mr")
        (div:SI (match_operand:SI 1 "register_operand" "r")
                (match_operand:SI 2 "reg_or_mem_operand" "0")))]
  ""
  "idivm %1,%0")

(define_expand "divmodsi4"
  [(parallel [(set (match_operand:SI 0 "reg_or_mem_operand" "")
		   (div:SI (match_operand:SI 1 "register_operand" "")
			   (match_operand:SI 2 "general_operand" "")))
	      (set (match_operand:SI 3 "register_operand" "")
		   (mod:SI (match_dup 1) (match_dup 2)))])]
  ""
  "{
     rtx temp = gen_reg_rtx (DImode);
     emit_move_insn (gen_rtx_SUBREG (SImode, temp, 0), operands[1]);
     emit_insn (gen_IDIV (temp, temp, operands[2]));
     emit_move_insn (operands[0], gen_rtx_SUBREG (SImode, temp, 0));
     emit_move_insn (operands[3], gen_rtx_SUBREG (SImode, temp, 1));
     DONE;
   }")

(define_insn "*DIV"
  [(set (subreg:SI (match_operand:DI 0 "register_operand" "=r,r,r") 0)
        (truncate:SI
	 (div:DI
	  (match_operand:DI 1 "register_operand" "0,0,0")
	  (sign_extend:DI (match_operand:SI 2 "general_operand" "rm,I,i")))))
   (set (subreg:SI (match_dup 0) 1)
        (truncate:SI (mod:DI (match_dup 1) (sign_extend:DI (match_dup 2)))))]
  ""
  "@
   div %0,%2
   divi %0,%2
   div %0,[%2]")

(define_insn "*DIVM"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=mr")
	(truncate:SI
	 (div:DI
	  (match_operand:DI 1 "register_operand" "r")
	  (sign_extend:DI (match_operand:SI 2 "memory_operand" "0")))))]
  ""
  "divm %1,%0")

(define_insn "DDIV"
  [(set (subreg:DI (match_operand:TI 0 "register_operand" "=r,r") 0)
        (truncate:DI
	 (div:TI
	  (match_operand:TI 1 "register_operand" "0,0")
	  (sign_extend:TI (match_operand:DI 2 "general_operand" "rm,i")))))
   (set (subreg:DI (match_dup 0) 1)
        (truncate:DI (mod:TI (match_dup 1) (sign_extend:TI (match_dup 2)))))]
  ""
  "@
   ddiv %0,%2
   ddiv %0,[%2]")

(define_expand "divmoddi4"
  [(parallel [(set (match_operand:DI 0 "reg_or_mem_operand" "")
		   (div:DI (match_operand:DI 1 "register_operand" "")
			   (match_operand:DI 2 "general_operand" "")))
	      (set (match_operand:DI 3 "register_operand" "")
		   (mod:DI (match_dup 1) (match_dup 2)))])]
  ""
  "{
     rtx temp = gen_reg_rtx (TImode);
     emit_move_insn (gen_rtx_SUBREG (DImode, temp, 0), operands[1]);
     emit_insn (gen_DDIV (temp, temp, operands[2]));
     emit_move_insn (operands[0], gen_rtx_SUBREG (DImode, temp, 0));
     emit_move_insn (operands[3], gen_rtx_SUBREG (DImode, temp, 1));
     DONE;
   }")

(define_insn "negsi2"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,m,r,m")
	(neg:SI (match_operand:SI 1 "general_operand" "rm,r,I,0")))]
  ""
  "@
   movn %0,%1
   movnm %1,%0
   movni %0,%N1
   movns %0")

(define_insn "*MOVNS"
  [(set (match_operand:SI 0 "register_operand" "=x")
	(neg:SI (match_operand:SI 1 "memory_operand" "+m")))
   (set (match_dup 1)
	(neg:SI (match_dup 1)))]
  ""
  "movns %0,%1")

(define_insn "negdi2"
  [(set (match_operand:DI 0 "reg_or_mem_operand" "=r,m")
	(neg:DI (match_operand:DI 1 "reg_or_mem_operand" "rm,r")))]
  "TARGET_KI10up"
  "@
   dmovn %0,%1
   dmovnm %1,%0")

(define_insn "abssi2"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,m,m")
	(abs:SI (match_operand:SI 1 "reg_or_mem_operand" "rm,r,0")))]
  ""
  "@
   movm %0,%1
   movmm %1,%0
   movms %0")

(define_insn "*MOVMS"
  [(set (match_operand:SI 0 "register_operand" "=x")
	(abs:SI (match_operand:SI 1 "memory_operand" "+m")))
   (set (match_dup 1)
	(abs:SI (match_dup 1)))]
  ""
  "movms %0,%1")

;; The JFFO instruction searches from the most significant bit while
;; ffs searches from the least significant bit.  The bit index and
;; treatment of zero also differ.
(define_expand "ffssi2"
  [(set (match_operand:SI 0 "register_operand" "")
	(ffs:SI (match_operand:SI 1 "register_operand" "")))]
  "TARGET_KA10up"
  "if (!TARGET_XKL2)
     {
       rtx t1 = gen_reg_rtx (SImode);
       rtx t2 = gen_reg_rtx (SImode);
       rtx t3 = gen_reg_rtx (DImode);
       rtx t4 = gen_reg_rtx (SImode);
       rtx label = gen_label_rtx ();

       emit_insn (gen_negsi2 (t1, operands[1]));
       emit_insn (gen_andsi3 (t2, operands[1], t1));
       emit_move_insn (gen_rtx_SUBREG (SImode, t3, 0), t2);
       emit_jump_insn (gen_JFFO (t3, gen_rtx_LABEL_REF (PSImode, label)));
       emit_move_insn (gen_rtx_SUBREG (SImode, t3, 1),
		       GEN_INT (36));
       emit_label (label);
       emit_insn (gen_subsi3 (t4,
			      gen_rtx_SUBREG (SImode, t3, 1),
			      GEN_INT (36)));
       emit_insn (gen_negsi2 (operands[0], t4));
       DONE;
     }")

(define_insn "*ffssi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ffs:SI (match_operand:SI 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2"
  "ffs %0,%1")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Integer Conversions

(define_insn "zero_extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (zero_extend:SI (match_operand:QI 1 "register_operand" "0")))]
  ""
  "andi %0,777\\t%; zero_extendqisi2")

(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (zero_extend:SI (match_operand:HI 1 "register_operand" "r")))]
  ""
  "hrrz %0,%1\\t%; zero_extendhisi2")

(define_insn "*extendqisi2_reg"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (sign_extend:SI
	 (subreg:QI (match_operand:SI 1 "register_operand" "r") 0)))]
  ""
  "*if (REGNO (operands[0]) == REGNO (operands[1]))
      return \"\";
    else
      return \"move %0,%1\\t%; extendqisi2_reg\";")

(define_expand "extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "")
	(sign_extend:SI (match_operand:QI 1 "register_operand" "")))]
  ""
  "if (!TARGET_XKL2)
     {
       emit_insn (gen_ashlsi3 (operands[0], operands[1], GEN_INT (27)));
       emit_insn (gen_ASH_right (operands[0], operands[0], GEN_INT (-27)));
       DONE;
     }")

(define_insn "*extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extend:SI (match_operand:QI 1 "register_operand" "0")))]
  "TARGET_XKL2"
  "extend %0,[sext 9]")

(define_insn "extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extend:SI (match_operand:HI 1 "register_operand" "r")))]
   ""
   "hrre %0,%1\\t%; extendhisi2")

(define_insn "truncsiqi2"
  [(set (match_operand:QI 0 "register_operand" "=r")
	(truncate:QI (match_operand:SI 1 "register_operand" "0")))]
  ""
  "andi %0,777\\t%; truncsiqi2")

(define_insn "truncsihi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(truncate:HI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "hrr %0,%1\\t%; truncsihi2")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Shifting and Rotating

;(define_insn "lshlsi3"
;  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
;	(lshift:SI (match_operand:SI 1 "register_operand" "0,0,0")
;		   (match_operand:SI 2 "general_operand" "x,K,m")))]
;  ""
;  "@
;   lsh %0,(%2)
;   lsh %0,%2
;   lsh %0,@%2")
;
;(define_insn "*LSH_left_plus"
;  [(set (match_operand:SI 0 "register_operand" "=r")
;	(lshift:SI (match_operand:SI 1 "register_operand" "0")
;		   (plus:SI (match_operand:SI 2 "register_operand" "x")
;			    (match_operand:SI 3 "immediate_operand" "K"))))]
;  ""
;  "lsh %0,%3(%2)")

(define_expand "lshrsi3"
  [(set (match_dup 3) (neg:SI (match_operand:SI 2 "general_operand" "")))
   (set (match_operand:SI 0 "register_operand" "")
	(lshiftrt:SI (match_operand:SI 1 "register_operand" "")
		     (neg:SI (match_dup 3))))]
  ""
  "operands[3] = gen_reg_rtx (SImode);")

(define_insn "*LSH_right"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(lshiftrt:SI (match_operand:SI 1 "register_operand" "0,0,0")
		     (neg:SI (match_operand:SI 2 "general_operand" "x,K,m"))))]
  ""
  "@
   lsh %0,(%2)
   lsh %0,%2
   lsh %0,@%2")

(define_insn "*LSH_right_plus"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lshiftrt:SI
	 (match_operand:SI 1 "register_operand" "0")
	 (neg:SI (plus:SI (match_operand:SI 2 "register_operand" "x")
			  (match_operand:SI 3 "const_int_operand" "K")))))]
  ""
  "lsh %0,%3(%2)")

;(define_insn "lshldi3"
;  [(set (match_operand:DI 0 "register_operand" "=r,r,r")
;	(lshift:DI (match_operand:DI 1 "register_operand" "0,0,0")
;		   (match_operand:SI 2 "general_operand" "x,K,m")))]
;  ""
;  "@
;   lshc %0,(%2)
;   lshc %0,%2
;   lshc %0,@%2")
;
;(define_insn "*LSHC"
;  [(set (match_operand:DI 0 "register_operand" "=r")
;	(lshift:DI (match_operand:DI 1 "register_operand" "0")
;		   (plus:SI (match_operand:SI 2 "register_operand" "x")
;			    (match_operand:SI 3 "immediate_operand" "K"))))]
;  ""
;  "lshc %0,%3(%2)")

(define_expand "lshrdi3"
  [(set (match_dup 3) (neg:SI (match_operand:SI 2 "general_operand" "")))
   (set (match_operand:DI 0 "register_operand" "")
	(lshiftrt:DI (match_operand:DI 1 "register_operand" "")
		     (neg:SI (match_dup 3))))]
  ""
  "operands[3] = gen_reg_rtx (SImode);")

(define_insn "*LSHC_right"
  [(set (match_operand:DI 0 "register_operand" "=r,r,r")
	(lshiftrt:DI (match_operand:DI 1 "register_operand" "0,0,0")
		     (neg:SI (match_operand:SI 2 "general_operand" "x,K,m"))))]
  ""
  "@
   lshc %0,(%2)
   lshc %0,%2
   lshc %0,@%2")

(define_insn "*LSHC_right_plus"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(lshiftrt:DI
	 (match_operand:DI 1 "register_operand" "0")
	 (neg:SI (plus:SI (match_operand:SI 2 "register_operand" "x")
			  (match_operand:SI 3 "immediate_operand" "K")))))]
  ""
  "lshc %0,%3(%2)")

(define_insn "ashlsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(ashift:SI (match_operand:SI 1 "register_operand" "0,0,0")
		   (match_operand:SI 2 "general_operand" "x,K,m")))]
  ""
  "@
   lsh %0,(%2)
   lsh %0,%2
   lsh %0,@%2")

(define_insn "*LSH_left_plus"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashift:SI (match_operand:SI 1 "register_operand" "0")
		   (plus:SI (match_operand:SI 2 "register_operand" "x")
			    (match_operand:SI 3 "immediate_operand" "K"))))]
  ""
  "lsh %0,%3(%2)")

(define_expand "ashrsi3"
  [(set (match_dup 3) (neg:SI (match_operand:SI 2 "general_operand" "")))
   (set (match_operand:SI 0 "register_operand" "")
	(ashiftrt:SI (match_operand:SI 1 "register_operand" "")
		     (neg:SI (match_dup 3))))]
  ""
  "operands[3] = gen_reg_rtx (SImode);")

(define_insn "ASH_right"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(ashiftrt:SI (match_operand:SI 1 "register_operand" "0,0,0")
		     (neg:SI (match_operand:SI 2 "general_operand" "x,K,m"))))]
  ""
  "@
   ash %0,(%2)
   ash %0,%2
   ash %0,@%2")

(define_insn "*ASH_right_plus"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI
	 (match_operand:SI 1 "register_operand" "0")
	 (neg:SI (plus:SI (match_operand:SI 2 "register_operand" "x")
			  (match_operand:SI 3 "immediate_operand" "K")))))]
  ""
  "ash %0,%3(%2)")

(define_insn "ashldi3"
  [(set (match_operand:DI 0 "register_operand" "=r,r,r")
	(ashift:DI (match_operand:DI 1 "register_operand" "0,0,0")
		   (match_operand:SI 2 "general_operand" "x,K,m")))]
  ""
  "@
   ashc %0,(%2)
   ashc %0,%2
   ashc %0,@%2")

(define_insn "*ASHC_left_plus"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(ashift:DI (match_operand:DI 1 "register_operand" "0")
		   (plus:SI (match_operand:SI 2 "register_operand" "x")
			    (match_operand:SI 3 "immediate_operand" "K"))))]
  ""
  "ashc %0,%3(%2)")

(define_expand "ashrdi3"
  [(set (match_dup 3) (neg:SI (match_operand:SI 2 "general_operand" "")))
   (set (match_operand:DI 0 "register_operand" "")
	(ashiftrt:DI (match_operand:DI 1 "register_operand" "")
		     (neg:SI (match_dup 3))))]
  ""
  "operands[3] = gen_reg_rtx (SImode);")

(define_insn "*ASHC_right"
  [(set (match_operand:DI 0 "register_operand" "=r,r,r")
	(ashiftrt:DI (match_operand:DI 1 "register_operand" "0,0,0")
		     (neg:SI (match_operand:SI 2 "general_operand" "x,K,m"))))]
  ""
  "@
   ashc %0,(%2)
   ashc %0,%2
   ashc %0,@%2")

(define_insn "*ASHC_right_plus"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(ashiftrt:DI
	 (match_operand:DI 1 "register_operand" "0")
	 (neg:SI (plus:SI (match_operand:SI 2 "register_operand" "x")
			  (match_operand:SI 3 "immediate_operand" "K")))))]
  ""
  "ashc %0,%3(%2)")

(define_insn "rotlsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(rotate:SI (match_operand:SI 1 "register_operand" "0,0,0")
		   (match_operand:SI 2 "general_operand" "x,K,m")))]
  ""
  "@
   rot %0,(%2)
   rot %0,%2
   rot %0,@%2")

(define_insn "*ROT_left_plus"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(rotate:SI (match_operand:SI 1 "register_operand" "0")
		   (plus:SI (match_operand:SI 2 "register_operand" "x")
			    (match_operand:SI 3 "immediate_operand" "K"))))]
  ""
  "rot %0,%3(%2)")

(define_expand "rotrsi3"
  [(set (match_dup 3) (neg:SI (match_operand:SI 2 "general_operand" "")))
   (set (match_operand:SI 0 "register_operand" "")
	(rotatert:SI (match_operand:SI 1 "register_operand" "")
		     (neg:SI (match_dup 3))))]
  ""
  "operands[3] = gen_reg_rtx (SImode);")

(define_insn "*ROT_right"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(rotatert:SI (match_operand:SI 1 "register_operand" "0,0,0")
		     (neg:SI (match_operand:SI 2 "general_operand" "x,K,m"))))]
  ""
  "@
   rot %0,(%2)
   rot %0,%2
   rot %0,@%2")

(define_insn "*ROT_right_plus"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(rotatert:SI
	 (match_operand:SI 1 "register_operand" "0")
	 (neg:SI (plus:SI (match_operand:SI 2 "register_operand" "x")
			  (match_operand:SI 3 "immediate_operand" "K")))))]
  ""
  "rot %0,%3(%2)")

(define_insn "rotldi3"
  [(set (match_operand:DI 0 "register_operand" "=r,r,r")
	(rotate:DI (match_operand:DI 1 "register_operand" "0,0,0")
		   (match_operand:SI 2 "general_operand" "x,K,m")))]
  ""
  "@
   rotc %0,(%2)
   rotc %0,%2
   rotc %0,@%2")

(define_insn "*ROTC_left_plus"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(rotate:DI (match_operand:DI 1 "register_operand" "0")
		   (plus:SI (match_operand:SI 2 "register_operand" "x")
			    (match_operand:SI 3 "immediate_operand" "K"))))]
  ""
  "rotc %0,%3(%2)")

(define_expand "rotrdi3"
  [(set (match_dup 3) (neg:SI (match_operand:SI 2 "general_operand" "")))
   (set (match_operand:DI 0 "register_operand" "")
	(rotatert:DI (match_operand:DI 1 "register_operand" "")
		     (neg:SI (match_dup 3))))]
  ""
  "operands[3] = gen_reg_rtx (SImode);")

(define_insn "*ROTC_right"
  [(set (match_operand:DI 0 "register_operand" "=r,r,r")
	(rotatert:DI (match_operand:DI 1 "register_operand" "0,0,0")
		     (neg:SI (match_operand:SI 2 "general_operand" "x,K,m"))))]
  ""
  "@
   rotc %0,(%2)
   rotc %0,%2
   rotc %0,@%2")

(define_insn "*ROTC_right_plus"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(rotatert:DI
	 (match_operand:DI 1 "register_operand" "0")
	 (neg:SI (plus:SI (match_operand:SI 2 "register_operand" "x")
			  (match_operand:SI 3 "immediate_operand" "K")))))]
  ""
  "rotc %0,%3(%2)")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Logical Operations

;; SETZ -- in Data Movement

;; AND -- andsi3

(define_insn "andsi3"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,r,r,r,m")
        (and:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0,0,0,0,0")
                (match_operand:SI 2 "general_operand" "rm,I,N,P,i,r")))]
  ""
  "@
   and %0,%2
   andi %0,%2
   tlz %0,%T2\\t%; and %0,[%2]
   andcmi %0,%C2\\t%; and %0,[%2]
   and %0,[%2]
   andm %2,%0")
;  trz %0,%C2\\t%; and %0,[%2]	Slower than ANDCMI?

(define_insn "*ANDB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(and:SI (match_operand:SI 1 "register_operand" "%0")
		(match_operand:SI 2 "reg_or_mem_operand" "+rm")))
   (set (match_dup 2)
	(and:SI (match_dup 1) (match_dup 2)))]
  ""
  "andb %0,%2")

;; ANDCA

(define_insn "*ANDCA"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,m")
        (and:SI (not:SI (match_operand:SI 1 "register_operand" "0,0,r"))
                (match_operand:SI 2 "general_operand" "rm,I,0")))]
  ""
  "@
   andca %0,%2
   andcai %0,%2
   andcam %2,%0")

(define_insn "*ANDCAB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(and:SI (not:SI (match_operand:SI 1 "register_operand" "0"))
		(match_operand:SI 2 "memory_operand" "+m")))
   (set (match_dup 2)
	(and:SI (not:SI (match_dup 1)) (match_dup 2)))]
  ""
  "andcab %0,%2")

;; SETM -- not useful

;; ANDCM

(define_insn "*ANDCM"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,m")
        (and:SI (not:SI (match_operand:SI 1 "general_operand" "rm,0"))
                (match_operand:SI 2 "register_operand" "0,r")))]
  ""
  "@
   andcm %0,%2
   andcmm %2,%0")

(define_insn "*ANDCMB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(and:SI (not:SI (match_operand:SI 1 "memory_operand" "+m"))
		(match_operand:SI 2 "register_operand" "0")))
   (set (match_dup 1)
	(and:SI (not:SI (match_dup 1)) (match_dup 2)))]
  ""
  "andcmb %0,%1")

;; SETA -- not useful

;; XOR -- xorsi3

(define_insn "xorsi3"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,r,r,m")
        (xor:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0,0,0,0")
                (match_operand:SI 2 "general_operand" "rm,I,L,i,r")))]
  ""
  "@
   xor %0,%2
   xori %0,%2
   tlc %0,%S2\\t%; xor %0,[%2]
   xor %0,[%2]
   xorm %2,%0")

(define_insn "*XORB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(xor:SI (match_operand:SI 1 "register_operand" "0")
		(match_operand:SI 2 "memory_operand" "+m")))
   (set (match_dup 2)
	(xor:SI (match_dup 1) (match_dup 2)))]
  ""
  "xorb %0,%2")

;; IOR -- iorsi3

(define_insn "iorsi3"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,r,r,r,m")
        (ior:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0,0,0,0,0")
                (match_operand:SI 2 "general_operand" "rm,I,L,P,i,r")))]
  ""
  "@
   ior %0,%2
   iori %0,%2
   tlo %0,%S2\\t%; ior %0,[%2]
   orcmi %0,%C2\\t%; ior %0,[%2]
   ior %0,[%2]
   iorm %2,%0")

(define_insn "*IORB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ior:SI (match_operand:SI 1 "register_operand" "0")
		(match_operand:SI 2 "memory_operand" "+m")))
   (set (match_dup 2)
	(ior:SI (match_dup 1) (match_dup 2)))]
  ""
  "iorb %0,%2")

;; ANDCB

(define_insn "*ANCDB"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,m")
        (and:SI (not:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0"))
                (not:SI (match_operand:SI 2 "reg_or_mem_operand" "rm,r"))))]
  ""
  "@
   andcb %0,%2
   andcbm %2,%0")

(define_insn "*ANDCBI"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (and:SI (not:SI (match_operand:SI 1 "register_operand" "%0"))
                (match_operand:SI 2 "immediate_operand" "P")))]
  ""
  "andcbi %0,%C2")

(define_insn "*ANDCBB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(and:SI (not:SI (match_operand:SI 1 "register_operand" "%0"))
		(not:SI (match_operand:SI 2 "memory_operand" "+m"))))
   (set (match_dup 2)
	(and:SI (not:SI (match_dup 1)) (not:SI (match_dup 2))))]
  ""
  "andcbb %0,%2")

;; EQV

(define_insn "*EQV"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,m")
	(not:SI (xor:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0,0")
			(match_operand:SI 2 "general_operand" "rm,I,r"))))]
  ""
  "@
   eqv %0,%2
   eqvi %0,%2
   eqvm %2,%0")

(define_insn "*EQVB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(not:SI (xor:SI (match_operand:SI 1 "register_operand" "0")
			(match_operand:SI 2 "memory_operand" "+m"))))
   (set (match_dup 2)
	(not:SI (xor:SI (match_dup 1) (match_dup 2))))]
  ""
  "eqvb %0,%2")

;; SETCA, SETCM -- one_cmplsi2

(define_insn "one_cmplsi2"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,m,m")
        (not:SI (match_operand:SI 1 "reg_or_mem_operand" "0,rm,r,0")))]
  ""
  "@
   setca %0,
   setcm %0,%1
   setcam %1,%0
   setcmm %0")

(define_insn "*SETCAB"
  [(set (match_operand:SI 1 "memory_operand" "+m")
	(not:SI (match_operand:SI 0 "register_operand" "+r")))
   (set (match_dup 0) (not:SI (match_dup 0)))]
  ""
  "setcab %0,%1")

(define_insn "*SETCMB"
  [(set (match_operand:SI 0 "register_operand" "+r")
	(not:SI (match_operand:SI 1 "memory_operand" "+m")))
   (set (match_dup 1) (not:SI (match_dup 1)))]
  ""
  "setcmb %0,%1")

;; ORCA

(define_insn "*ORCA"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,r,m")
        (ior:SI (not:SI (match_operand:SI 1 "register_operand" "0,0,r"))
                (match_operand:SI 2 "general_operand" "rm,I,0")))]
  ""
  "@
   orca %0,%2
   orcai %0,%2
   orcam %2,%0")

(define_insn "orcab"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ior:SI (not:SI (match_operand:SI 1 "register_operand" "0"))
		(match_operand:SI 2 "memory_operand" "+m")))
   (set (match_dup 2)
	(ior:SI (not:SI (match_dup 1)) (match_dup 2)))]
  ""
  "orcab %0,%2")

;; ORCM

(define_insn "ORCM"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,m")
        (ior:SI (not:SI (match_operand:SI 1 "general_operand" "rm,0"))
                (match_operand:SI 2 "register_operand" "0,r")))]
  ""
  "@
   orcm %0,%2
   orcmm %2,%0")

(define_insn "*ORCMB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ior:SI (not:SI (match_operand:SI 1 "memory_operand" "+m"))
		(match_operand:SI 2 "register_operand" "0")))
   (set (match_dup 1) (ior:SI (not:SI (match_dup 1)) (match_dup 2)))]
  ""
  "orcmb %0,%2")

;; ORCB

(define_insn "*ORCB"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=r,m")
        (ior:SI (not:SI (match_operand:SI 1 "reg_or_mem_operand" "%0,0"))
                (not:SI (match_operand:SI 2 "reg_or_mem_operand" "rm,r"))))]
  ""
  "@
   orcb %0,%2
   orcbm %2,%0")

(define_insn "*ORCBI"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (ior:SI (not:SI (match_operand:SI 1 "register_operand" "%0"))
                (match_operand:SI 2 "immediate_operand" "P")))]
  ""
  "orcbi %0,%C2")

(define_insn "*ORCBB"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ior:SI (not:SI (match_operand:SI 1 "register_operand" "%0"))
		(not:SI (match_operand:SI 2 "memory_operand" "+m"))))
   (set (match_dup 2)
	(ior:SI (not:SI (match_dup 1)) (not:SI (match_dup 2))))]
  ""
  "orcbb %0,%2")

;; SETO -- in Data Movement


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Floating-point Arithmetic

(define_insn "addsf3"
  [(set (match_operand:SF 0 "reg_or_mem_operand" "=r,r,r,m")
	(plus:SF (match_operand:SF 1 "reg_or_mem_operand" "%0,0,0,0")
		 (match_operand:SF 2 "general_operand" "rm,G,F,r")))]
  ""
  "@
   fadr %0,%2
   fadri %0,%G2
   fadr %0,[%2]
   fadrm %2,%0")

(define_insn "*FADRB"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(plus:SF (match_operand:SF 1 "register_operand" "%0")
		 (match_operand:SF 2 "reg_or_mem_operand" "+rm")))
   (set (match_dup 2)
	(plus:SF (match_dup 1) (match_dup 2)))]
  ""
  "fadrb %0,%2")

(define_expand "adddf3"
  [(set (match_operand:DF 0 "register_operand" "")
	(plus:DF (match_operand:DF 1 "register_operand" "")
		 (match_operand:DF 2 "general_operand" "")))]
  "TARGET_KI10up"
  "")

(define_insn "*DFAD"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(plus:DF (match_operand:DF 1 "register_operand" "%0,0")
		 (match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KI10up && !TARGET_GFLOAT"
  "@
   dfad %0,%2
   dfad %0,[%2]")

(define_insn "*GFAD"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(plus:DF (match_operand:DF 1 "register_operand" "%0,0")
		 (match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "@
   gfad %0,%2
   gfad %0,[%2]")

(define_insn "subsf3"
  [(set (match_operand:SF 0 "reg_or_mem_operand" "=r,r,r,m")
	(minus:SF (match_operand:SF 1 "reg_or_mem_operand" "%0,0,0,0")
		  (match_operand:SF 2 "general_operand" "rm,G,F,r")))]
  ""
  "@
   fsbr %0,%2
   fsbri %0,%G2
   fsbr %0,[%2]
   fsbrm %2,%0")

(define_insn "*FSBRB"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(minus:SF (match_operand:SF 1 "register_operand" "0")
		  (match_operand:SF 2 "reg_or_mem_operand" "+rm")))
   (set (match_dup 2)
	(minus:SF (match_dup 1) (match_dup 2)))]
  ""
  "fsbrb %0,%2")

(define_expand "subdf3"
  [(set (match_operand:DF 0 "register_operand" "")
	(minus:DF (match_operand:DF 1 "register_operand" "")
		  (match_operand:DF 2 "general_operand" "")))]
  "TARGET_KI10up"
  "")

(define_insn "*DFSB"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(minus:DF (match_operand:DF 1 "register_operand" "%0,0")
		  (match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KI10up && !TARGET_GFLOAT"
  "@
   dfsb %0,%2
   dfsb %0,[%2]")

(define_insn "*GFSB"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(minus:DF (match_operand:DF 1 "register_operand" "%0,0")
		  (match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "@
   gfsb %0,%2
   gfsb %0,[%2]")

(define_insn "abssf2"
  [(set (match_operand:SF 0 "reg_or_mem_operand" "=r,m,m")
	(abs:SF (match_operand:SF 1 "reg_or_mem_operand" "rm,r,0")))]
  ""
  "@
   movm %0,%1
   movmm %1,%0
   movms %0")

(define_insn "*MOVMSsf"
  [(set (match_operand:SF 0 "register_operand" "=x")
	(abs:SF (match_operand:SF 1 "memory_operand" "+m")))
   (set (match_dup 1)
	(abs:SF (match_dup 1)))]
  ""
  "movms %0,%1")

(define_insn "negsf2"
  [(set (match_operand:SF 0 "reg_or_mem_operand" "=r,m,m")
	(neg:SF (match_operand:SF 1 "reg_or_mem_operand" "rm,r,0")))]
  ""
  "@
   movn %0,%1
   movnm %1,%0
   movns %0")

(define_insn "*MOVNSsf"
  [(set (match_operand:SF 0 "register_operand" "=x")
	(neg:SF (match_operand:SF 1 "memory_operand" "+m")))
   (set (match_dup 1)
	(neg:SF (match_dup 1)))]
  ""
  "movns %0,%1")

(define_insn "negdf2"
  [(set (match_operand:DF 0 "reg_or_mem_operand" "=r,m")
	(neg:DF (match_operand:DF 1 "reg_or_mem_operand" "rm,r")))]
  "TARGET_KI10up"
  "@
   dmovn %0,%1
   dmovnm %1,%0")

(define_insn "mulsf3"
  [(set (match_operand:SF 0 "reg_or_mem_operand" "=r,r,r,m")
	(mult:SF (match_operand:SF 1 "reg_or_mem_operand" "%0,0,0,0")
		 (match_operand:SF 2 "general_operand" "rm,G,F,r")))]
  ""
  "@
   fmpr %0,%2
   fmpri %0,%G2
   fmpr %0,[%2]
   fmprm %2,%0")

(define_insn "*FMPRB"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(mult:SF (match_operand:SF 1 "register_operand" "%0")
		 (match_operand:SF 2 "reg_or_mem_operand" "+rm")))
   (set (match_dup 2)
	(mult:SF (match_dup 1) (match_dup 2)))]
  ""
  "fmprb %0,%2")

(define_expand "muldf3"
  [(set (match_operand:DF 0 "register_operand" "")
	(mult:DF (match_operand:DF 1 "register_operand" "")
		 (match_operand:DF 2 "general_operand" "")))]
  "TARGET_KI10up"
  "")

(define_insn "*DFMP"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(mult:DF (match_operand:DF 1 "register_operand" "%0,0")
		 (match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KI10up && !TARGET_GFLOAT"
  "@
   dfmp %0,%2
   dfmp %0,[%2]")

(define_insn "*GFMP"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(mult:DF (match_operand:DF 1 "register_operand" "%0,0")
		 (match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "@
   gfmp %0,%2
   gfmp %0,[%2]")

(define_insn "divsf3"
  [(set (match_operand:SF 0 "reg_or_mem_operand" "=r,r,r,m")
	(div:SF (match_operand:SF 1 "reg_or_mem_operand" "%0,0,0,0")
		(match_operand:SF 2 "general_operand" "rm,G,F,r")))]
  ""
  "@
   fdvr %0,%2
   fdvri %0,%G2
   fdvr %0,[%2]
   fdvrm %2,%0")

(define_insn "*FDVRB"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(div:SF (match_operand:SF 1 "register_operand" "%0")
		(match_operand:SF 2 "reg_or_mem_operand" "=rm")))
   (set (match_dup 2)
	(div:SF (match_dup 1) (match_dup 2)))]
  ""
  "fdvrb %0,%2")

(define_expand "divdf3"
  [(set (match_operand:DF 0 "register_operand" "")
	(div:DF (match_operand:DF 1 "register_operand" "")
		(match_operand:DF 2 "general_operand" "")))]
  "TARGET_KI10up"
  "")

(define_insn "*DFDV"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(div:DF (match_operand:DF 1 "register_operand" "%0,0")
		(match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KI10up && !TARGET_GFLOAT"
  "@
   dfdv %0,%2
   dfdv %0,[%2]")

(define_insn "*GFDV"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(div:DF (match_operand:DF 1 "register_operand" "%0,0")
		(match_operand:DF 2 "general_operand" "rm,F")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "@
   gfdv %0,%2
   gfdv %0,[%2]")

(define_insn "sqrtsf2"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(sqrt:SF (match_operand:SF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2"
  "extend %0,[sqrt %1]")

(define_insn "sqrtdf2"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(sqrt:DF (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2"
  "")

(define_insn "*DSQRT"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(sqrt:DF (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2 && !TARGET_GFLOAT"
  "extend %0,[dsqrt %1]")

(define_insn "*GSQRT"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(sqrt:DF (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2 && TARGET_GFLOAT"
  "extend %0,[gsqrt %1]")

(define_insn "FSC"
  [(set (match_operand:SF 0 "register_operand" "=r,r")
	(unspec:SF
	 [(match_operand:SF 1 "register_operand" "0,0")
	  (match_operand:SI 2 "nonmemory_operand" "K,r")]
	 UNSPEC_FSC))]
  ""
  "@
   fsc %0,%2
   fsc %0,(%2)")

(define_insn "GFSC"
  [(set (match_operand:DF 0 "register_operand" "=r,r")
	(unspec:DF
	 [(match_operand:DF 1 "register_operand" "0,0")
	  (match_operand:SI 2 "nonmemory_operand" "K,r")]
	 UNSPEC_FSC))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "@
   extend %0,[gfsc %2]
   extend %0,[gfsc (%2)]")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   Floating-point Conversions

(define_insn "fix_truncsfsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(fix:SI (match_operand:SF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KI10up"
  "fix %0,%1")

; (define_insn "fix_truncsfdi2" ...)

(define_expand "fix_truncdfsi2"
  [(set (match_operand:SI 0 "register_operand" "")
	(fix:SI (match_operand:DF 1 "reg_or_mem_operand" "")))]
  "(TARGET_KL10up && TARGET_GFLOAT) || (TARGET_XKL2 && !TARGET_GFLOAT)"
  "")

(define_insn "*DFIX"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(fix:SI (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2 && !TARGET_GFLOAT"
  "extend %0,[dfix %1]")

(define_insn "*GFIX"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(fix:SI (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "extend %0,[gfix %1]")

(define_expand "fix_truncdfdi2"
  [(set (match_operand:DI 0 "register_operand" "")
	(fix:DI (match_operand:DF 1 "reg_or_mem_operand" "")))]
  "(TARGET_KL10up && TARGET_GFLOAT) || (TARGET_XKL2 && !TARGET_GFLOAT)"
  "")

(define_insn "*DDFIX"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(fix:DI (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2 && !TARGET_GFLOAT"
  "extend %0,[ddfix %1]")

(define_insn "*GDFIX"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(fix:DI (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "extend %0,[gdfix %1]")

(define_insn "floatsisf2"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(float:SF (match_operand:SI 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KI10up"
  "fltr %0,%1")

; (define_insn "floatdisf2" ...)

(define_expand "floatsidf2"
  [(set (match_operand:DF 0 "register_operand" "")
	(float:DF (match_operand:SI 1 "reg_or_mem_operand" "")))]
  "(TARGET_KL10up && TARGET_GFLOAT) || (TARGET_XKL2 && !TARGET_GFLOAT)"
  "")

(define_insn "*DFLTR"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(float:DF (match_operand:SI 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2 && !TARGET_GFLOAT"
  "extend %0,[dfltr %1]")

(define_insn "*GFLTR"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(float:DF (match_operand:SI 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "extend %0,[gfltr %1]")

(define_expand "floatdidf2"
  [(set (match_operand:DF 0 "register_operand" "")
	(float:DF (match_operand:DI 1 "reg_or_mem_operand" "")))]
  "(TARGET_KL10up && TARGET_GFLOAT) || (TARGET_XKL2 && !TARGET_GFLOAT)"
  "")

(define_insn "*DDFLTR"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(float:DF (match_operand:DI 1 "reg_or_mem_operand" "rm")))]
  "TARGET_XKL2 && !TARGET_GFLOAT"
  "extend %0,[ddfltr %1]")

(define_insn "*DGFLTR"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(float:DF (match_operand:DI 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "extend %0,[dgfltr %1]")

(define_expand "extendsfdf2"
  [(set (match_operand:DF 0 "register_operand" "")
	(float_extend:DF (match_operand:SF 1 "reg_or_mem_operand" "")))]
  ""
  "")

;; Extend to D-float by just zeroing AC+1.
(define_insn "*extendsfdf2"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(float_extend:DF (match_operand:SF 1 "reg_or_mem_operand" "0")))]
  "!TARGET_GFLOAT"
  "movei %Z0,0")

(define_insn "*GSNGL"
  [(set (match_operand:DF 0 "register_operand" "=r")
	(float_extend:DF (match_operand:SF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "extend %0,[gsngl %1]")

(define_expand "truncdfsf2"
  [(set (match_operand:SF 0 "register_operand" "")
	(float_truncate:SF (match_operand:DF 1 "reg_or_mem_operand" "")))]
  ""
  "")

;; Truncate from D-float by just dropping AC+1.
(define_insn "*truncdfsf2"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(float_truncate:SF (match_operand:DF 1 "reg_or_mem_operand" "0")))]
  "!TARGET_GFLOAT"
  ""
  [(set_attr "length" "0")])

(define_insn "*GDBLE"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(float_truncate:SF (match_operand:DF 1 "reg_or_mem_operand" "rm")))]
  "TARGET_KL10up && TARGET_GFLOAT"
  "extend %0,[gdble %1]")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Pointer Arithmetic

(define_insn "MOVEI"
  [(set (match_operand:PSI 0 "register_operand" "=r,r")
	(unspec:PSI [(match_operand:PSI 1 "address_operand" "r,p")]
		    UNSPEC_ADDRESS))]
  ""
  "@
   move %0,%1\\t%; movpsi
   movei %0,%a1\\t%; movpsi")

(define_insn "MOVSI"
  [(set (zero_extract:PSI (match_operand:PSI 0 "register_operand" "+r,r")
			  (const_int 18)
			  (const_int 0))
	(unspec:PSI [(match_operand:PSI 1 "address_operand" "r,p")]
		    UNSPEC_ADDRESS))
   (set (zero_extract:PSI (match_dup 0) (const_int 18) (const_int 18))
	(const_int 0))]
  ""
  "@
   movs %0,%1\\t%; movpsi
   movsi %0,%a1\\t%; movpsi")

(define_insn "HRRI"
  [(set (zero_extract:PSI (match_operand:PSI 0 "register_operand" "+r,r")
			  (const_int 18)
			  (const_int 18))
	(unspec:PSI [(match_operand:PSI 1 "address_operand" "r,p")]
		    UNSPEC_ADDRESS))]
  ""
  "@
   hrr %0,%1\\t%; movpsi
   hrri %0,%a1\\t%; movpsi")

(define_insn "HRLI"
  [(set (zero_extract:PSI (match_operand:PSI 0 "register_operand" "+r,r")
			  (const_int 18)
			  (const_int 0))
	(unspec:PSI [(match_operand:PSI 1 "address_operand" "r,p")]
		    UNSPEC_ADDRESS))]
  ""
  "@
   hrl %0,%1\\t%; movpsi
   hrli %0,%a1\\t%; movpsi")

(define_insn "ADDI"
  [(set (match_operand:PSI 0 "register_operand" "=r,r")
	(unspec:PSI [(plus:PSI (match_operand:PSI 1 "register_operand" "0,0")
			       (match_operand:PSI 2 "address_operand" "r,p"))]
		    UNSPEC_ADDRESS))]
  ""
  "@
   add %0,%2\\t%; addpsi
   addi %0,%a2\\t%; addpsi")

(define_insn "movpsi"
  [(set (match_operand:PSI 0 "reg_or_mem_operand" "=r,r,r,r,r,m")
        (match_operand:PSI 1 "general_operand" "rm,IS,J,L,i,r"))]
  ""
  "@
   move %0,%1\\t%; movpsi
   movei %0,%1\\t%; movpsi
   movni %0,%N1\\t%; movpsi
   movsi %0,%S1\\t%; movpsi
   move %0,[%1]\\t%; movpsi
   movem %1,%0\\t%; movpsi")

(define_expand "addpsi3"
  [(set (match_operand:PSI 0 "reg_or_mem_operand" "")
	(plus:PSI (match_operand:PSI 1 "reg_or_mem_operand" "")
		  (match_operand:PSI 2 "general_operand" "")))]
  ""
  "pdp10_expand_addpsi3 (operands);
   DONE;")

(define_expand "subpsi3"
  [(set (match_operand:PSI 0 "reg_or_mem_operand" "")
	(minus:PSI (match_operand:PSI 1 "reg_or_mem_operand" "")
		   (match_operand:PSI 2 "general_operand" "")))]
  ""
  "pdp10_expand_subpsi3 (operands);
   DONE;")

(define_insn "*AOSwp"
  [(set (match_operand:PSI 0 "memory_operand" "=m")
	(plus:PSI (match_operand:PSI 1 "memory_operand" "0")
		  (const_int 1)))]
  ""
  "aos %0\\t%; addwp3")

(define_insn "*SOSwp"
  [(set (match_operand:PSI 0 "memory_operand" "=m")
	(plus:PSI (match_operand:PSI 1 "memory_operand" "0")
		  (const_int -1)))]
  ""
  "sos %0\\t%; addwp3")

;; Add to a word pointer.
(define_insn "addwp3"
  [(set (match_operand:PSI 0 "reg_or_mem_operand" "=r,r,r,r,m")
	(plus:PSI (match_operand:PSI 1 "reg_or_mem_operand" "%0,0,0,0,0")
		  (match_operand:PSI 2 "general_operand" "rm,IS,J,i,r")))]
  ""
  "@
   add %0,%2\\t%; addwp3
   addi %0,%2\\t%; addwp3
   subi %0,%N2\\t%; addwp3
   add %0,[%2]\\t%; addwp3
   addm %2,%0\\t%; addwp3")
	 
(define_insn "*addwp3_unspec"
  [(set (match_operand:PSI 0 "register_operand" "=r,r,r,r")
	(plus:PSI (match_operand:PSI 1 "register_operand" "0,0,0,0")
		  (unspec:PSI
		   [(match_operand:PSI 2 "" "r,I,J,i")]
		   UNSPEC_ADDRESS)))]
  ""
  "@
   add %0,%2\\t%; addwp3
   addi %0,%2\\t%; addwp3
   subi %0,%N2\\t%; addwp3
   add %0,[%2]\\t%; addwp3")
	 
(define_insn "subwp3"
  [(set (match_operand:PSI 0 "reg_or_mem_operand" "=r,r,r,r,m")
	(minus:PSI (match_operand:PSI 1 "reg_or_mem_operand" "0,0,0,0,0")
		   (match_operand:PSI 2 "general_operand" "rm,IS,J,i,r")))]
  ""
  "@
   sub %0,%2\\t%; subwp3
   subi %0,%2\\t%; subwp3
   addi %0,%N2\\t%; subwp3
   sub %0,[%2]\\t%; subwp3
   subm %2,%0\\t%; subwp3")

(define_insn "divpsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (div:SI (match_operand:SI 1 "register_operand" "r")
                (match_operand:SI 2 "register_operand" "0")))]
  ""
  "idivm %1,%0\\t%; divpsi3")

(define_insn "ashlpsi3"
  [(set (match_operand:PSI 0 "register_operand" "=r")
	(unspec:PSI [(match_operand:PSI 1 "register_operand" "0")
		     (match_operand:PSI 2 "immediate_operand" "K")]
		    UNSPEC_SHIFT))]
  ""
  "ash %0,%2\\t%; ashlpsi3")

(define_insn "ashrpsi3"
  [(set (match_operand:PSI 0 "register_operand" "=r")
	(unspec:PSI [(match_operand:PSI 1 "register_operand" "0")
		     (match_operand:PSI 2 "immediate_operand" "K")]
		    UNSPEC_SHIFTRT))]
  ""
  "ash %0,%N2\\t%; ashrpsi3")

(define_insn "ashrsipsi3"
  [(set (match_operand:PSI 0 "register_operand" "=r")
	(unspec:PSI
	 [(truncate:PSI (match_operand:SI 1 "register_operand" "0"))
	  (match_operand:PSI 2 "immediate_operand" "K")]
	 UNSPEC_SHIFTRT))]
  ""
  "ash %0,%N2\\t%; ashrsipsi3")

(define_insn "ashlsi3_pointer"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(unspec:SI [(match_operand:SI 1 "register_operand" "0")
		    (match_operand:SI 2 "immediate_operand" "K")]
		   UNSPEC_SHIFT))]
  ""
  "ash %0,%2%; ashrpsi3_pointer")

(define_insn "negpsi2"
  [(set (match_operand:PSI 0 "reg_or_mem_operand" "=r,m,m")
	(neg:PSI (match_operand:PSI 1 "reg_or_mem_operand" "rm,r,0")))]
  ""
  "@
   movn %0,%1\\t%; negpsi2
   movnm %1,%0\\t%; negpsi2
   movns %0\\t%; negpsi2")

(define_insn "IBP"
  [(set (match_operand:PSI 0 "reg_or_mem_operand" "=rm")
	(unspec:PSI
	 [(match_operand:PSI 1 "reg_or_mem_operand" "0")
	  (const_int 1)]
	 UNSPEC_ADJBP))]
  ""
  "ibp %0")

(define_insn "ADJBP"
  [(set (match_operand:PSI 0 "register_operand" "=x")
	(unspec:PSI
	 [(match_operand:PSI 1 "reg_or_mem_operand" "rm")
	  (match_operand:PSI 2 "register_operand" "0")]
	 UNSPEC_ADJBP))]
  "TARGET_KL10up"
  "adjbp %0,%1")

(define_expand "cmppsi"
  [(set (cc0) (compare (match_operand:PSI 0 "register_operand" "")
                       (match_operand:PSI 1 "general_operand" "")))]
  ""
  "pdp10_compare_op0 = operands[0];
   pdp10_compare_op1 = operands[1];
   DONE;")

(define_insn "zero_extendpsisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (zero_extend:SI (match_operand:PSI 1 "register_operand" "r")))]
  ""
  "*if (REGNO (operands[0]) == REGNO (operands[1]))
      return \"\";
    else
      return \"move %0,%1\\t%; zero_extendpsisi2\";")

(define_insn "truncsipsi2"
  [(set (match_operand:PSI 0 "register_operand" "=r")
        (truncate:PSI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "*if (REGNO (operands[0]) == REGNO (operands[1]))
      return \"\";
    else
      return \"move %0,%1\\t%; truncsipsi2\";")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Unconditional Jumps

(define_insn "nop"
  [(const_int 0)]
  ""
  "jfcl\\t%; nop")

(define_insn "*nothing"
  [(const_int 1)]
  ""
  ""
  [(set_attr "length" "0")])

(define_insn "jump"
  [(set (pc)
	(label_ref (match_operand 0 "" "")))]
  ""
  "jrst %l0")

(define_insn "indirect_jump"
  [(set (pc)
	(match_operand:PSI 0 "general_operand" "x,IS,i,rm"))]
  ""
  "@
   jrst (%0)
   jrst %0
   jrst @[%0]
   jrst @%0")

(define_expand "call"
  [(call (match_operand 0 "memory_operand" "")
	 (match_operand 1 "general_operand" ""))]
  ""
  "")

(define_insn "*call_reg"
  [(call (mem:SI (match_operand:PSI 0 "register_operand" "x"))
         (match_operand 1 "" "g"))]
  ""
  "pushj 17,(%0)")

(define_insn "*call_mem1"
  [(call (mem:SI (match_operand:PSI 0 "memory_operand" "m"))
	 (match_operand:SI 1 "general_operand" "g"))]
  ""
  "pushj 17,%0")

(define_insn "*call_mem2"
  [(call (mem:SI (match_operand 0 "immediate_operand" "i"))
	 (match_operand 1 "general_operand" "g"))]
  "GET_CODE (operands[0]) != SYMBOL_REF"
  "pushj 17,%a0")

(define_insn "*call_symbol"
  [(call (mem:SI (match_operand:PSI 0 "" "X"))
	 (match_operand:SI 1 "general_operand" "g"))]
  "GET_CODE (operands[0]) == SYMBOL_REF"
  "pushj 17,%L0")

(define_expand "call_value"
  [(set (match_operand 0 "" "")
	(call (match_operand 1 "memory_operand" "")
	      (match_operand 2 "general_operand" "")))]
  ""
  "")

(define_insn "*call_value_reg"
  [(set (match_operand 0 "" "=r")
        (call (mem:SI (match_operand:PSI 1 "register_operand" "x"))
	      (match_operand:SI 2 "general_operand" "g")))]
  ""
  "pushj 17,(%1)")

(define_insn "*call_value_mem1"
  [(set (match_operand 0 "" "=r")
	(call (mem:SI (match_operand 1 "memory_operand" "m"))
	      (match_operand 2 "general_operand" "g")))]
  "CONSTANT_ADDRESS_P (XEXP (operands[1], 0))"
  "pushj 17,%1")

(define_insn "*call_value_mem2"
  [(set (match_operand 0 "" "=r")
	(call (mem:SI (match_operand 1 "immediate_operand" "i"))
	      (match_operand 2 "general_operand" "g")))]
  "GET_CODE (operands[1]) != SYMBOL_REF"
  "pushj 17,%a1")

(define_insn "*call_value_symbol"
  [(set (match_operand 0 "register_operand" "=r")
	(call (mem:SI (match_operand:PSI 1 "" "X"))
	      (match_operand:SI 2 "general_operand" "g")))]
  "GET_CODE (operands[1]) == SYMBOL_REF"
  "pushj 17,%L1")

(define_insn "untyped_call"
  [(parallel [(call (match_operand 0 "" "")
		    (const_int 0))
	      (match_operand 1 "" "")
	      (match_operand 2 "" "")])]
  ""
  "untypedcall %0, %1, %2")

; (define_insn "untyped_return"
;   ...)

; (define_insn "trap"
;   [(trap_if (const_int 1) (const_int 0))]
;   ""
;   "<trap 0>\\t%; trap")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Conditional Jumps

(define_expand "cmpsi"
  [(set (cc0) (compare (match_operand:SI 0 "register_operand" "")
                       (match_operand:SI 1 "general_operand" "")))]
  ""
  "pdp10_compare_op0 = operands[0];
   pdp10_compare_op1 = operands[1];
   DONE;")

(define_expand "cmpsf"
  [(set (cc0) (compare (match_operand:SF 0 "register_operand" "")
                       (match_operand:SF 1 "general_operand" "")))]
  ""
  "pdp10_compare_op0 = operands[0];
   pdp10_compare_op1 = operands[1];
   DONE;")

(define_expand "cmpdf"
  [(set (match_dup 2)
	(minus:DF (match_operand:DF 0 "register_operand" "")
		  (match_operand:DF 1 "general_operand" "")))]
  ""
  "operands[2] = gen_reg_rtx (DFmode);
   pdp10_compare_op0 = operands[2];
   pdp10_compare_op1 = CONST0_RTX (DFmode);")

(define_insn "*test_and_skip"
  [(set (pc)
	(if_then_else
	 (match_operator 0 "pdp10_equality_operator"
	  [(and:SI (match_operand:SI 1 "register_operand" "r,r,r,r")
		   (match_operand:SI 2 "general_operand" "I,L,i,rm"))
	   (const_int 0)])
	 (label_ref (match_operand 3 "" ""))
	 (pc)))]
  ""
  "*{
      const char *eq[][2] =
      { { \"trn%0 %1,%2\",	\"trn%R0 %1,%2\; jrst %l3\" },
	{ \"tln%0 %1,%S2\",	\"tln%R0 %1,%S2\; jrst %l3\" },
	{ \"tdn%0 %1,[%2]\",	\"tdn%R0 %1,[%2]\; jrst %l3\" },
	{ \"tdn%0 %1,%2\",	\"tdn%R0 %1,%2\; jrst %l3\" } };
      const char *ne[][2] =
      { { \"trn%R0 %1,%2\",	\"trn%0 %1,%2\; jrst %l3\" },
	{ \"tln%R0 %1,%S2\",	\"tln%0 %1,%S2\; jrst %l3\" },
	{ \"tdn%R0 %1,[%2]\",	\"tdn%0 %1,[%2]\; jrst %l3\" },
	{ \"tdn%R0 %1,%2\",	\"tdn%0 %1,%2\; jrst %l3\" } };
      if (GET_CODE (operands[0]) == EQ)
	return eq[which_alternative][get_attr_length (insn) - 1];
      else
	return ne[which_alternative][get_attr_length (insn) - 1];
    }"
  [(set (attr "length")
	(if_then_else (eq (minus (match_dup 3) (pc)) (const_int 1))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "yes,yes,yes,yes")])

;; Equality comparisons of a group of bits with zero will be written
;; using zero_extract rather than the equivalent AND operations.
(define_insn "test_extract_skip"
  [(set (pc)
	(if_then_else
	 (match_operator 0 "pdp10_equality_operator"
	  [(zero_extract:SI
	    (match_operand:SI 1 "register_operand" "r,r,r,r")
	    (match_operand:SI 2 "const_int_operand" "")
	    (match_operand:SI 3 "const_int_operand" ""))
	   (const_int 0)])
	 (label_ref (match_operand 4 "" ""))
	 (pc)))]
  ""
  "*{
      const char *eq[][2] =
      { { \"trn%0 %1,%2\",	\"trn%R0 %1,%2\; jrst %l4\" },
	{ \"tln%0 %1,%S2\",	\"tln%R0 %1,%S2\; jrst %l4\" },
	{ \"tdn%0 %1,[%2]\",	\"tdn%R0 %1,[%2]\; jrst %l4\" },
	{ \"tdn%0 %1,%2\",	\"tdn%R0 %1,%2\; jrst %l4\" } };
      const char *ne[][2] =
      { { \"trn%R0 %1,%2\",	\"trn%0 %1,%2\; jrst %l4\" },
	{ \"tln%R0 %1,%S2\",	\"tln%0 %1,%S2\; jrst %l4\" },
	{ \"tdn%R0 %1,[%2]\",	\"tdn%0 %1,[%2]\; jrst %l4\" },
	{ \"tdn%R0 %1,%2\",	\"tdn%0 %1,%2\; jrst %l4\" } };
      int len = INTVAL (operands[2]) - 1;
      int pos = 36 - INTVAL (operands[3]) - len - 1;
      unsigned HOST_WIDE_INT mask = (2 << len) - 1;
      operands[2] = GEN_INT (mask << pos);
      if (GET_CODE (operands[0]) == EQ)
	return eq[which_alternative][get_attr_length (insn) - 1];
      else
	return ne[which_alternative][get_attr_length (insn) - 1];
    }"
  [(set (attr "length")
	(if_then_else (eq (minus (match_dup 4) (pc)) (const_int 1))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "yes,yes,yes,yes")])

(define_expand "bne"
  [(set (pc)
	(if_then_else (ne (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;")

(define_expand "beq"
  [(set (pc)
	(if_then_else (eq (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;")

(define_expand "bge"
  [(set (pc)
	(if_then_else (ge (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;")

(define_expand "bgt"
  [(set (pc)
	(if_then_else (gt (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;")

(define_expand "ble"
  [(set (pc)
	(if_then_else (le (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;")

(define_expand "blt"
  [(set (pc)
	(if_then_else (lt (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;")

(define_expand "bgeu"
  [(set (match_dup 3) (xor:SI (match_dup 1) (const_int SIGNBIT)))
   (set (match_dup 4) (xor:SI (match_dup 2) (const_int SIGNBIT)))
   (set (pc)
	(if_then_else (ge (match_dup 3) (match_dup 4))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;
   operands[3] = gen_reg_rtx (SImode);
   operands[4] = gen_reg_rtx (SImode);")

(define_expand "bgtu"
  [(set (match_dup 3) (xor:SI (match_dup 1) (const_int SIGNBIT)))
   (set (match_dup 4) (xor:SI (match_dup 2) (const_int SIGNBIT)))
   (set (pc)
	(if_then_else (gt (match_dup 3) (match_dup 4))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;
   operands[3] = gen_reg_rtx (SImode);
   operands[4] = gen_reg_rtx (SImode);")

(define_expand "bleu"
  [(set (match_dup 3) (xor:SI (match_dup 1) (const_int SIGNBIT)))
   (set (match_dup 4) (xor:SI (match_dup 2) (const_int SIGNBIT)))
   (set (pc)
	(if_then_else (le (match_dup 3) (match_dup 4))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;
   operands[3] = gen_reg_rtx (SImode);
   operands[4] = gen_reg_rtx (SImode);")

(define_expand "bltu"
  [(set (match_dup 3) (xor:SI (match_dup 1) (const_int SIGNBIT)))
   (set (match_dup 4) (xor:SI (match_dup 2) (const_int SIGNBIT)))
   (set (pc)
	(if_then_else (lt (match_dup 3) (match_dup 4))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "operands[1] = pdp10_compare_op0;
   operands[2] = pdp10_compare_op1;
   operands[3] = gen_reg_rtx (SImode);
   operands[4] = gen_reg_rtx (SImode);")

; (define_insn "cbranchsi4"
;   [(set (pc)
; 	(if_then_else (match_operator 0 "pdp10_comparison_operator"
; 		       [(match_operand:SI 1 "reg_or_mem_operand" "r,m,r,r,r")
; 		        (match_operand:SI 2 "general_operand" "O,O,I,i,rm")])
; 		      (label_ref (match_operand 3 "" ""))
; 		      (pc)))]
;   ""
;   "*{
;       const char *asms[][2] =
;       { { \"jump%0 %1,%l3\",	0 },
; 	{ \"skip%0 %1\",	\"skip%R0 %1\; jrst %l3\" },
; 	{ \"cai%0 %1,%2\",	\"cai%R0 %1,%2\; jrst %l3\" },
; 	{ \"cam%0 %1,[%2]\",	\"cam%R0 %1,[%2]\; jrst %l3\" },
; 	{ \"cam%0 %1,%2\",	\"cam%R0 %1,%2\; jrst %l3\" } };
;       return asms[which_alternative][get_attr_length (insn) - 1];
;     }"
;   [(set (attr "length")
; 	(if_then_else (ior (eq_attr "alternative" "0")
; 			   (eq (minus (match_dup 3) (pc)) (const_int 1)))
; 		      (const_int 1)
; 		      (const_int 2)))
;    (set_attr "skip" "no,yes,yes,yes,yes")])

(define_insn "*cbranchsi"
  [(set (pc)
	(if_then_else (match_operator 3 "comparison_operator"
		       [(match_operand:SI 1 "reg_or_mem_operand" "r,m,r,r,r")
		        (match_operand:SI 2 "general_operand" "O,O,I,i,rm")])
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*{
      const char *asms[][2] =
      { { \"jump%3 %1,%l0\",	0 },
	{ \"skip%3 %1\",	\"skip%R3 %1\; jrst %l0\" },
	{ \"cai%3 %1,%2\",	\"cai%R3 %1,%2\; jrst %l0\" },
	{ \"cam%3 %1,[%2]\",	\"cam%R3 %1,[%2]\; jrst %l0\" },
	{ \"cam%3 %1,%2\",	\"cam%R3 %1,%2\; jrst %l0\" } };
      return asms[which_alternative][get_attr_length (insn) - 1];
    }"
  [(set (attr "length")
	(if_then_else (ior (eq_attr "alternative" "0")
			   (eq (minus (match_dup 0) (pc)) (const_int 1)))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "no,yes,yes,yes,yes")])

(define_insn "*cbranchpsi"
  [(set (pc)
	(if_then_else (match_operator 3 "pdp10_comparison_operator"
		       [(match_operand:PSI 1 "reg_or_mem_operand" "r,m,r,r")
		        (match_operand:PSI 2 "general_operand" "O,O,I,rm")])
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*{
      const char *asms[][2] =
      { { \"jump%3 %1,%l0\",	0 },
	{ \"skip%3 %1\",	\"skip%R3 %1\; jrst %l0\" },
	{ \"cai%3 %1,%2\",	\"cai%R3 %1,%2\; jrst %l0\" },
	{ \"cam%3 %1,%2\",	\"cam%R3 %1,%2\; jrst %l0\" } };
      return asms[which_alternative][get_attr_length (insn) - 1];
    }"
  [(set (attr "length")
	(if_then_else (ior (eq_attr "alternative" "0")
			   (eq (minus (match_dup 0) (pc)) (const_int 1)))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "no,yes,yes,yes")])

(define_insn "*cbranchsf"
  [(set (pc)
	(if_then_else (match_operator 3 "pdp10_comparison_operator"
		       [(match_operand:SF 1 "reg_or_mem_operand" "r,m,r,r,r")
		        (match_operand:SF 2 "general_operand" "O,O,I,i,rm")])
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*{
      const char *asms[][2] =
      { { \"jump%3 %1,%l0\",	0 },
	{ \"skip%3 %1\",	\"skip%R3 %1\; jrst %l0\" },
	{ \"cai%3 %1,%2\",	\"cai%R3 %1,%2\; jrst %l0\" },
	{ \"cam%3 %1,[%2]\",	\"cam%R3 %1,[%2]\; jrst %l0\" },
	{ \"cam%3 %1,%2\",	\"cam%R3 %1,%2\; jrst %l0\" } };
      return asms[which_alternative][get_attr_length (insn) - 1];
    }"
  [(set (attr "length")
	(if_then_else (ior (eq_attr "alternative" "0")
			   (eq (minus (match_dup 0) (pc)) (const_int 1)))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "no,yes,yes,yes,yes")])

(define_insn "*cbranchdf"
  [(set (pc)
	(if_then_else
	 (match_operator 3 "pdp10_comparison_operator"
          [(match_operand:DF 1 "reg_or_mem_operand" "r,m")
	   (match_operand:DF 2 "pdp10_const_double_0_operand" "")])
	 (label_ref (match_operand 0 "" ""))
	 (pc)))]
  ""
  "*if (which_alternative == 0)
    	return \"jump%3 %1,%l0\";
    else if (get_attr_length (insn) == 1)
    	return \"skip%3 %1\";
    else
    	return \"skip%R3 %1\; jrst %l0\";"
  [(set (attr "length")
	(if_then_else (ior (eq_attr "alternative" "0")
			   (eq (minus (match_dup 0) (pc)) (const_int 1)))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "no,yes")])

(define_insn "doloop_end"
  [(set (pc)
	(if_then_else
	 (ge (plus:SI (match_operand:SI 0 "reg_or_mem_operand" "+r")
		      (const_int -1))
	     (const_int 0))
	 (label_ref (match_operand 4 "" ""))
	 (pc)))
   (set (match_dup 0)
	(plus:SI (match_dup 0) (const_int -1)))
   (use (match_operand 1 "" ""))	; iterations; zero if unknown
   (use (match_operand 2 "" ""))	; max iterations
   (use (match_operand 3 "" ""))]	; loop level
  ""
  "sojge %0,%l4\\t%; doloop_end")

(define_insn "decrement_and_branch_until_zero"
  [(set (pc)
	(if_then_else
	 (match_operator 2 "pdp10_comparison_operator"
	  [(plus:SI (match_operand:SI 0 "reg_or_mem_operand" "+r,m")
		    (const_int -1))
	   (const_int 0)])
	 (label_ref (match_operand 1 "" ""))
	 (pc)))
   (set (match_dup 0)
	(plus:SI (match_dup 0) (const_int -1)))]
  ""
  "*if (which_alternative == 0)
      return \"soj%2 %0,%l1\\t%; decrement_and_branch_until_zero\";
    else if (get_attr_length (insn) == 1)
      return \"sos%2 %0\\t%; decrement_and_branch_until_zero\";
    else
      return \"sos%R2 %0\; jrst %l1\\t%; decrement_and_branch_until_zero\";"
  [(set (attr "length")
	(if_then_else (ior (eq_attr "alternative" "0")
			   (eq (minus (match_dup 1) (pc)) (const_int 1)))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "no,yes")])

(define_insn "*AOJ"
  [(set (pc)
	(if_then_else
	 (match_operator 0 "pdp10_comparison_operator"
	  [(plus:SI (match_operand:SI 1 "reg_or_mem_operand" "+r,m")
		    (const_int 1))
	   (const_int 0)])
	 (label_ref (match_operand 2 "" ""))
	 (pc)))
   (set (match_dup 1)
	(plus:SI (match_dup 1) (const_int 1)))]
  ""
  "*if (which_alternative == 0)
      return \"aoj%0 %1,%l2\";
    else if (get_attr_length (insn) == 1)
      return \"sos%0 %1\";
    else
      return \"sos%R0 %1\; jrst %l2\";"
  [(set (attr "length")
	(if_then_else (ior (eq_attr "alternative" "0")
			   (eq (minus (match_dup 2) (pc)) (const_int 1)))
		      (const_int 1)
		      (const_int 2)))
   (set_attr "skip" "no,yes")])

(define_insn "JFFO"
  [(set (pc)
	(if_then_else (ne (subreg:SI
			   (match_operand:DI 0 "register_operand" "+r") 0)
			  (const_int 0))
		      (match_operand 1 "address_operand" "")
		      (pc)))
   (set (subreg:SI (match_dup 0) 1)
	(if_then_else:SI (ne (subreg:SI (match_dup 0) 0) (const_int 0))
			 (unspec:SI [(subreg:SI (match_dup 0) 0)] UNSPEC_FFO)
			 (const_int 0)))]
  "TARGET_KA10up"
  "*if (GET_CODE (operands[1]) == LABEL_REF)
      return \"jffo %0,%l1\";
    else
      return \"jffo %0,%1\";")

; (define_expand "conditional_trap"
;   [(trap_if (match_operator 0 "pdp10_comparison_operator" 
;              [(match_dup 2) (const_int 0)])
;             (match_operand 1 "const_int_operand" ""))]
;   ""
;   "operands[2] = pdp10_compare_op0;")

; (define_insn "*trap"
;   [(trap_if (match_operator 0 "pdp10_comparison_operator" 
;              [(match_operand 2 "memory_operand" "m") (const_int 0)])
;             (match_operand 1 "const_int_operand" "i"))]
;   ""
;   "skip%R0 %2\; <trap %1>\\t%; conditional_trap"
;   [(set_attr "length" "2")])


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Function Prologue and Epilogue

(define_expand "prologue"
  [(const_int 2)]
  ""
  "pdp10_expand_prologue ();
   DONE;")

(define_expand "epilogue"
  [(return)]
  ""
  "pdp10_expand_epilogue ();
   DONE;")

;(define_expand "sibcall_epilogue"
;  [(const_int 3)]
;  ""
;  "return pdp10_expand_sibcall_epilogue ();")

(define_insn "POPJ"
  [(set (pc) (match_operand:PSI 0 "pdp10_pop_operand" ""))]
  ""
  "popj %P0,")

(define_insn "POPJ_17"
  [(return)]
  ""
  "popj 17,")

(define_expand "allocate_stack"
  [(set (match_operand:PSI 0 "register_operand" "")
	(reg:PSI SP_REGNUM))
   (set (reg:PSI SP_REGNUM)
	(unspec:PSI
	 [(reg:PSI SP_REGNUM)
	  (match_operand:SI 1 "general_operand" "")]
	 UNSPEC_ADJSP))]
  ""
  "if (GET_CODE (operands[1]) == CONST_INT)
     operands[1] = GEN_INT (INTVAL (operands[1]) / 4);
   else
     emit_insn (gen_ASH_right (operands[1], operands[1], GEN_INT (-2)));")

(define_expand "ADJSP"
  [(set (match_operand:PSI 0 "register_operand" "")
	(unspec:PSI [(match_dup 0) (match_operand:SI 1 "general_operand" "")]
		    UNSPEC_ADJSP))]
  ""
  "")

(define_insn "*ADJSP_KL10up"
  [(set (match_operand:PSI 0 "register_operand" "+r,r,r,r")
	(unspec:PSI [(match_dup 0)
		     (match_operand:SI 1 "general_operand" "K,i,r,m")]
		    UNSPEC_ADJSP))]
  "TARGET_KL10up"
  "@
   adjsp %0,%1
   adjsp %0,[%1]
   adjsp %0,(%1)
   adjsp %0,@%1")

(define_insn "*ADJSP_KI10down"
  [(set (match_operand:PSI 0 "register_operand" "+r")
	(unspec:PSI
	 [(match_dup 0) (match_operand:SI 1 "immediate_operand" "K")]
	 UNSPEC_ADJSP))]
  "!TARGET_KL10up && !TARGET_EXTENDED"
  "add %0,[%1,,%1]")

(define_insn "PUSH"
  [(set (match_operand:SI 0 "pdp10_push_operand" "=m")
	(match_operand:SI 1 "reg_or_mem_operand" "rm"))]
  ""
  "push %P0,%1")

(define_insn "POP"
  [(set (match_operand:SI 0 "reg_or_mem_operand" "=rm")
	(match_operand:SI 1 "pdp10_pop_operand" "m"))]
  ""
  "pop %P1,%0")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Miscellaneous

(define_insn "*movqi_subreg"
  [(set (match_operand:QI 0 "register_operand" "=r")
	(subreg:QI (match_operand:SI 1 "register_operand" "r") 0))]
  ""
  "move %0,%1\\t%; movqi_subreg")

;; The following four insns compensate for unnecessary shifts
;; generated by pointer arithmetic.

(define_insn "*combine_nop_shift_left_right"
  [(set (match_operand:PSI 0 "register_operand" "=r")
	(unspec:PSI
	 [(truncate:PSI
	   (ashift:SI (match_operand:SI 1 "register_operand" "0")
		      (match_operand:SI 2 "const_int_operand" "")))
	  (match_operand:SI 3 "const_int_operand" "")]
	 UNSPEC_SHIFTRT))]
  "INTVAL (operands[2]) == INTVAL (operands[3])"
  ""
  [(set_attr "length" "0")])

(define_insn "*combine_shift_left_right"
  [(set (match_operand:PSI 0 "register_operand" "=r")
	(unspec:PSI
	 [(truncate:PSI
	   (ashift:SI (match_operand:SI 1 "register_operand" "0")
		      (match_operand:SI 2 "const_int_operand" "")))
	  (match_operand:SI 3 "const_int_operand" "")]
	 UNSPEC_SHIFTRT))]
  ""
  "*operands[2] = GEN_INT (INTVAL (operands[2]) - INTVAL (operands[3]));
    return \"ash %0,%2\";")

(define_insn "*combine_nop_shift_right_left"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI
	 (unspec:SI [(match_operand:SI 1 "register_operand" "0")
		     (match_operand:SI 2 "const_int_operand" "")]
		    UNSPEC_SHIFT)
	 (match_operand:SI 3 "const_int_operand" "")))]
  "INTVAL (operands[2]) == INTVAL (operands[3])"
  ""
  [(set_attr "length" "0")])

(define_insn "*combine_shift_right_left"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI
	 (unspec:SI [(match_operand:SI 1 "register_operand" "0")
		     (match_operand:SI 2 "const_int_operand" "")]
		    UNSPEC_SHIFT)
	 (match_operand:SI 3 "const_int_operand" "")))]
  ""
  "*operands[2] = GEN_INT (INTVAL (operands[2]) - INTVAL (operands[3]));
    return \"ash %0,%2\";")

;; The following two peepholes remove unnecessary shifts generated by
;; pointer arithmetic.

(define_peephole2
  [(set (match_operand:PSI 0 "register_operand" "=r")
	(unspec:PSI
	 [(truncate:PSI
	   (ashift:SI (match_operand:SI 1 "register_operand" "0")
		      (match_operand:SI 2 "const_int_operand" "")))
	  (match_operand:SI 3 "const_int_operand" "")]
	 UNSPEC_SHIFTRT))]
  "INTVAL (operands[2]) == INTVAL (operands[3])"
  [(set (match_dup 0) (truncate:PSI (match_dup 1)))])

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI
	 (unspec:SI [(match_operand:SI 1 "register_operand" "0")
		     (match_operand:SI 2 "const_int_operand" "")]
		    UNSPEC_SHIFT)
	 (match_operand:SI 3 "const_int_operand" "")))]
  "INTVAL (operands[2]) == INTVAL (operands[3])"
  [(set (match_dup 0) (match_dup 1))])

;; The following two peepholes generate EXCH instructions.

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "+r")
	(match_operand:SI 1 "register_operand" "+r"))
   (set (match_dup 1)
	(match_operand:SI 2 "reg_or_mem_operand" "+rm"))
   (set (match_dup 2)
	(match_dup 0))]
  "dead_or_set_p (insn, operands[0])"
  [(parallel [(set (match_dup 1) (match_dup 2))
	      (set (match_dup 2) (match_dup 1))])])

(define_peephole2
  [(set (match_operand:SI 0 "register_operand" "+r")
	(match_operand:SI 1 "reg_or_mem_operand" "+rm"))
   (set (match_dup 1)
	(match_operand:SI 2 "register_operand" "+r"))
   (set (match_dup 2)
	(match_dup 0))]
  "dead_or_set_p (insn, operands[0])"
  [(parallel [(set (match_dup 2) (match_dup 1))
	      (set (match_dup 1) (match_dup 2))])])

;; This peephole removes an unneccesary MOVEI instruction.

(define_peephole2
  [(set (match_operand:PSI 0 "register_operand" "")
        (unspec:PSI [(plus:PSI (match_dup 0)
			       (match_operand 1 "const_int_operand" ""))]
		    UNSPEC_ADDRESS))
   (parallel [(unspec_volatile:BLK
	       [(match_operand:PSI 2 "register_operand" "")]
	       VUNSPEC_BLT)
	      (use (mem:BLK (match_dup 0)))])]
  "dead_or_set_p (insn, operands[0])"
  [(parallel [(unspec_volatile:BLK [(match_dup 2)] VUNSPEC_BLT)
	      (use (mem:BLK (plus:PSI (match_dup 0) (match_dup 1))))])])

;; This peephole optimizes a block-clear operation with a constant address.

(define_peephole2
  [(set (match_operand:PSI 0 "register_operand" "")
	(match_operand:PSI 1 "immediate_operand" "IS"))
   (parallel
    [(set (zero_extract:PSI (match_operand:PSI 2 "register_operand" "")
			    (const_int 18)
			    (const_int 0))
	  (unspec:PSI [(match_dup 0)] UNSPEC_ADDRESS))
     (set (zero_extract:PSI (match_dup 2) (const_int 18) (const_int 18))
	  (const_int 0))])
   (set (zero_extract:PSI (match_dup 2) (const_int 18) (const_int 18))
	(unspec:PSI [(plus:PSI (match_dup 0) (const_int 4))] UNSPEC_ADDRESS))]
  "dead_or_set_p (insn, operands[0])"
  [(parallel
    [(set (zero_extract:PSI (match_dup 2) (const_int 18) (const_int 0))
	  (match_dup 1))
     (set (zero_extract:PSI (match_dup 2) (const_int 18) (const_int 18))
	  (const:PSI (plus:PSI (match_dup 1) (const_int 4))))])])
